/* ChanServ routines

   Cygnus IRC Services - Copyright (c) 2001-2002 Darcy Grexton
   Contact: skold@habber.net, skold @ HabberNet, ICQ #6285178

   This program is free but copyrighted software; see the file COPYING
   for details.
 */

#include "../inc/services.h"

ChanInfo *cs_chanlist[1024];

static void do_help (const char *source);
static void do_register (const char *source);
static void do_identify (const char *source);
static void do_drop (const char *source);
static void do_info (const char *source);
static void do_auth (const char *source);
static void do_set (const char *source);
static void do_count (const char *source);
static void do_vop (const char *source);
#ifdef UNREAL3
static void do_hop (const char *source);
#endif
static void do_aop (const char *source);
static void do_sop (const char *source);
static void do_op (const char *source);
static void do_deop (const char *source);
#ifdef UNREAL3
static void do_halfop (const char *source);
static void do_dehalfop (const char *source);
#endif
static void do_voice (const char *source);
static void do_devoice (const char *source);
static void do_akick (const char *source);
static void do_freeze (const char *source);
static void do_unfreeze (const char *source);
static void do_getpass (const char *source);
static void do_setpass (const char *source);
static void do_sendpass (const char *source);

/* Main ChanServ routine. */
void chanserv (const char *source, char *buf)  
{
    char *cmd, *s;

    cmd = strtok (buf, " ");

    if (!cmd)
	return;
    
    if (!stricmp (cmd, "\1PING"))
    {
	if (!(s = strtok (NULL, "")))
	    s = "0";
	strip (s);
	ctcpreply (s_ChanServ, source, "PING %s", s);
	return;
    }
    else if (!stricmp (cmd, "\1VERSION\1"))
    {
	ctcpreply (s_ChanServ, source, "VERSION Cygnus IRC Services v%s",
	    version);
	return;
    }

    /* A CTCP we don't recognize.. Ignore it. */
    else if (*cmd == '\1')
	return;

    else
    {
	Hash *command, hash_table[] =
	{
	    {"HELP",		H_NONE,		do_help},
	    {"REGISTER",	H_NONE,		do_register},
	    {"IDENTIFY",	H_NONE,		do_identify},
	    {"ID",		H_NONE,		do_identify},
	    {"INFO",		H_NONE,		do_info},
	    {"DROP",		H_NONE,		do_drop},
	    {"SET",		H_NONE,		do_set},
	    {"COUNT",		H_NONE,		do_count},
	    {"VOP",		H_NONE,		do_vop},
#ifdef UNREAL3
	    {"HOP",		H_NONE,		do_hop},
#endif
	    {"AOP",		H_NONE,		do_aop},
	    {"SOP",		H_NONE,		do_sop},
	    {"OP",		H_NONE,		do_op},
	    {"DEOP",		H_NONE,		do_deop},
#ifdef UNREAL3
	    {"HALFOP",		H_NONE,		do_halfop},
	    {"DEHALFOP",	H_NONE,		do_dehalfop},
#endif
	    {"VOICE",		H_NONE,		do_voice},
	    {"DEVOICE",		H_NONE,		do_devoice},
	    {"AKICK",		H_NONE,		do_akick},
	    {"FREEZE",		H_CSOP,		do_freeze},
	    {"UNFREEZE",	H_CSOP,		do_unfreeze},
	    {"GETPASS",		H_CSOP,		do_getpass},
	    {"SETPASS",		H_CSOP,		do_setpass},
	    {"SENDPASS",	H_CSOP,		do_sendpass},
	    {"AUTH",		H_CSOP,		do_auth},
	    {NULL}
	};
            
	if ((command = get_hash (s_ChanServ, source, strupper (cmd),
				 hash_table)))
	{
	    if (csregistertype != 4 && command->process == do_auth)
		errmoreinfo (s_ChanServ, source, "AUTH");
	    else
		(*command->process) (source);
	}
    }
}

/* Return a help message. */
static void do_help (const char *source) 
{
    char *cmd = strtok (NULL, "");
            
    if (!cmd)
	chanserv_help_index (source);
    else
    {
	Hash_HELP *command, hash_table[] =
	{
	    {"REGISTER",	H_NONE,		chanserv_help_register},
	    {"DROP",		H_NONE,		chanserv_help_drop},
	    {"INFO",		H_NONE,		chanserv_help_info},
	    {"IDENTIFY",	H_NONE,		chanserv_help_identify},
	    {"SET",		H_NONE,		chanserv_help_set},
	    {"COUNT",		H_NONE,		chanserv_help_count},
	    {"VOP",		H_NONE,		chanserv_help_vop},
#ifdef UNREAL3
	    {"HOP",		H_NONE,		chanserv_help_hop},
#endif
	    {"AOP",		H_NONE,		chanserv_help_aop},
	    {"SOP",		H_NONE,		chanserv_help_sop},
	    {"OP",		H_NONE,		chanserv_help_op},
	    {"DEOP",		H_NONE,		chanserv_help_deop},
#ifdef UNREAL3
	    {"HALFOP",		H_NONE,		chanserv_help_halfop},
	    {"DEHALFOP",	H_NONE,		chanserv_help_dehalfop},
#endif
	    {"VOICE",		H_NONE,		chanserv_help_voice},
	    {"DEVOICE",		H_NONE,		chanserv_help_devoice},
	    {"AKICK",		H_NONE,		chanserv_help_akick},
	    {"FREEZE",		H_CSOP,		chanserv_help_freeze},
	    {"UNFREEZE",	H_CSOP,		chanserv_help_unfreeze},
	    {"GETPASS",		H_CSOP,		chanserv_help_getpass},
	    {"SETPASS",		H_CSOP,		chanserv_help_setpass},
	    {"SENDPASS",	H_CSOP,		chanserv_help_sendpass},
	    {"AUTH",		H_CSOP,		chanserv_help_auth},
	    {NULL}
	};

	if ((command = get_help_hash (s_ChanServ, source, strupper (cmd),
				      hash_table)))
	    (*command->process) (source);
    }
}

/* Register a channel. */
static void do_register (const char *source)
{
    ChanInfo *ci, **list;
    ChanAccess *ca;
    NickInfo *ni;
    Channel *c;
    User *u = finduser (source);
    time_t now = time (NULL);
    char *param1 = strtok (NULL, " ");
    char *param2 = strtok (NULL, " ");
    char *param3 = strtok (NULL, " ");
    int i, reason = 0;

    if (!u)
	return;

    /* Make sure this nick is fully registered */
    ni = get_nick (u);

    if (!ni)
    {
	notice (s_ChanServ, source, ERR_NEED_IDENT, "REGISTER");
	return;
    }

    if (!check_ns_auth (s_ChanServ, source, ni))
	return;

    /* Check here if the're completing registration from E-Mail AUTH */
    if (param1 && param2 && (csregistertype == 1 || csregistertype == 3))
    {
	ci = cs_findchan (param1);
	c = findchan (param1);

	/* Is this channel waiting for completion? */
	if (ci && (ci->flags & CF_WAITAUTH))
	{
	    if (ci->key && ci->key == atol (param2))
	    {
		notice (s_ChanServ, source, CS_REGISTERED, ci->name, ni->nick);
		notice (s_ChanServ, source, RPL_PASSWORD_SET, ci->pass);

		if (strlen (csregister_url))
		{
		    notice (s_ChanServ, source, CS_REGISTER_INFO);
		    notice (s_ChanServ, source, csregister_url);
		}

		/* Set the default topic if there is one, and set the channel +r */
		if (c && !strlen (c->topic) && strlen (def_topic))
		    set_topic (ci, def_topic, s_ChanServ);

		if (c)
		    check_modes (ci->name);

		/* Turn the key off */
		ci->key = 0;

		/* No longer waiting */
		ci->flags &= ~CF_WAITAUTH;

		return;
	    }
	}
    }

    /* Check that we have all needed parameters and that they're valid. */
    if ((!csregistertype || csregistertype == 4) && (!param1 || !param2))
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "REGISTER Channel Password");
	errmoreinfo (s_ChanServ, source, "REGISTER");
        return;
    }

    /* REGISTER #Channel */
    if (csregistertype == 1 && !param1)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "REGISTER Channel");
	errmoreinfo (s_ChanServ, source, "REGISTER");
	return;
    }

    /* REGISTER Channel Pass Pass */
    if (csregistertype == 2 || csregistertype == 3)
    {
	if (!param1 || !param2 || !param3)
	{
	    notice (s_ChanServ, source, RPL_SYNTAX, "REGISTER Channel Password Password");
	    errmoreinfo (s_ChanServ, source, "REGISTER");
	    return;
	}

	if (strcmp (param2, param3))
	{
	    notice (s_ChanServ, source, ERR_PASS_DONT_MATCH);
	    return;
	}
    }

    ci = cs_findchan (param1);

    if (ci)
    {
	if (ci->flags & CF_FROZEN)
	    notice (s_ChanServ, source, CS_CANNOT_REGISTER, param1);
	else
	    notice (s_ChanServ, source, CS_ALREADY_REGISTERED, param1);

	return;
    }

    /* We only register #Channels.. */
    if (!*param1 == '#')
    {
	notice (s_ChanServ, source, CS_NOT_VALID, param1);
	return;
    }

    if (!is_opped (source, param1))
    {
	notice (s_ChanServ, source, CS_BE_OPPED, param1);
	return;
    }

    /* Password sanity checks */
    if (csregistertype != 1)
    {
	if (strlen (param2) < 5)
	    reason = 1;
	if (strlen (param2) > PASSWDLEN)
	    reason = 2;
	if (!stricmp (source, param2))
	    reason = 3;
	if (!stricmp ("password", param2))
	    reason = 4;

	if (reason)
	{
	    notice (s_ChanServ, source, ERR_BETTER_PASS);

	    if (reason == 1)
		notice (s_ChanServ, source, ERR_PASS_REJECTED_SHORT);
	    if (reason == 2)
		notice (s_ChanServ, source, ERR_PASS_REJECTED_LONG);
	    if (reason == 3)
		notice (s_ChanServ, source, ERR_PASS_REJECTED_NICK);
	    if (reason == 4)
		notice (s_ChanServ, source, ERR_PASS_REJECTED_WEAK);

	    errmoreinfo (s_ChanServ, source, "REGISTER");
	    return;
	}
    }

    /* No register floods! */
    if ((chan_delay && (now - u->lastcreg < chan_delay)) && !is_oper (source))
    {
	notice (s_ChanServ, source, CS_WAIT_REG,
	    duration (u->lastcreg - now + chan_delay, 2));

	return;
    }

    /* Make sure they have an email if the registertype calls for it. */
    if (csregistertype == 1 || csregistertype == 3)
	if (!ni->email)
	{
	    notice (s_ChanServ, source, CS_NEED_EMAIL, s_NickServ);
	    return;
	}

    /* Do they already have enough channels? */
    if (ni->chancnt >= chanlimit)
    {
	notice (s_ChanServ, source, CS_CHAN_LIMIT, ni->chancnt, chanlimit);
	notice (s_ChanServ, source, CS_DROP_SOME);
	notice (s_ChanServ, source, CS_LIST_OWNED, haveserv_on == TRUE ? "" : "MSG ",
	    s_ChanServ, haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
	    haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");
	return;
    }

    /* Register the Channel. */
    ci = scalloc (sizeof (ChanInfo), 1);
    strscpy (ci->name, param1, CHANNELLEN);
    strscpy (ci->founder, ni->nick, NICKLEN);

    if (csregistertype != 1)
	strscpy (ci->pass, param2, PASSWDLEN);

    /* We'll make a pw */
    if (csregistertype == 1 || csregistertype == 3)
	ci->key = makekey ();

    /* They don't have a password. We'll use the key as their pw. */
    if (csregistertype == 1)
	strscpy (ci->pass, itoa (ci->key), PASSWDLEN);

    ci->registered = ci->lastused = now;

    if (defchanflags)
	ci->flags = defchanflags;

    /* Set the default modelock */
    ci->mlock_on |= defmlock_on;

    if (defmlock_off)
	ci->mlock_off |= defmlock_off;

    u->lastcreg = now;

    /* Add this channel to the founder's channel list */
    ni->chancnt++;
    ni->chans = srealloc (ni->chans, sizeof (char *) *ni->chancnt);
    ni->chans[ni->chancnt - 1] = sstrdup (ci->name);

    list = &cs_chanlist[CSHASH(ci->name)];
    ci->next = *list;
    if (*list)
        (*list)->prev = ci;
    *list = ci;

    /* Add them as founder */
    for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	if (ca->level)
	    break;

    if (i == ci->accesscnt)
    {
	ci->accesscnt++;
	ci->access = srealloc (ci->access, sizeof (ChanAccess) * ci->accesscnt);
	ca = &ci->access[ci->accesscnt - 1];
    }
 
    ca->mask = sstrdup (ni->nick);
    ca->level = 5;

    /* Make note of the occurance. */
    log ("CS:REGISTER: %s %s!%s@%s", param1, source, u->user, u->host);
    snoop (s_ChanServ, "CS:REGISTER: %s %s!%s@%s", param1, source, u->user,
	u->host);

    /* If this was an E-Mail auth register, we'll tell them to check their
       email, email the instructions, and bail. We don't need to go beyond
       this point.
     */
    if (csregistertype == 1 || csregistertype == 3)
    {
	sendemail (ci->name, itoa (ci->key), 3);

	notice (s_ChanServ, source, CS_REGISTERED_TEMP, ci->name);
	notice (s_ChanServ, source, RPL_EMAIL_INSTRUCTIONS);
	notice (s_ChanServ, source, CS_TEMP_EXPIRE, duration (tempexpire, 2));

	/* Set a flag so we know we're waiting for them */
	ci->flags |= CF_WAITAUTH;

	return;
    }

    /* List the channel as awaiting verification, then tell them they
       have to wait for it.
     */
    if (csregistertype == 4)
    {
	Verify *vf, **vlist;

	ci->flags |= CF_WAITING;

	vf = scalloc (sizeof (Verify), 1);
	strscpy (vf->name, ci->name, CHANNELLEN);
	vf->since = now;

	vlist = &verifylist[VHASH(vf->name)];
	vf->next = *vlist;
	if (*vlist)
	    (*vlist)->prev = vf;
	*vlist = vf;

	verify_notice (vf);

	notice (s_ChanServ, source, CS_REGISTERED_TEMP, ci->name);
	notice (s_ChanServ, source, CS_WAIT_AUTH);
	notice (s_ChanServ, source, CS_WAIT_AWHILE);

	return;
    }

    /* Tell them it worked */
    notice (s_ChanServ, source, CS_REGISTERED, ci->name, ni->nick);
    notice (s_ChanServ, source, RPL_PASSWORD_SET, ci->pass);

    if (strlen (csregister_url))
    {
	notice (s_ChanServ, source, CS_REGISTER_INFO);
	notice (s_ChanServ, source, csregister_url);
    }

    c = findchan (param1);

    /* Set the default topic if there is one, and set the channel +r */
    if (c && !strlen (c->topic) && strlen (def_topic))
	set_topic (ci, def_topic, s_ChanServ);

    check_modes (ci->name);
}

/* Identify for founder access on a channel. */
static void do_identify (const char *source)
{
    ChanInfo *ci;
    User *u = finduser (source);
    char *chan = strtok (NULL, " ");
    char *pass = strtok (NULL, " ");

    if (!u)
	return;

    if (!chan || !pass)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "IDENTIFY Channel Password");
	errmoreinfo (s_ChanServ, source, "IDENTIFY");
	return;
    }

    ci = cs_findchan (chan);

    if (!ci)
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_CHAN_FROZEN, ci->name);
	return;
    }

    /* Check the passwords. */
    if (!strcmp (pass, ci->pass))
    {
	u->idchancnt++;
        u->idchans = srealloc (u->idchans, sizeof (char *) *u->idchancnt);
        u->idchans[u->idchancnt - 1] = sstrdup (ci->name);

	notice (s_ChanServ, source, RPL_PASS_ACCEPTED, ci->name);

	ci->lastused = time (NULL);

#ifdef UNREAL3
	send_cmd (s_ChanServ, "MODE %s +q %s", ci->name, source);
#endif
    }
    else
    {
	/* Passwords don't match, tell them and update their passfail count */
	u->passfail++;

	notice (s_ChanServ, source, RPL_WRONG_PASS, ci->name);

	/* If they've exceeded the limit, kill them. */
	if (u->passfail >= passfails)
	{
	    snoop (s_ChanServ, "CS:IDENTIFY:BP: %s %s!%s@%s (Killed)",
		ci->name, source, u->user, u->host);
	    log ("CS:IDENTIFY:BP: %s %s!%s@%s (Killed)",
		ci->name, source, u->user, u->host);
	    kill_user (s_ChanServ, source, RPL_KILLED_PASSFAIL);
	    return;
	}

	if (u->passfail >= (passfails - 1))
	    notice (s_ChanServ, source, RPL_LAST_TRY);
	snoop (s_ChanServ, "CS:IDENTIFY:BP: %s %s!%s@%s (%d/%d)",
	    ci->name, source, u->user, u->host, u->passfail, passfails);
	log ("CS:IDENTIFY:BP: %s %s!%s@%s (%d/%d)",
	    ci->name, source, u->user, u->host, u->passfail, passfails);

	return;
    }
}

/* Drop a registered channel */
static void do_drop (const char *source)
{
    ChanInfo *ci;
    NickInfo *ni;
    User *u = finduser (source);
    char *chan = strtok (NULL, " ");
    char *pass = strtok (NULL, " ");
    char **chans;
    int i, id = 0;

    if (!u)
	return;

    ni = get_nick (u);

    if (!ni)
    {
	notice (s_ChanServ, source, ERR_NEED_IDENT, "DROP");
	return;
    }

    if (!chan || !pass)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "DROP Channel Password");
	errmoreinfo (s_ChanServ, source, "DROP");
	return;
    }

    ci = cs_findchan (chan);

    if (!ci)
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_CHAN_FROZEN, ci->name);
	return;
    }

    /* Make sure they've identified for it. */
    if (u->idchancnt)
	for (chans = u->idchans, i = 0; i < u->idchancnt; chans++, i++)
	    if (!stricmp (*chans, ci->name))
		id = 1;

    if (!id)
    {
	notice (s_ChanServ, source, CS_IDENTIFY, ci->name, "\2DROP\2 it");
	return;
    }

    /* Check the passwords. */
    if (!strcmp (pass, ci->pass))
    {
	log ("CS:DROP: %s %s!%s@%s", ci->name, source, u->user, u->host);
	snoop (s_ChanServ, "CS:DROP: %s %s!%s@%s", ci->name, source, u->user, u->host);
	notice (s_ChanServ, source, CS_DROPPED, ci->name);

	delchan (ci);
    }
    else
    {
	/* Passwords don't match, tell them and update their passfail count */
	u->passfail++;

        notice (s_ChanServ, source, RPL_WRONG_PASS, ci->name);

	/* If they've exceeded the limit, kill them. */
	if (u->passfail >= passfails)
	{
	    snoop (s_ChanServ, "CS:DROP:BP: %s %s!%s@%s (Killed)",
		ci->name, source, u->user, u->host);
	    log ("CS:DROP:BP: %s %s!%s@%s (Killed)",
		ci->name, source, u->user, u->host);
	    kill_user (s_ChanServ, source, RPL_KILLED_PASSFAIL);
	    return;
	}

	if (u->passfail >= (passfails - 1))
	    notice (s_ChanServ, source, RPL_LAST_TRY);
	snoop (s_ChanServ, "CS:DROP:BP: %s %s!%s@%s (%d/%d)",
	    ci->name, source, u->user, u->host, u->passfail, passfails);
	log ("CS:DROP:BP: %s %s!%s@%s (%d/%d)",
	    ci->name, source, u->user, u->host, u->passfail, passfails);

	return;
    }
}

/* Show information on a registered channel */
static void do_info (const char *source)
{
    ChanInfo *ci;
    NickInfo *ni, *tni, *hni;
    char *chan = strtok (NULL, " ");
    char buf[BUFSIZE];

    if (!chan)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "INFO Channel");
	errmoreinfo (s_ChanServ, source, "INFO");
	return;
    }

    if (!(ci = cs_findchan (chan)))
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    notice (s_ChanServ, source, CS_INFO_START, ci->name);

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_INFO_FROZEN);
	notice (s_ChanServ, source, CS_INFO_FROZEN_BY, ci->topicsetter);
	notice (s_ChanServ, source, CS_INFO_FROZEN_WHY, ci->topic);
	notice (s_ChanServ, source, CS_INFO_FROZEN_TIME, get_time (ci->topictime, 4));
    }

    if (ci->flags & (CF_WAITING | CF_WAITAUTH))
	notice (s_ChanServ, source, CS_INFO_WAITING);

    ni = findnick (ci->founder);

    if (ni->host)
	hni = findnick (ni->host);
    else
	hni = ni;

    notice (s_ChanServ, source, CS_INFO_FOUNDER, ci->founder,
	hni ? " (" : "", hni ? hni->usermask : "", hni ? ")" : "");

    if (ci->successor)
    {
	tni = findnick (ci->successor);

	notice (s_ChanServ, source, CS_INFO_SUCCESSOR, ci->successor,
	    tni ? " (" : "", tni ? tni->usermask : "", tni ? ")" : "");
    }

    notice (s_ChanServ, source, CS_INFO_REGISTERED,
	get_time (ci->registered, 4), time_ago (ci->registered));

    if (!findchan (ci->name))
	notice (s_ChanServ, source, CS_INFO_LAST_USED,
	    get_time (ci->lastused, 4), time_ago (ci->lastused));

    if (ci->mlock_on || ci->mlock_off)
    {
	char params[BUFSIZE];

	*buf = 0;
	*params = 0;

	if (ci->mlock_on)
	{
	    strcat (buf, "+");
	    strcat (buf, flags_to_string (ci->mlock_on));

	    /* Add these in manually */
	    if (ci->mlock_limit)
	    {
		strcat (buf, "l");
		strcat (params, " ");
		strcat (params, itoa (ci->mlock_limit));
	    }

	    if (ci->mlock_key)
		strcat (buf, "k");

#ifdef UNREAL3
	    if (ci->mlock_flood)
	    {
		strcat (buf, "f");
		strcat (params, " ");
		strcat (params, ci->mlock_flood);
	    }

	    if (ci->mlock_link)
	    {
		strcat (buf, "L");
		strcat (params, " ");
		strcat (params, ci->mlock_link);
	    }
#endif
	}

	if (ci->mlock_off)
	{
	    strcat (buf, "-");
	    strcat (buf, flags_to_string (ci->mlock_off));
	}

	if (*buf)
	    notice (s_ChanServ, source, CS_INFO_MLOCK, buf, params ? params : "");
    }

    if (ci->url)
	notice (s_ChanServ, source, CS_INFO_URL, ci->url);

    if (ci->topiclock)
	notice (s_ChanServ, source, CS_INFO_TOPICLOCK,
	    ci->topiclock == 1 ? "VOP and above" :
#ifdef UNREAL3
	    ci->topiclock == 2 ? "HOP and above" :
#endif
	    ci->topiclock == 3 ? "AOP and above" :
	    ci->topiclock == 4 ? "SOP and above" :
	    ci->topiclock == 5 ? "Founder" : "Unknown!");

    *buf = 0;

    if (ci->flags & CF_SECURE)
	strcat (buf, FLAG_SECURE);

    if (ci->flags & CF_RESTRICTED)
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_RESTRICTED);
    }

    if (ci->flags & CF_VERBOSE)
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_VERBOSE);
    }

    if (ci->flags & CF_VOPALL)
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_VOPALL);
    }

    if (ci->flags & CF_HELD)
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_HELD);
    }

    if (ci->flags & CF_MARKED)
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_MARKED);
    }

    if (*buf)
	notice (s_ChanServ, source, CS_INFO_FLAGS, buf);
    else
	notice (s_ChanServ, source, CS_INFO_FLAGS, "None");

    notice (s_ChanServ, source, CS_INFO_END);
}

/* Handle an AUTH command */
static void do_auth (const char *source)
{
    Verify *vf;
    User *u;
    ChanInfo *ci;
    Channel *c;
    char *param = strtok (NULL, " ");
    char *chan = strtok (NULL, " ");
    char *reason = strtok (NULL, "");
    int i, j = 0;

    if (!param)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "AUTH VERIFY|DENY|LIST [Channel] [Reason]");
	errmoreinfo (s_ChanServ, source, "AUTH");
	return;
    }

    if (!stricmp (param, "VERIFY") || !stricmp (param, "DENY"))
    {
	if (!chan)
	{
	    notice (s_ChanServ, source, RPL_SYNTAX, "AUTH VERIFY|DENY Channel [Reason]");
	    errmoreinfo (s_ChanServ, source, "AUTH");
	    return;
	}

	vf = findverify (chan);

	if (!vf)
	{
	    notice (s_ChanServ, source, CS_VERIFY_NOT_FOUND, chan);
	    return;
	}

	ci = cs_findchan (chan);
	c = findchan (chan);
	u = finduser (ci->founder);

	if (!ci)
	    return;

	u = finduser (ci->founder);

	/* If they're online, we'll tell them now. If not, we'll memo them. */
	if (u)
	{
	    if (!stricmp (param, "VERIFY"))
		notice (s_ChanServ, u->nick, CS_VERIFIED, ci->name);
	    else
		notice (s_ChanServ, u->nick, CS_DENIED, ci->name, reason ? " Reason: " : "",
		    reason ? reason : "");
	}
	else
	{
	    NickInfo *ni = findnick (ci->founder);
	    MemoInfo *mi;
	    Memo *m;
	    char buf[256];

	    mi = &ni->memos;
	    mi->memocnt++;
	    mi->memos = srealloc (mi->memos, sizeof (Memo) * mi->memocnt);
	    m = &mi->memos[mi->memocnt - 1];

	    strscpy (m->sender, s_ChanServ, NICKLEN);

	    if (mi->memocnt > 1)
	    {
		m->number = m[-1].number + 1;

		if (m->number < 1)
		    for (i = 0; i < mi->memocnt; i++)
		mi->memos[i].number = i + 1;
	    }
	    else
		m->number = 1;

	    if (!stricmp (param, "VERIFY"))
		snprintf (buf, sizeof (buf), CS_VERIFIED, ci->name);
	    else
		snprintf (buf, sizeof (buf), CS_DENIED, ci->name, reason ? " Reason: " : "",
		    reason ? reason : "");

	    m->time = time (NULL);
	    m->text = sstrdup (buf);
	    m->flags = MF_UNREAD;
	}

	/* Now tell the CSOps */
	if (!stricmp (param, "VERIFY"))
	{
	    noticecsops (s_ChanServ, CS_CSOP_VERIFIED, ci->name, source, time_ago (vf->since));
	    log ("CS:AUTH:VERIFY: %s %s (Waited for %s)", source, ci->name, time_ago (vf->since));
	    snoop (s_ChanServ, "CS:AUTH:VERIFY: %s %s (Waited for %s)", source, ci->name, time_ago (vf->since));
	}
	else
	{
	    noticecsops (s_ChanServ, CS_CSOP_DENIED, ci->name, source, time_ago (vf->since));
	    log ("CS:AUTH:DENY: %s %s (Waited for %s)", source, ci->name, time_ago (vf->since));
	    snoop (s_ChanServ, "CS:AUTH:DENY: %s %s (Waited for %s)", source, ci->name, time_ago (vf->since));
	}

	/* Delete it from the verify list */
	if (vf->prev)
	    vf->prev->next = vf->next;
	else
	    verifylist[VHASH(vf->name)] = vf->next;

	if (vf->next)
	    vf->next->prev = vf->prev;

	free (vf);

	/* If it was denied, drop the channel */
	if (!stricmp (param, "DENY"))
	    delchan (ci);

	/* Turn the waiting flag off if it was verified */
	if (!stricmp (param, "VERIFY"))
	{
	    ci->flags &= ~CF_WAITING;

	    /* Set the default topic if there isn't one */
	    if (c && !strlen (c->topic) && strlen (def_topic))
		set_topic (ci, def_topic, s_ChanServ);

	    if (c)
		check_modes (ci->name);
	}

	return;
    }

    if (!stricmp (param, "LIST"))
    {
	notice (s_ChanServ, source, CS_VERIFY_LIST_START);
	notice (s_ChanServ, source, CS_VERIFY_LIST);

	for (i = 0; i < 1024; i++)
	    for (vf = verifylist[i]; vf; vf = vf->next)
	    {
		ci = cs_findchan (vf->name);

		if (!ci)
		{
		    /* It's not there now .. Delete it. */
		    if (vf->prev)
			vf->prev->next = vf->next;
		    else
			verifylist[VHASH(vf->name)] = vf->next;

		    if (vf->next)
			vf->next->prev = vf->prev;

		    free (vf);
		}
		else
		{
		    /* This should NOT be null. */
		    NickInfo *ni = findnick (ci->founder);

		    notice (s_ChanServ, source, "%-24s %-24s %-7d %s", vf->name, ci->founder,
			ni->chancnt, time_ago (vf->since));
		    j++;
		}
	    }

	notice (s_ChanServ, source, CS_VERIFY_LIST_END, j);

	return;
    }

    notice (s_ChanServ, source, ERR_UNKNOWN_OPTION, param,
	haveserv_on == TRUE ? "" : "MSG ", s_ChanServ,
	haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
	haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");
}

/* Set various channel options. */
static void do_set (const char *source)
{
    ChanInfo *ci;
    NickInfo *ni;
    User *u = finduser (source);
    char *chan = strtok (NULL, " ");
    char *option = strtok (NULL, " ");
    char *param;
    char **chans;
    int i, id = 0;

    if (!u)
	return;

    /* Get their nickname */
    ni = get_nick (u);

    if (!ni)
    {
	notice (s_ChanServ, source, ERR_NEED_IDENT, "SET");
	return;
    }

    /* Special handling for this for some options.. */
    if (option && !stricmp (option, "MLOCK"))
	param = strtok (NULL, " ");
    else
	param = strtok (NULL, "");

    /* Check for everything we need */
    if (!chan || !option || !param)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "SET Channel Option Parameters");
	errmoreinfo (s_ChanServ, source, "SET");
	return;
    }

    /* Find the channel and make sure it's usable */
    ci = cs_findchan (chan);

    if (!ci)
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_CHAN_FROZEN, ci->name);
	return;
    }

    if (!check_cs_auth (s_ChanServ, source, ci))
	return;

    /* Only someone with founder access can modify settings. Check if
       they've identified for this channel.
     */
    if (u->idchancnt)
	for (chans = u->idchans, i = 0; i < u->idchancnt; chans++, i++)
	    if (!stricmp (*chans, ci->name))
		id = 1;

    /* Also check for identification to the founder's nick, which we
       accept as well.
     */
    if (!stricmp (ni->nick, ci->founder))
	id = 1;

    if (!id)
    {
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "the channel founder");
	return;
    }

    /* Set the MLock? */
    if (!stricmp (option, "MLOCK"))
    {
	char *s, modebuf[32], *end, c;
	int add = -1;
	int32 newlock_on = 0, newlock_off = 0, newlock_limit = 0;
	char *newlock_key = NULL;
#ifdef UNREAL3
	char *newlock_flood = NULL, *newlock_link = NULL, **chans;
#endif

	while (*param)
	{
	    if (*param != '+' && *param != '-' && add < 0)
	    {
		param++;
		continue;
	    }

	    switch ((c = *param++))
	    {
		case '+':
		    add = 1;
		    break;

		case '-':
		    add = 0;
		    break;
		case 'k':
		    if (add)
		    {
			if (!(s = strtok (NULL, " ")))
			{
			    notice (s_ChanServ, source, CS_MLOCK_NEED_PARAM, "key",
				"+k");
			    return;
			}

			if (newlock_key)
			    free (newlock_key);

			newlock_key = sstrdup (s);
			newlock_off &= ~CMODE_k;
		    }
		    else
		    {
			if (newlock_key)
			{
			    free (newlock_key);
			    newlock_key = NULL;
			}

			newlock_off |= CMODE_k;
		    }

		    break;
		case 'l':
		    if (add)
		    {
			if (!(s = strtok (NULL, " ")))
			{
			    notice (s_ChanServ, source, CS_MLOCK_NEED_PARAM, "limit", "+l");
			    return;
			}

			if (atol (s) <= 0)
			{
			    notice (s_ChanServ, source, CS_MLOCK_LIMIT_POSITIVE);
			    return;
			}

			newlock_limit = atol (s);
			newlock_off &= ~CMODE_l;
		    }
		    else
		    {
			newlock_limit = 0;
			newlock_off |= CMODE_l;
		    }

		    break;

#ifdef UNREAL3
		case 'f':
		    if (add)
		    {
			if (!(s = strtok (NULL, " ")))
			{
			    notice (s_ChanServ, source, CS_MLOCK_NEED_PARAM, "flood trigger", "+f");
			    return;
			}

			if (newlock_flood)
			    free (newlock_flood);

			newlock_flood = sstrdup (s);
			newlock_off &= ~CMODE_f;
		    }
		    else
		    {
			if (newlock_flood)
			{
			    free (newlock_flood);
			    newlock_flood = NULL;
			}

			newlock_off |= CMODE_f;
		    }

		    break;

		case 'L':
			/* Only the channel founder can set +L. */
			if (u->idchancnt)
			    for (chans = u->idchans, i = 0; i < u->idchancnt; chans++, i++)
				if (stricmp (*chans, ci->name))
				{
				    notice (s_ChanServ, source, CS_IDENTIFY, ci->name, "MLock +L");
				    return;
				}

		    if (add)
		    {
			if (!(s = strtok (NULL, " ")))
			{
			    notice (s_ChanServ, source, CS_MLOCK_NEED_PARAM, "channel", "+L");
			    return;
			}

			if (newlock_link)
			    free (newlock_link);

			newlock_link = sstrdup (s);
			newlock_off &= ~CMODE_L;
		    }
		    else
		    {
			if (newlock_link)
			{
			    free (newlock_link);
			    newlock_link = NULL;
			}

			newlock_off |= CMODE_L;
		    }

		    break;
#endif
		default:
		{
		    int32 flag = mode_to_flag (c);

#ifdef UNREAL3
		    if ((flag & (CMODE_A | CMODE_H)) && (!is_csop (source) ||
			(u->mode & (UMODE_A | UMODE_N | UMODE_T))))
			continue;
#endif

#if defined (BAHAMUT) || defined (PROMETHEUS) || defined (UNREAL3)
		    if ((flag & CMODE_O) && !is_oper (source))
			continue;
#endif
		    if (flag)
		    {
			if (add)
			    newlock_on |= flag, newlock_off &= ~flag;
			else
			    newlock_off |= flag, newlock_on &= ~flag;
		    }
		}
	    }
	}

	/* Registered channels are always +r. Add it to mlock on and remove it
	   from mlock off.
	 */
#ifndef IRCUP9
	newlock_on |= CMODE_r, newlock_off &= ~CMODE_r;
#endif

	/* Save it to ChanInfo. */
	ci->mlock_on = newlock_on;
	ci->mlock_off = newlock_off;

	ci->mlock_limit = newlock_limit;

	if (ci->mlock_key)
	    free (ci->mlock_key);

	ci->mlock_key = newlock_key;

#ifdef UNREAL3
	if (ci->mlock_flood)
	    free (ci->mlock_flood);

	ci->mlock_flood = newlock_flood;

	if (ci->mlock_link)
	    free (ci->mlock_link);

	ci->mlock_link = newlock_link;
#endif

	/* Tell the user about it. */
	end = modebuf;
	*end = 0;

#ifdef UNREAL3
	if (ci->mlock_on || ci->mlock_key || ci->mlock_limit || ci->mlock_flood || ci->mlock_link)
	    end += snprintf (end, sizeof (modebuf) - (end - modebuf), "+%s%s%s%s%s",
		flags_to_string (ci->mlock_on),
		ci->mlock_key ? "k" : "",
		ci->mlock_limit ? "l" : "",
		ci->mlock_flood ? "f" : "",
		ci->mlock_link ? "L" : "");
#else
	if (ci->mlock_on || ci->mlock_key || ci->mlock_limit)
	    end += snprintf (end, sizeof (modebuf) - (end - modebuf), "+%s%s%s",
		flags_to_string (ci->mlock_on),
		ci->mlock_key ? "k" : "",
		ci->mlock_limit ? "l" : "");
#endif

	if (ci->mlock_off)
	    end += snprintf (end, sizeof (modebuf) - (end - modebuf), "-%s",
		flags_to_string (ci->mlock_off));

	if (*modebuf)
	    notice (s_ChanServ, source, RPL_ADDED, "Your channel's \2MLOCK\2", modebuf);
	else
	    notice (s_ChanServ, source, RPL_REMOVED, "Your channel's \2MLOCK\2");

	/* Finally, check the modes and make any nessecary changes. */
	check_modes (ci->name);

	return;
    }

    /* Set the Successor? */
    if (!stricmp (option, "SUCCESSOR"))
    {
	NickInfo *ni = findnick (param);

	if (!stricmp (param, "NONE") || !stricmp (param, "OFF"))
	{
	    /* Remove this channel from old successor's list, and
	       free the successor.
	     */
	    if (ci->successor)
	    {
		NickInfo *sni = findnick (ci->successor);
		int i;

		/* Take this channel out of their successor channel list. */
		for (i = 0; i < sni->successorcnt; i++)
		{
		    if (!stricmp (sni->successor[i], ci->name))
		    {
			free (sni->successor[i]);
			sni->successorcnt--;
			break;
		    }
		}

		if (!sni->successorcnt)
		    free (sni->successor);

		free (ci->successor);
	    }

	    notice (s_ChanServ, source, RPL_REMOVED, "Your channel's \2SUCCESSOR\2");
	    return;
	}

	if (!ni)
	{
	    notice (s_ChanServ, source, NS_NOT_REGISTERED, param);
	    return;
	}

	/* Nuh-uh. */
	if (!stricmp (ni->nick, ci->founder))
	{
	    notice (s_ChanServ, source, CS_SUCCESSOR_SELF);
	    return;
	}

	if (ni->chancnt >= chanlimit)
	{
	    notice (s_ChanServ, source, CS_CHAN_LIMIT_OTHER, ni->nick);
	    return;
	}

	/* Check for the NoSuccessor flag */
	if (ni->flags & NF_NOSUCCESSOR)
	{
	    notice (s_ChanServ, source, CS_NO_SUCCESSOR, ni->nick);
	    return;
	}

	/* If theres already a successor, we have to do a bit of cleanup. */
	if (ci->successor)
	{
	    NickInfo *sni = findnick (ci->successor);
	    int i;

	    /* Take this channel out of their successor channel list. */
	    for (i = 0; i < sni->successorcnt; i++)
	    {
		if (!stricmp (sni->successor[i], ci->name))
		{
		    free (sni->successor[i]);
		    sni->successorcnt--;
		    break;
		}
	    }

	    if (!sni->successorcnt)
		free (sni->successor);

	    free (ci->successor);
	}

	ci->successor = sstrdup (ni->nick);

	/* Add this channel to the users successor channel list */
	ni->successorcnt++;
	ni->successor = srealloc (ni->successor, sizeof (char *) *ni->successorcnt);
	ni->successor[ni->successorcnt - 1] = sstrdup (ci->name);

	notice (s_ChanServ, source, RPL_ADDED, "Your channel's \2SUCCESSOR\2", ni->nick);

	return;
    }

    /* Change password? */
    if (!stricmp (option, "PASS") || !stricmp (option, "PASSWD") ||
	!stricmp (option, "PASSWORD"))
    {
	int reason = 0;
	char *pass1 = strtok (param, " ");
	char *pass2 = strtok (NULL, " ");

	if (!pass2)
	{
	    notice (s_ChanServ, source, RPL_SYNTAX,
		"SET Channel PASS Password Password");
	    errmoreinfo (s_ChanServ, source, "SET PASS");
	    return;
	}

	/* Password sanity checks */
	if (strlen (pass1) < 5)
	    reason = 1;
	if (strlen (pass1) > PASSWDLEN)
	    reason = 2;
	if (!stricmp (source, pass1))
	    reason = 3;
	if (!stricmp ("password", pass1))
	    reason = 4;
	if (strcmp (pass1, pass2))
	    reason = 5;
	if (!stricmp (ci->name, pass1))
	    reason = 6;
	if (!stricmp (ci->founder, pass1))
	    reason = 7;

	if (reason)
	{
	    notice (s_ChanServ, source, ERR_BETTER_PASS);

	    if (reason == 1)
		notice (s_ChanServ, source, ERR_PASS_REJECTED_SHORT);
	    if (reason == 2)
		notice (s_ChanServ, source, ERR_PASS_REJECTED_LONG);
	    if (reason == 3)
		notice (s_ChanServ, source, ERR_PASS_REJECTED_NICK);
	    if (reason == 4)
		notice (s_ChanServ, source, ERR_PASS_REJECTED_WEAK);
	    if (reason == 5)
		notice (s_ChanServ, source, ERR_PASS_REJECTED_DIFF);
	    if (reason == 6)
		notice (s_ChanServ, source, ERR_PASS_REJECTED_CHAN);
	    if (reason == 7)
		notice (s_ChanServ, source, ERR_PASS_REJECTED_FNDR);

	    errmoreinfo (s_ChanServ, source, "SET PASS");
	    return;
	}

	/* Passed checks, set it */
	strscpy (ci->pass, pass1, PASSWDLEN);
	notice (s_ChanServ, source, RPL_PASSWORD_SET, ci->pass);

	/* Log this. This might be an attempt to steal a nick.. */
	log ("CS:SET:PASS: %s %s!%s@%s", ci->name, source, u->user, u->host);
	return;
    }

    /* Change URL? */
    if (!stricmp (option, "URL"))
    {
	if (!stricmp (param, "NONE") || !stricmp (param, "OFF"))
	{
	    if (ci->url)
		free (ci->url);
	    ci->url = NULL;
	    notice (s_ChanServ, source, RPL_REMOVED, "Your channel's \2URL\2");
	    return;
	}
	else
	{
	    if (strstr (param, "://"))
	    {
		if (ci->url)
		    free (ci->url);
		ci->url = sstrdup (param);
		notice (s_ChanServ, source, RPL_ADDED, "Your channel's \2URL\2", param);
	    }
	    else
		notice (s_ChanServ, source, ERR_INVALID, "URL", "");

	    return;
	}
    }

    /* Change GREET? */
    if (!stricmp (option, "GREET"))
    {
	if (!stricmp (param, "NONE") || !stricmp (param, "OFF"))
	{
	    if (ci->greet)
		free (ci->greet);
	    ci->greet = NULL;
	    notice (s_ChanServ, source, RPL_REMOVED, "Your channel's \2GREET\2 message");
	    return;
	}
	else
	{
	    if (strlen (param) <= 50)
	    {
		if (ci->greet)
		    free (ci->greet);
		ci->greet = sstrdup (param);
		notice (s_ChanServ, source, RPL_ADDED, "Your channel's \2GREET\2 message", param);
	    }
	    else
		notice (s_ChanServ, source, ERR_INVALID, "GREET message", " It is too long.");

	    return;
	}
    }

    /* Change TOPIC? */
    if (!stricmp (option, "TOPIC"))
    {
	Channel *c = findchan (ci->name);

	if (!c)
	    return;

	if (!stricmp (param, "NONE") || !stricmp (param, "OFF"))
	{
	    strscpy (ci->topic, "", TOPICLEN);
	    strscpy (ci->topicsetter, s_ChanServ, NICKLEN);
	    strscpy (c->topic, "", TOPICLEN);
	    strscpy (c->topicsetter, s_ChanServ, NICKLEN);
	    c->topictime = ci->topictime = time (NULL);
	    send_cmd (s_ChanServ, "TOPIC %s %s %lu :", ci->name,
		s_ChanServ, time (NULL));

	    notice (s_ChanServ, source, RPL_REMOVED, "Your channel's \2TOPIC\2");
	    return;
	}
	else
	{
	    /* This differs on some IRCds, but this is as low as it gets AFAIK. */
	    if (strlen (param) <= 307)
	    {
		strscpy (ci->topic, param, TOPICLEN);
		strscpy (ci->topicsetter, source, NICKLEN);
		strscpy (c->topic, param, TOPICLEN);
		strscpy (c->topicsetter, source, NICKLEN);
		c->topictime = ci->topictime = time (NULL);
		send_cmd (s_ChanServ, "TOPIC %s %s %lu :%s", ci->name,
		    source, time (NULL), param);

		notice (s_ChanServ, source, RPL_ADDED, "Your channel's \2TOPIC\2", param);
	    }
	    else
		notice (s_ChanServ, source, ERR_INVALID, "TOPIC", " It is too long.");

	    return;
	}
    }

    /* Change TOPICLOCK? */
    if (!stricmp (option, "TOPICLOCK"))
    {
	if (!stricmp (param, "NONE") || !stricmp (param, "OFF"))
	{
	    ci->topiclock = 0;

	    notice (s_ChanServ, source, RPL_REMOVED, "Your channel's \2TOPICLOCK\2");
	    return;
	}

	if (!stricmp (param, "VOP"))
	{
	    ci->topiclock = 1;

	    notice (s_ChanServ, source, RPL_ADDED, "Your channel's \2TOPICLOCK\2", "VOP and above");
	    return;
	}

#ifdef UNREAL3
	if (!stricmp (param, "HOP"))
	{
	    ci->topiclock = 2;

	    notice (s_ChanServ, source, RPL_ADDED, "Your channel's \2TOPICLOCK\2", "HOP and above");
	    return;
	}
#endif

	if (!stricmp (param, "AOP"))
	{
	    ci->topiclock = 3;

	    notice (s_ChanServ, source, RPL_ADDED, "Your channel's \2TOPICLOCK\2", "AOP and above");
	    return;
	}

	if (!stricmp (param, "SOP"))
	{
	    ci->topiclock = 4;

	    notice (s_ChanServ, source, RPL_ADDED, "Your channel's \2TOPICLOCK\2", "SOP and above");
	    return;
	}

	if (!stricmp (param, "FOUNDER"))
	{
	    ci->topiclock = 5;

	    notice (s_ChanServ, source, RPL_ADDED, "Your channel's \2TOPICLOCK\2", "Founder");
	    return;
	}

	/* Dunno what they want.. */
#ifdef UNREAL3
	notice (s_ChanServ, source, RPL_SYNTAX, "SET Channel TOPICLOCK VOP|HOP|AOP|SOP|Founder");
#else
	notice (s_ChanServ, source, RPL_SYNTAX, "SET Channel TOPICLOCK VOP|AOP|SOP|Founder");
#endif
	errmoreinfo (s_ChanServ, source, "SET");
	return;
    }

    /* Toggle VERBOSE? */
    if (!stricmp (option, "VERBOSE"))
    {
	if (!stricmp (param, "ON"))
	{
	    if (ci->flags & CF_VERBOSE)
	    {
		notice (s_ChanServ, source, CS_OPTION_ALREADY, "Verbose", "on");
		return;
	    }

	    ci->flags |= CF_VERBOSE;
	    notice (s_ChanServ, source, CS_OPTION_NOW, "Verbose", "on");
	    return;
	}

	if (!stricmp (param, "OFF"))
	{
	    if (!(ci->flags & CF_VERBOSE))
	    {
		notice (s_ChanServ, source, CS_OPTION_ALREADY, "Verbose", "off");
		return;
	    }

	    ci->flags &= ~CF_VERBOSE;
	    notice (s_ChanServ, source, CS_OPTION_NOW, "Verbose", "off");
	    return;
	}

	else
	{
	    notice (s_ChanServ, source, ERR_UNKNOWN_OPTION, strupper (param),
		haveserv_on == TRUE ? "" : "MSG ", s_ChanServ,
 		haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");
	    return;
	}
    }

    /* Toggle VOPALL? */
    if (!stricmp (option, "VOPALL"))
    {
	if (!stricmp (param, "ON"))
	{
	    if (ci->flags & CF_VOPALL)
	    {
		notice (s_ChanServ, source, CS_OPTION_ALREADY, "VOPAll", "on");
		return;
	    }

	    ci->flags |= CF_VOPALL;
	    notice (s_ChanServ, source, CS_OPTION_NOW, "VOPAll", "on");
	    return;
	}

	if (!stricmp (param, "OFF"))
	{
	    if (!(ci->flags & CF_VOPALL))
	    {
		notice (s_ChanServ, source, CS_OPTION_ALREADY, "VOPAll", "off");
		return;
	    }

	    ci->flags &= ~CF_VOPALL;
	    notice (s_ChanServ, source, CS_OPTION_NOW, "VOPAll", "off");
	    return;
	}

	else
	{
	    notice (s_ChanServ, source, ERR_UNKNOWN_OPTION, strupper (param),
		haveserv_on == TRUE ? "" : "MSG ", s_ChanServ,
 		haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");
	    return;
	}
    }

    /* Toggle SECURE? */
    if (!stricmp (option, "SECURE"))
    {
	if (!stricmp (param, "ON"))
	{
	    if (ci->flags & CF_SECURE)
	    {
		notice (s_ChanServ, source, CS_OPTION_ALREADY, "Secure", "on");
		return;
	    }

	    ci->flags |= CF_SECURE;
	    notice (s_ChanServ, source, CS_OPTION_NOW, "Secure", "on");
	    return;
	}

	if (!stricmp (param, "OFF"))
	{
	    if (!(ci->flags & CF_SECURE))
	    {
		notice (s_ChanServ, source, CS_OPTION_ALREADY, "Secure", "off");
		return;
	    }

	    ci->flags &= ~CF_SECURE;
	    notice (s_ChanServ, source, CS_OPTION_NOW, "Secure", "off");
	    return;
	}

	else
	{
	    notice (s_ChanServ, source, ERR_UNKNOWN_OPTION, strupper (param),
		haveserv_on == TRUE ? "" : "MSG ", s_ChanServ,
 		haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");
	    return;
	}
    }

    /* Toggle RESTRICTED? */
    if (!stricmp (option, "RESTRICTED"))
    {
	if (!stricmp (param, "ON"))
	{
	    if (ci->flags & CF_RESTRICTED)
	    {
		notice (s_ChanServ, source, CS_OPTION_ALREADY, "Restricted", "on");
		return;
	    }

	    ci->flags |= CF_RESTRICTED;
	    notice (s_ChanServ, source, CS_OPTION_NOW, "Restricted", "on");
	    return;
	}

	if (!stricmp (param, "OFF"))
	{
	    if (!(ci->flags & CF_RESTRICTED))
	    {
		notice (s_ChanServ, source, CS_OPTION_ALREADY, "Restricted", "off");
		return;
	    }

	    ci->flags &= ~CF_RESTRICTED;
	    notice (s_ChanServ, source, CS_OPTION_NOW, "Restricted", "off");
	    return;
	}

	else
	{
	    notice (s_ChanServ, source, ERR_UNKNOWN_OPTION, strupper (param),
		haveserv_on == TRUE ? "" : "MSG ", s_ChanServ,
 		haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");
	    return;
	}
    }

    /* Don't know what they want .. */
    notice (s_ChanServ, source, ERR_UNKNOWN_OPTION, strupper (option),
	haveserv_on == TRUE ? "" : "MSG ", s_ChanServ,
	haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
	haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");
}

/* Count various lists on a channel */
static void do_count (const char *source)
{
    ChanInfo *ci;
    ChanAccess *ca;
    AKick *ak;
    User *u = finduser (source);
    char *chan = strtok (NULL, " ");
    int i, ulev, vop = 0, aop = 0, sop = 0, akick = 0;
#ifdef UNREAL3
    int hop = 0;
#endif

    if (!u)
	return;

    if (!chan)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "COUNT Channel");
	errmoreinfo (s_ChanServ, source, "COUNT");
	return;
    }

    if (!(ci = cs_findchan (chan)))
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_CHAN_FROZEN, ci->name);
	return;
    }

    if (!check_cs_auth (s_ChanServ, source, ci))
	return;

    ulev = get_access (u, ci);

    if (ulev < 1)
    {
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "VOP");
	return;
    }

    /* Count VOPs */
    for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
    {
	if (!ca->level || !(ca->level == 1))
	    continue;

	vop++;
    }

#ifdef UNREAL3
    /* Count HOPs */
    for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
    {
	if (!ca->level || !(ca->level == 2))
	    continue;

	hop++;
    }
#endif

    /* Count AOPs */
    for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
    {
	if (!ca->level || !(ca->level == 3))
	    continue;

	aop++;
    }

    /* Count SOPs */
    for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
    {
	if (!ca->level || !(ca->level == 4))
	    continue;

	sop++;
    }

    /* Count AKicks */
    for (ak = ci->akick, i = 0; i < ci->akickcnt; ak++, i++)
    {
	if (!ak->level)
	    continue;

	akick++;
    }

    notice (s_ChanServ, source, CS_COUNT, vop,
#ifdef UNREAL3
	itoa (hop), " HOPs, ",
#else
	"", "",
#endif
	aop, sop, akick, ci->name);
}

/* Add, Remove, or List VOP entries. */
static void do_vop (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *option = strtok (NULL, " ");
    char *mask = strtok (NULL, " ");
    ChanInfo *ci;
    ChanAccess *ca;
    NickInfo *ni, *hni, *tni;
    User *u = finduser (source);
    int i, j = 0, ulev;

    if (!u)
	return;

    /* Get their nickname */
    ni = get_nick (u);

    if (!ni)
    {
	notice (s_ChanServ, source, ERR_NEED_IDENT, "VOP");
	return;
    }

    /* Check for everything we need */
    if (!chan || !option)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "VOP Channel ADD|DEL|LIST [Mask|Nickname]");
	errmoreinfo (s_ChanServ, source, "VOP");
	return;
    }

    /* Find the channel and make sure it's usable */
    ci = cs_findchan (chan);

    if (!ci)
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_CHAN_FROZEN, ci->name);
	return;
    }

    if (!check_cs_auth (s_ChanServ, source, ci))
	return;

    ulev = get_access (u, ci);

    /* ADD */
    if (!stricmp (option, "ADD"))
    {
	if (!mask)
	{
	    notice (s_ChanServ, source, RPL_SYNTAX, "VOP Channel ADD Nick");
	    errmoreinfo (s_ChanServ, source, "VOP");
	    return;
	}

	if (ulev < 4)
	{
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "SOP");
	    return;
	}

	tni = findnick (mask);

	if (!tni)
	{
	    notice (s_ChanServ, source, CS_ACCESS_REGISTERED, mask);
	    return;
	}

	/* Check for a host nick */
	if (tni->host)
	    hni = findnick (tni->host);
	else
	    hni = tni;

	if (!check_ns_auth (s_ChanServ, source, hni))
	    return;

	/* Check for NoOp */
	if (hni->flags & NF_NOOP)
	{
	    notice (s_ChanServ, source, CS_NO_OP, hni->nick);
	    return;
	}

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	{
	    if (!ca->level)
		continue;

	    if (!stricmp (ca->mask, hni->nick))
	    {
		/* Do they already have this level? */
		if (ca->level == 1)
		{
		    notice (s_ChanServ, source, CS_ACCESS_ALREADY, hni->nick, "VOP", ci->name);
		    return;
		}

		/* Do they have greater or equal access than this user? */
		if (ca->level >= ulev)
		{
		    notice (s_ChanServ, source, CS_ACCESS_DENIED, "Founder");
		    return;
		}

		notice (s_ChanServ, source, CS_ACCESS_CHANGED, hni->nick, ci->name,
		    ca->level == 2 ? "HOP" : ca->level == 3 ? "AOP" :
		    ca->level == 4 ? "SOP" : "Unknown", "VOP");

		if (ci->flags & CF_VERBOSE)
		    opnotice (s_ChanServ, ci->name, CS_VERBOSE_CHANGED, source, hni->nick,
			ca->level == 2 ? "HOP" : ca->level == 3 ? "AOP" : ca->level == 4 ? "SOP" :
			"Unknown", "VOP");

		ca->level = 1;

		return;
	    }
	}

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	    if (!ca->level)
		break;

	if (i == ci->accesscnt)
	{
	    ci->accesscnt++;
	    ci->access = srealloc (ci->access, sizeof (ChanAccess) * ci->accesscnt);
	    ca = &ci->access[ci->accesscnt - 1];
	}

	ca->mask = sstrdup (hni->nick);
	ca->level = 1;

	notice (s_ChanServ, source, CS_ACCESS_ADDED, hni->nick, ci->name, "VOP");

	if (ci->flags & CF_VERBOSE)
	    opnotice (s_ChanServ, ci->name, CS_VERBOSE_ADDED, source, hni->nick, "VOP");

	return;
    }

    /* DEL */
    if (!stricmp (option, "DEL"))
    {
	int success = 0;

	if (!mask)
	{
	    notice (s_ChanServ, source, RPL_SYNTAX, "VOP Channel DEL Nick");
	    errmoreinfo (s_ChanServ, source, "VOP");
	    return;
	}

	if (ulev < 4)
	{
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "SOP");
	    return;
	}

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	{
	    if (!ca->level && !(ca->level == 1))
		continue;

	    j++;

	    if (!stricmp (mask, ca->mask))
	    {
		success = 1;
		break;
	    }

	    if (atoi (mask) == j)
	    {
		success = 1;
		break;
	    }
	}

	if (!success)
	{
	    notice (s_ChanServ, source, CS_ACCESS_NOT_FOUND, mask, ci->name, "VOP");
	    return;
	}
	else
	{
	    notice (s_ChanServ, source, CS_ACCESS_REMOVED, ca->mask, ci->name, "VOP");

	    if (ci->flags & CF_VERBOSE)
		opnotice (s_ChanServ, ci->name, CS_VERBOSE_REMOVED, source, ca->mask, "VOP");

	    if (ca->mask)
		free (ca->mask);

	    ca->level = 0;
	}

	return;
    }

    /* LIST */
    if (!stricmp (option, "LIST"))
    {
	if (ulev < 1)
	{
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "VOP");
	    return;
	}

	notice (s_ChanServ, source, CS_ACCESS_LIST_START, "VOP", ci->name);
	notice (s_ChanServ, source, CS_ACCESS_LIST_HEADER);

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	{
	    if (!ca->level || !(ca->level == 1))
		continue;

	    j++;

	    tni = findnick (ca->mask);

	    notice (s_ChanServ, source, CS_ACCESS_LIST, j, ca->mask,
		tni->usermask ? "(" : "", tni->usermask ? tni->usermask : "",
		tni->usermask ? ")" : "");
	}

	notice (s_ChanServ, source, CS_ACCESS_LIST_END, j);
    }
}

#ifdef UNREAL3
/* Add, Remove, or List HOP entries. */
static void do_hop (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *option = strtok (NULL, " ");
    char *mask = strtok (NULL, " ");
    ChanInfo *ci;
    ChanAccess *ca;
    NickInfo *ni, *hni, *tni;
    User *u = finduser (source);
    int i, j = 0, ulev;

    if (!u)
	return;

    /* Get their nickname */
    ni = get_nick (u);

    if (!ni)
    {
	notice (s_ChanServ, source, ERR_NEED_IDENT, "HOP");
	return;
    }

    /* Check for everything we need */
    if (!chan || !option)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "HOP Channel ADD|DEL|LIST [Mask|Nickname]");
	errmoreinfo (s_ChanServ, source, "HOP");
	return;
    }

    /* Find the channel and make sure it's usable */
    ci = cs_findchan (chan);

    if (!ci)
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_CHAN_FROZEN, ci->name);
	return;
    }

    if (!check_cs_auth (s_ChanServ, source, ci))
	return;

    ulev = get_access (u, ci);

    /* ADD */
    if (!stricmp (option, "ADD"))
    {
	if (!mask)
	{
	    notice (s_ChanServ, source, RPL_SYNTAX, "HOP Channel ADD Nick");
	    errmoreinfo (s_ChanServ, source, "HOP");
	    return;
	}

	if (ulev < 4)
	{
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "SOP");
	    return;
	}

	tni = findnick (mask);

	if (!tni)
	{
	    notice (s_ChanServ, source, CS_ACCESS_REGISTERED, mask);
	    return;
	}

	/* Check for a host nick */
	if (tni->host)
	    hni = findnick (tni->host);
	else
	    hni = tni;

	if (!check_ns_auth (s_ChanServ, source, hni))
	    return;

	/* Check for NoOp */
	if (hni->flags & NF_NOOP)
	{
	    notice (s_ChanServ, source, CS_NO_OP, hni->nick);
	    return;
	}

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	{
	    if (!ca->level)
		continue;

	    if (!stricmp (ca->mask, hni->nick))
	    {
		/* Do they already have this level? */
		if (ca->level == 2)
		{
		    notice (s_ChanServ, source, CS_ACCESS_ALREADY, hni->nick, "HOP", ci->name);
		    return;
		}

		/* Do they have greater or equal access than this user? */
		if (ca->level >= ulev)
		{
		    notice (s_ChanServ, source, CS_ACCESS_DENIED, "Founder");
		    return;
		}

		notice (s_ChanServ, source, CS_ACCESS_CHANGED, hni->nick, ci->name,
		    ca->level == 1 ? "VOP" : ca->level == 3 ? "AOP" :
		    ca->level == 4 ? "SOP" : "Unknown", "HOP");

		if (ci->flags & CF_VERBOSE)
		    opnotice (s_ChanServ, ci->name, CS_VERBOSE_CHANGED, source, hni->nick,
			ca->level == 1 ? "VOP" : ca->level == 3 ? "AOP" : ca->level == 4 ? "SOP" :
			"Unknown", "HOP");

		ca->level = 2;

		return;
	    }
	}

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	    if (!ca->level)
		break;

	if (i == ci->accesscnt)
	{
	    ci->accesscnt++;
	    ci->access = srealloc (ci->access, sizeof (ChanAccess) * ci->accesscnt);
	    ca = &ci->access[ci->accesscnt - 1];
	}

	ca->mask = sstrdup (hni->nick);
	ca->level = 2;

	notice (s_ChanServ, source, CS_ACCESS_ADDED, hni->nick, ci->name, "HOP");

	if (ci->flags & CF_VERBOSE)
	    opnotice (s_ChanServ, ci->name, CS_VERBOSE_ADDED, source, hni->nick, "HOP");

	return;
    }

    /* DEL */
    if (!stricmp (option, "DEL"))
    {
	int success = 0;

	if (!mask)
	{
	    notice (s_ChanServ, source, RPL_SYNTAX, "HOP Channel DEL Nick");
	    errmoreinfo (s_ChanServ, source, "HOP");
	    return;
	}

	if (ulev < 4)
	{
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "SOP");
	    return;
	}

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	{
	    if (!ca->level && !(ca->level == 2))
		continue;

	    j++;

	    if (!stricmp (mask, ca->mask))
	    {
		success = 1;
		break;
	    }

	    if (atoi (mask) == j)
	    {
		success = 1;
		break;
	    }
	}

	if (!success)
	{
	    notice (s_ChanServ, source, CS_ACCESS_NOT_FOUND, mask, ci->name, "HOP");
	    return;
	}
	else
	{
	    notice (s_ChanServ, source, CS_ACCESS_REMOVED, ca->mask, ci->name, "HOP");

	    if (ci->flags & CF_VERBOSE)
		opnotice (s_ChanServ, ci->name, CS_VERBOSE_REMOVED, source, ca->mask, "HOP");

	    if (ca->mask)
		free (ca->mask);

	    ca->level = 0;
	}

	return;
    }

    /* LIST */
    if (!stricmp (option, "LIST"))
    {
	if (ulev < 1)
	{
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "VOP");
	    return;
	}

	notice (s_ChanServ, source, CS_ACCESS_LIST_START, "HOP", ci->name);
	notice (s_ChanServ, source, CS_ACCESS_LIST_HEADER);

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	{
	    if (!ca->level || !(ca->level == 2))
		continue;

	    j++;

	    tni = findnick (ca->mask);

	    notice (s_ChanServ, source, CS_ACCESS_LIST, j, ca->mask,
		tni->usermask ? "(" : "", tni->usermask ? tni->usermask : "",
		tni->usermask ? ")" : "");
	}

	notice (s_ChanServ, source, CS_ACCESS_LIST_END, j);
    }
}
#endif

/* Add, Remove, or List AOP entries. */
static void do_aop (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *option = strtok (NULL, " ");
    char *mask = strtok (NULL, " ");
    ChanInfo *ci;
    ChanAccess *ca;
    NickInfo *ni, *hni, *tni;
    User *u = finduser (source);
    int i, j = 0, ulev;

    if (!u)
	return;

    /* Get their nickname */
    ni = get_nick (u);

    if (!ni)
    {
	notice (s_ChanServ, source, ERR_NEED_IDENT, "AOP");
	return;
    }

    /* Check for everything we need */
    if (!chan || !option)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "AOP Channel ADD|DEL|LIST [Mask|Nickname]");
	errmoreinfo (s_ChanServ, source, "AOP");
	return;
    }

    /* Find the channel and make sure it's usable */
    ci = cs_findchan (chan);

    if (!ci)
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_CHAN_FROZEN, ci->name);
	return;
    }

    if (!check_cs_auth (s_ChanServ, source, ci))
	return;

    ulev = get_access (u, ci);

    /* ADD */
    if (!stricmp (option, "ADD"))
    {
	if (!mask)
	{
	    notice (s_ChanServ, source, RPL_SYNTAX, "AOP Channel ADD Nick");
	    errmoreinfo (s_ChanServ, source, "AOP");
	    return;
	}

	if (ulev < 4)
	{
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "SOP");
	    return;
	}

	tni = findnick (mask);

	if (!tni)
	{
	    notice (s_ChanServ, source, CS_ACCESS_REGISTERED, mask);
	    return;
	}

	/* Check for a host nick */
	if (tni->host)
	    hni = findnick (tni->host);
	else
	    hni = tni;

	if (!check_ns_auth (s_ChanServ, source, hni))
	    return;

	/* Check for NoOp */
	if (hni->flags & NF_NOOP)
	{
	    notice (s_ChanServ, source, CS_NO_OP, hni->nick);
	    return;
	}

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	{
	    if (!ca->level)
		continue;

	    if (!stricmp (ca->mask, hni->nick))
	    {
		/* Do they already have this level? */
		if (ca->level == 3)
		{
		    notice (s_ChanServ, source, CS_ACCESS_ALREADY, hni->nick, "AOP", ci->name);
		    return;
		}

		/* Check if they're trying to change their own access. In this case,
		   the check for greater or equal access below will work, but we want to
		   return a different message.
		 */
		if (!stricmp (ni->nick, ca->mask))
		{
		    notice (s_ChanServ, source, CS_ACCESS_SELF);
		    return;
		}

		/* Do they have greater or equal access than this user? */
		if (ca->level >= ulev)
		{
		    notice (s_ChanServ, source, CS_ACCESS_DENIED, "Founder");
		    return;
		}

		notice (s_ChanServ, source, CS_ACCESS_CHANGED, hni->nick, ci->name,
		    ca->level == 1 ? "VOP" : ca->level == 2 ? "HOP" :
		    ca->level == 4 ? "SOP" : "Unknown", "AOP");

		if (ci->flags & CF_VERBOSE)
		    opnotice (s_ChanServ, ci->name, CS_VERBOSE_CHANGED, source, hni->nick,
			ca->level == 1 ? "VOP" : ca->level == 2 ? "HOP" : ca->level == 4 ? "SOP" :
			"Unknown", "AOP");

		ca->level = 3;

		return;
	    }
	}

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	    if (!ca->level)
		break;

	if (i == ci->accesscnt)
	{
	    ci->accesscnt++;
	    ci->access = srealloc (ci->access, sizeof (ChanAccess) * ci->accesscnt);
	    ca = &ci->access[ci->accesscnt - 1];
	}

	ca->mask = sstrdup (hni->nick);
	ca->level = 3;

	notice (s_ChanServ, source, CS_ACCESS_ADDED, hni->nick, ci->name, "AOP");

	if (ci->flags & CF_VERBOSE)
	    opnotice (s_ChanServ, ci->name, CS_VERBOSE_ADDED, source, hni->nick, "AOP");

	return;
    }

    /* DEL */
    if (!stricmp (option, "DEL"))
    {
	int success = 0;

	if (!mask)
	{
	    notice (s_ChanServ, source, RPL_SYNTAX, "AOP Channel DEL Nick");
	    errmoreinfo (s_ChanServ, source, "AOP");
	    return;
	}

	if (ulev < 4)
	{
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "SOP");
	    return;
	}

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	{
	    if (!ca->level && !(ca->level == 3))
		continue;

	    j++;

	    if (!stricmp (mask, ca->mask))
	    {
		success = 1;
		break;
	    }

	    if (atoi (mask) == j)
	    {
		success = 1;
		break;
	    }
	}

	if (!success)
	{
	    notice (s_ChanServ, source, CS_ACCESS_NOT_FOUND, mask, ci->name, "AOP");
	    return;
	}
	else
	{
	    notice (s_ChanServ, source, CS_ACCESS_REMOVED, ca->mask, ci->name, "AOP");

	    if (ci->flags & CF_VERBOSE)
		opnotice (s_ChanServ, ci->name, CS_VERBOSE_REMOVED, source, ca->mask, "AOP");

	    if (ca->mask)
		free (ca->mask);

	    ca->level = 0;
	}

	return;
    }

    /* LIST */
    if (!stricmp (option, "LIST"))
    {
	if (ulev < 1)
	{
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "VOP");
	    return;
	}

	notice (s_ChanServ, source, CS_ACCESS_LIST_START, "AOP", ci->name);
	notice (s_ChanServ, source, CS_ACCESS_LIST_HEADER);

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	{
	    if (!ca->level || !(ca->level == 3))
		continue;

	    j++;

	    tni = findnick (ca->mask);

	    notice (s_ChanServ, source, CS_ACCESS_LIST, j, ca->mask,
		tni->usermask ? "(" : "", tni->usermask ? tni->usermask : "",
		tni->usermask ? ")" : "");
	}

	notice (s_ChanServ, source, CS_ACCESS_LIST_END, j);
    }
}

/* Add, Remove, or List SOP entries. */
static void do_sop (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *option = strtok (NULL, " ");
    char *mask = strtok (NULL, " ");
    ChanInfo *ci;
    ChanAccess *ca;
    NickInfo *ni, *hni, *tni;
    User *u = finduser (source);
    int i, j = 0, ulev;

    if (!u)
	return;

    /* Get their nickname */
    ni = get_nick (u);

    if (!ni)
    {
	notice (s_ChanServ, source, ERR_NEED_IDENT, "SOP");
	return;
    }

    /* Check for everything we need */
    if (!chan || !option)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "SOP Channel ADD|DEL|LIST [Mask|Nickname]");
	errmoreinfo (s_ChanServ, source, "SOP");
	return;
    }

    /* Find the channel and make sure it's usable */
    ci = cs_findchan (chan);

    if (!ci)
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_CHAN_FROZEN, ci->name);
	return;
    }

    if (!check_cs_auth (s_ChanServ, source, ci))
	return;

    ulev = get_access (u, ci);

    /* ADD */
    if (!stricmp (option, "ADD"))
    {
	if (!mask)
	{
	    notice (s_ChanServ, source, RPL_SYNTAX, "SOP Channel ADD Nick");
	    errmoreinfo (s_ChanServ, source, "SOP");
	    return;
	}

	if (ulev < 4)
	{
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "SOP");
	    return;
	}

	tni = findnick (mask);

	if (!tni)
	{
	    notice (s_ChanServ, source, CS_ACCESS_REGISTERED, mask);
	    return;
	}

	/* Check for a host nick */
	if (tni->host)
	    hni = findnick (tni->host);
	else
	    hni = tni;

	if (!check_ns_auth (s_ChanServ, source, hni))
	    return;

	/* Check for NoOp */
	if (hni->flags & NF_NOOP)
	{
	    notice (s_ChanServ, source, CS_NO_OP, hni->nick);
	    return;
	}

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	{
	    if (!ca->level)
		continue;

	    if (!stricmp (ca->mask, hni->nick))
	    {
		/* Do they already have this level? */
		if (ca->level == 4)
		{
		    notice (s_ChanServ, source, CS_ACCESS_ALREADY, hni->nick, "SOP", ci->name);
		    return;
		}

		/* Do they have greater or equal access than this user? */
		if (ca->level >= ulev)
		{
		    notice (s_ChanServ, source, CS_ACCESS_DENIED, "Founder");
		    return;
		}

		notice (s_ChanServ, source, CS_ACCESS_CHANGED, hni->nick, ci->name,
		    ca->level == 1 ? "VOP" : ca->level == 2 ? "HOP" :
		    ca->level == 3 ? "AOP" : "Unknown", "SOP");

		if (ci->flags & CF_VERBOSE)
		    opnotice (s_ChanServ, ci->name, CS_VERBOSE_CHANGED, source, hni->nick,
			ca->level == 1 ? "VOP" : ca->level == 2 ? "HOP" : ca->level == 3 ? "AOP" :
			"Unknown", "SOP");

		ca->level = 4;

		return;
	    }
	}

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	    if (!ca->level)
		break;

	if (i == ci->accesscnt)
	{
	    ci->accesscnt++;
	    ci->access = srealloc (ci->access, sizeof (ChanAccess) * ci->accesscnt);
	    ca = &ci->access[ci->accesscnt - 1];
	}

	ca->mask = sstrdup (hni->nick);
	ca->level = 4;

	notice (s_ChanServ, source, CS_ACCESS_ADDED, hni->nick, ci->name, "SOP");

	if (ci->flags & CF_VERBOSE)
	    opnotice (s_ChanServ, ci->name, CS_VERBOSE_ADDED, source, hni->nick, "SOP");

	return;
    }

    /* DEL */
    if (!stricmp (option, "DEL"))
    {
	int success = 0;

	if (!mask)
	{
	    notice (s_ChanServ, source, RPL_SYNTAX, "SOP Channel DEL Nick");
	    errmoreinfo (s_ChanServ, source, "SOP");
	    return;
	}

	if (ulev < 4)
	{
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "SOP");
	    return;
	}

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	{
	    if (!ca->level && !(ca->level == 4))
		continue;

	    j++;

	    if (!stricmp (mask, ca->mask))
	    {
		success = 1;
		break;
	    }

	    if (atoi (mask) == j)
	    {
		success = 1;
		break;
	    }
	}

	if (!success)
	{
	    notice (s_ChanServ, source, CS_ACCESS_NOT_FOUND, mask, ci->name, "SOP");
	    return;
	}
	else
	{
	    notice (s_ChanServ, source, CS_ACCESS_REMOVED, ca->mask, ci->name, "SOP");

	    if (ci->flags & CF_VERBOSE)
		opnotice (s_ChanServ, ci->name, CS_VERBOSE_REMOVED, source, ca->mask, "SOP");

	    if (ca->mask)
		free (ca->mask);

	    ca->level = 0;
	}

	return;
    }

    /* LIST */
    if (!stricmp (option, "LIST"))
    {
	if (ulev < 1)
	{
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "VOP");
	    return;
	}

	notice (s_ChanServ, source, CS_ACCESS_LIST_START, "SOP", ci->name);
	notice (s_ChanServ, source, CS_ACCESS_LIST_HEADER);

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	{
	    if (!ca->level || !(ca->level == 4))
		continue;

	    j++;

	    tni = findnick (ca->mask);

	    notice (s_ChanServ, source, CS_ACCESS_LIST, j, ca->mask,
		tni->usermask ? "(" : "", tni->usermask ? tni->usermask : "",
		tni->usermask ? ")" : "");
	}

	notice (s_ChanServ, source, CS_ACCESS_LIST_END, j);
    }
}

/* Op someone on the given channel */
static void do_op (const char *source)
{
    ChanInfo *ci;
    NickInfo *ni;
    User *u = finduser (source), *uo = NULL;
    char *chan = strtok (NULL, " ");
    char *nick = strtok (NULL, " ");
    int ulev = 0, uolev = 0;
    char *av[3];

    if (!u)
	return;

    /* Get their nickname */
    ni = get_nick (u);

    if (!ni)
    {
	notice (s_ChanServ, source, ERR_NEED_IDENT, "OP");
	return;
    }

    /* Check for everything we need */
    if (!chan)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "OP Channel [Nickname]");
	errmoreinfo (s_ChanServ, source, "OP");
	return;
    }

    /* Find the channel and make sure it's usable */
    ci = cs_findchan (chan);

    if (!ci)
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_CHAN_FROZEN, ci->name);
	return;
    }

    if (!check_cs_auth (s_ChanServ, source, ci))
	return;

    /* Get their access */
    ulev = get_access (u, ci);

    /* Get the other users access */
    if (nick)
    {
	uo = finduser (nick);

	if (!uo || !is_on_chan (nick, ci->name))
	{
	    notice (s_ChanServ, source, CS_NOT_ON, nick, ci->name);
	    return;
	}

	if (is_opped (nick, ci->name))
	{
	    notice (s_ChanServ, source, CS_IS_ALREADY, nick,
		"opped", ci->name);
	    return;
	}

	uolev = get_access (uo, ci);
    }

    if (!nick && is_opped (source, ci->name))
    {
	notice (s_ChanServ, source, CS_YOURE_ALREADY, "opped", ci->name);
	return;
    }

    /* Make sure they have enough access to op people. (AOP) */
    if (ulev < 3)
    {
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "AOP");
	return;
    }

    /* If Secure is on, make sure the person they want to op has
       AOP or higher.
     */
    if (uo && (ci->flags & CF_SECURE) && uolev < 3)
    {
	notice (s_ChanServ, source, CS_ACCESS_DENIED_OTHER, uo->nick,
	    "AOP");
	return;
    }

    /* Everything checks out, do the opping. */
    if (nick)
    {
	av[0] = ci->name;
	av[1] = "+o";
	av[2] = uo->nick;
	do_cmode (s_ChanServ, 3, av);

	send_cmode (s_ChanServ, ci->name, "+o", uo->nick);

	/* Only notice them if they didnt op themselves */
	if (stricmp (uo->nick, source))
	{
	    notice (s_ChanServ, source, CS_HAS_BEEN, uo->nick, "opped",
		ci->name);
	    notice (s_ChanServ, uo->nick, CS_YOU_WERE, "opped",
		ci->name, source);
	}
    }
    else
    {
	av[0] = ci->name;
	av[1] = "+o";
	av[2] = u->nick;
	do_cmode (s_ChanServ, 3, av);

	send_cmode (s_ChanServ, ci->name, "+o", source);
    }
}

/* DeOp someone on the given channel */
static void do_deop (const char *source)
{
    ChanInfo *ci;
    NickInfo *ni;
    User *u = finduser (source), *uo = NULL;
    char *chan = strtok (NULL, " ");
    char *nick = strtok (NULL, " ");
    int ulev = 0, uolev = 0;
    char *av[3];

    if (!u)
	return;

    /* Get their nickname */
    ni = get_nick (u);

    if (!ni)
    {
	notice (s_ChanServ, source, ERR_NEED_IDENT, "DEOP");
	return;
    }

    /* Check for everything we need */
    if (!chan)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "DEOP Channel [Nickname]");
	errmoreinfo (s_ChanServ, source, "DEOP");
	return;
    }

    /* Find the channel and make sure it's usable */
    ci = cs_findchan (chan);

    if (!ci)
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_CHAN_FROZEN, ci->name);
	return;
    }

    if (!check_cs_auth (s_ChanServ, source, ci))
	return;

    /* Get their access */
    ulev = get_access (u, ci);

    /* Get the other users access */
    if (nick)
    {
	uo = finduser (nick);

	if (!uo || !is_on_chan (nick, ci->name))
	{
	    notice (s_ChanServ, source, CS_NOT_ON, nick, ci->name);
	    return;
	}

	if (!is_opped (nick, ci->name))
	{
	    notice (s_ChanServ, source, CS_IS_ALREADY, nick,
		"deopped", ci->name);
	    return;
	}

	uolev = get_access (uo, ci);
    }

    if (!nick && !is_opped (source, ci->name))
    {
	notice (s_ChanServ, source, CS_YOURE_ALREADY, "deopped", ci->name);
	return;
    }

    /* Make sure they have enough access to deop people. (AOP) */
    if (ulev < 3)
    {
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "AOP");
	return;
    }

    /* Everything checks out, do the deopping. */
    if (nick)
    {
	av[0] = ci->name;
	av[1] = "-o";
	av[2] = uo->nick;
	do_cmode (s_ChanServ, 3, av);

	send_cmode (s_ChanServ, ci->name, "-o", uo->nick);

	/* Only notice them if they didnt deop themselves */
	if (stricmp (uo->nick, source))
	{
	    notice (s_ChanServ, source, CS_HAS_BEEN, uo->nick, "deopped",
		ci->name);
	    notice (s_ChanServ, uo->nick, CS_YOU_WERE, "deopped",
		ci->name, source);
	}
    }
    else
    {
	av[0] = ci->name;
	av[1] = "-o";
	av[2] = u->nick;
	do_cmode (s_ChanServ, 3, av);

	send_cmode (s_ChanServ, ci->name, "-o", source);
    }
}

#ifdef UNREAL3
/* HalfOp someone on the given channel */
static void do_halfop (const char *source)
{
    ChanInfo *ci;
    NickInfo *ni;
    User *u = finduser (source), *uo = NULL;
    char *chan = strtok (NULL, " ");
    char *nick = strtok (NULL, " ");
    int ulev = 0, uolev = 0;
    char *av[3];

    if (!u)
	return;

    /* Get their nickname */
    ni = get_nick (u);

    if (!ni)
    {
	notice (s_ChanServ, source, ERR_NEED_IDENT, "HALFOP");
	return;
    }

    /* Check for everything we need */
    if (!chan)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "HALFOP Channel [Nickname]");
	errmoreinfo (s_ChanServ, source, "HALFOP");
	return;
    }

    /* Find the channel and make sure it's usable */
    ci = cs_findchan (chan);

    if (!ci)
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_CHAN_FROZEN, ci->name);
	return;
    }

    if (!check_cs_auth (s_ChanServ, source, ci))
	return;

    /* Get their access */
    ulev = get_access (u, ci);

    /* Get the other users access */
    if (nick)
    {
	uo = finduser (nick);

	if (!uo || !is_on_chan (nick, ci->name))
	{
	    notice (s_ChanServ, source, CS_NOT_ON, nick, ci->name);
	    return;
	}

	if (is_halfopped (nick, ci->name))
	{
	    notice (s_ChanServ, source, CS_IS_ALREADY, nick,
		"halfopped", ci->name);
	    return;
	}

	uolev = get_access (uo, ci);
    }

    if (!nick && is_halfopped (source, ci->name))
    {
	notice (s_ChanServ, source, CS_YOURE_ALREADY, "halfopped", ci->name);
	return;
    }

    /* Make sure they have enough access to halfop people. (AOP) */
    if (ulev < 3)
    {
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "AOP");
	return;
    }

    /* If Secure is on, make sure the person they want to halfop has
       HOP or higher.
     */
    if (uo && (ci->flags & CF_SECURE) && uolev < 2)
    {
	notice (s_ChanServ, source, CS_ACCESS_DENIED_OTHER, uo->nick,
	    "HOP");
	return;
    }

    /* Everything checks out, do the halfopping. */
    if (nick)
    {
	av[0] = ci->name;
	av[1] = "+h";
	av[2] = uo->nick;
	do_cmode (s_ChanServ, 3, av);

	send_cmode (s_ChanServ, ci->name, "+h", uo->nick);

	/* Only notice them if they didnt halfop themselves */
	if (stricmp (uo->nick, source))
	{
	    notice (s_ChanServ, source, CS_HAS_BEEN, uo->nick, "halfopped",
		ci->name);
	    notice (s_ChanServ, uo->nick, CS_YOU_WERE, "halfopped",
		ci->name, source);
	}
    }
    else
    {
	av[0] = ci->name;
	av[1] = "+h";
	av[2] = u->nick;
	do_cmode (s_ChanServ, 3, av);

	send_cmode (s_ChanServ, ci->name, "+h", source);
    }
}

/* DeHalfOp someone on the given channel */
static void do_dehalfop (const char *source)
{
    ChanInfo *ci;
    NickInfo *ni;
    User *u = finduser (source), *uo = NULL;
    char *chan = strtok (NULL, " ");
    char *nick = strtok (NULL, " ");
    int ulev = 0, uolev = 0;
    char *av[3];

    if (!u)
	return;

    /* Get their nickname */
    ni = get_nick (u);

    if (!ni)
    {
	notice (s_ChanServ, source, ERR_NEED_IDENT, "DEHALFOP");
	return;
    }

    /* Check for everything we need */
    if (!chan)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "DEHALFOP Channel [Nickname]");
	errmoreinfo (s_ChanServ, source, "DEHALFOP");
	return;
    }

    /* Find the channel and make sure it's usable */
    ci = cs_findchan (chan);

    if (!ci)
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_CHAN_FROZEN, ci->name);
	return;
    }

    if (!check_cs_auth (s_ChanServ, source, ci))
	return;

    /* Get their access */
    ulev = get_access (u, ci);

    /* Get the other users access */
    if (nick)
    {
	uo = finduser (nick);

	if (!uo || !is_on_chan (nick, ci->name))
	{
	    notice (s_ChanServ, source, CS_NOT_ON, nick, ci->name);
	    return;
	}

	if (!is_halfopped (nick, ci->name))
	{
	    notice (s_ChanServ, source, CS_IS_ALREADY, nick,
		"dehalfopped", ci->name);
	    return;
	}

	uolev = get_access (uo, ci);
    }

    if (!nick && !is_halfopped (source, ci->name))
    {
	notice (s_ChanServ, source, CS_YOURE_ALREADY, "dehalfopped", ci->name);
	return;
    }

    /* Make sure they have enough access to dehalfop people. (AOP)
       If they want to dehalfop themselves, we'll always let them,
       regardless of access.
     */
    if (ulev < 3 && !nick)
    {
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "AOP");
	return;
    }

    /* Everything checks out, do the dehalfopping. */
    if (nick)
    {
	av[0] = ci->name;
	av[1] = "-h";
	av[2] = uo->nick;
	do_cmode (s_ChanServ, 3, av);

	send_cmode (s_ChanServ, ci->name, "-h", uo->nick);

	/* Only notice them if they didnt dehalfop themselves */
	if (stricmp (uo->nick, source))
	{
	    notice (s_ChanServ, source, CS_HAS_BEEN, uo->nick, "dehalfopped",
		ci->name);
	    notice (s_ChanServ, uo->nick, CS_YOU_WERE, "dehalfopped",
		ci->name, source);
	}
    }
    else
    {
	av[0] = ci->name;
	av[1] = "-h";
	av[2] = u->nick;
	do_cmode (s_ChanServ, 3, av);

	send_cmode (s_ChanServ, ci->name, "-h", source);
    }
}
#endif UNREAL3

/* Voice someone on the given channel */
static void do_voice (const char *source)
{
    ChanInfo *ci;
    NickInfo *ni;
    User *u = finduser (source), *uo = NULL;
    char *chan = strtok (NULL, " ");
    char *nick = strtok (NULL, " ");
    int ulev = 0, uolev = 0;
    char *av[3];

    if (!u)
	return;

    /* Get their nickname */
    ni = get_nick (u);

    if (!ni)
    {
	notice (s_ChanServ, source, ERR_NEED_IDENT, "VOICE");
	return;
    }

    /* Check for everything we need */
    if (!chan)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "VOICE Channel [Nickname]");
	errmoreinfo (s_ChanServ, source, "VOICE");
	return;
    }

    /* Find the channel and make sure it's usable */
    ci = cs_findchan (chan);

    if (!ci)
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_CHAN_FROZEN, ci->name);
	return;
    }

    if (!check_cs_auth (s_ChanServ, source, ci))
	return;

    /* Get their access */
    ulev = get_access (u, ci);

    /* Get the other users access */
    if (nick)
    {
	uo = finduser (nick);

	if (!uo || !is_on_chan (nick, ci->name))
	{
	    notice (s_ChanServ, source, CS_NOT_ON, nick, ci->name);
	    return;
	}

	if (is_voiced (nick, ci->name))
	{
	    notice (s_ChanServ, source, CS_IS_ALREADY, nick,
		"voiced", ci->name);
	    return;
	}

	uolev = get_access (uo, ci);
    }

    if (!nick && is_voiced (source, ci->name))
    {
	notice (s_ChanServ, source, CS_YOURE_ALREADY, "voiced", ci->name);
	return;
    }

    /* Make sure they have enough access to voice people. (AOP) */
    if (ulev < 3)
    {
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "AOP");
	return;
    }

    /* If Secure is on, make sure the person they want to voice has
       VOP or higher.
     */
    if (uo && (ci->flags & CF_SECURE) && uolev < 1)
    {
	notice (s_ChanServ, source, CS_ACCESS_DENIED_OTHER, uo->nick,
	    "VOP");
	return;
    }

    /* Everything checks out, do the halfopping. */
    if (nick)
    {
	av[0] = ci->name;
	av[1] = "+v";
	av[2] = uo->nick;
	do_cmode (s_ChanServ, 3, av);

	send_cmode (s_ChanServ, ci->name, "+v", uo->nick);

	/* Only notice them if they didnt voice themselves */
	if (stricmp (uo->nick, source))
	{
	    notice (s_ChanServ, source, CS_HAS_BEEN, uo->nick, "voiced",
		ci->name);
	    notice (s_ChanServ, uo->nick, CS_YOU_WERE, "voiced",
		ci->name, source);
	}
    }
    else
    {
	av[0] = ci->name;
	av[1] = "-v";
	av[2] = u->nick;
	do_cmode (s_ChanServ, 3, av);

	send_cmode (s_ChanServ, ci->name, "+v", source);
    }
}

/* DeVoice someone on the given channel */
static void do_devoice (const char *source)
{
    ChanInfo *ci;
    NickInfo *ni;
    User *u = finduser (source), *uo = NULL;
    char *chan = strtok (NULL, " ");
    char *nick = strtok (NULL, " ");
    int ulev = 0, uolev = 0;
    char *av[3];

    if (!u)
	return;

    /* Get their nickname */
    ni = get_nick (u);

    if (!ni)
    {
	notice (s_ChanServ, source, ERR_NEED_IDENT, "DEVOICE");
	return;
    }

    /* Check for everything we need */
    if (!chan)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "DEVOICE Channel [Nickname]");
	errmoreinfo (s_ChanServ, source, "DEVOICE");
	return;
    }

    /* Find the channel and make sure it's usable */
    ci = cs_findchan (chan);

    if (!ci)
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_CHAN_FROZEN, ci->name);
	return;
    }

    if (!check_cs_auth (s_ChanServ, source, ci))
	return;

    /* Get their access */
    ulev = get_access (u, ci);

    /* Get the other users access */
    if (nick)
    {
	uo = finduser (nick);

	if (!uo || !is_on_chan (nick, ci->name))
	{
	    notice (s_ChanServ, source, CS_NOT_ON, nick, ci->name);
	    return;
	}

	if (!is_voiced (nick, ci->name))
	{
	    notice (s_ChanServ, source, CS_IS_ALREADY, nick,
		"devoiced", ci->name);
	    return;
	}

	uolev = get_access (uo, ci);
    }

    if (!nick && !is_voiced (source, ci->name))
    {
	notice (s_ChanServ, source, CS_YOURE_ALREADY, "devoiced", ci->name);
	return;
    }

    /* Make sure they have enough access to devoice people. (AOP)
       If they want to devoice themselves, we'll always let them,
       regardless of access.
     */
    if (ulev < 3 && !nick)
    {
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "AOP");
	return;
    }

    /* Everything checks out, do the devoiceing. */
    if (nick)
    {
	av[0] = ci->name;
	av[1] = "-v";
	av[2] = uo->nick;
	do_cmode (s_ChanServ, 3, av);

	send_cmode (s_ChanServ, ci->name, "-v", uo->nick);

	/* Only notice them if they didnt devoice themselves */
	if (stricmp (uo->nick, source))
	{
	    notice (s_ChanServ, source, CS_HAS_BEEN, uo->nick, "devoiced",
		ci->name);
	    notice (s_ChanServ, uo->nick, CS_YOU_WERE, "devoiced",
		ci->name, source);
	}
    }
    else
    {
	av[0] = ci->name;
	av[1] = "-v";
	av[2] = u->nick;
	do_cmode (s_ChanServ, 3, av);

	send_cmode (s_ChanServ, ci->name, "-v", source);
    }
}

/* Add, Remove, or List AKick entries. */
static void do_akick (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *option = strtok (NULL, " ");
    char *mask = strtok (NULL, " ");
    char *reason = strtok (NULL, "");
    char *akmask;
    ChanInfo *ci;
    Channel *c;
    AKick *ak;
    NickInfo *ni, *tni;
    User *aku, *u = finduser (source);
    int i, j = 0, ulev, isnick = 0;

    if (!u)
	return;

    /* Get their nickname */
    ni = get_nick (u);

    if (!ni)
    {
	notice (s_ChanServ, source, ERR_NEED_IDENT, "SOP");
	return;
    }

    /* Check for everything we need */
    if (!chan || !option)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "AKICK Channel ADD|DEL|LIST [Mask] [Reason]");
	errmoreinfo (s_ChanServ, source, "AKICK");
	return;
    }

    /* Find the channel and make sure it's usable */
    ci = cs_findchan (chan);

    if (!ci)
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_CHAN_FROZEN, ci->name);
	return;
    }

    if (!check_cs_auth (s_ChanServ, source, ci))
	return;

    ulev = get_access (u, ci);

    /* ADD */
    if (!stricmp (option, "ADD"))
    {
	if (!mask)
	{
	    notice (s_ChanServ, source, RPL_SYNTAX, "AKICK Channel ADD Mask [Reason]");
	    errmoreinfo (s_ChanServ, source, "AKICK");
	    return;
	}

	if (ulev < 4)
	{
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "SOP");
	    return;
	}

	/* See if the channel has too many AKicks. */
	if (ci->akickcnt >= akickmax)
	{
	    notice (s_ChanServ, source, CS_AKICK_FULL, ci->name, akickmax);
	    return;
	}

	/* First, try to find a registered nickname. If we find this, we'll AKick
	   that nick (thus getting a new host for it whenever someone id'd to that
	   nick joins).
	 */
	tni = findnick (mask);

	/* Check for a host nick */
	if (tni)
	{
	    if (tni->host)
		akmask = tni->host;
	    else
		akmask = tni->nick;

	    isnick = 1;
	}
	else
	{
	    /* Not a registered nick, try to find an online user */
	    aku = finduser (mask);

	    if (aku)
		akmask = create_mask (aku);
	    else
	    {
		/* Not an online user either.. Check if it's already a hostmask */
		if (!strchr (mask, '@') || !strchr (mask, '.'))
		{
		    notice (s_ChanServ, source, ERR_INVALID, "hostmask",
			" Hostmasks must be in the format of user@host.tld.");
		    return;
		}
		else
		    akmask = mask;
	    }

	    isnick = 0;
	}

	/* Check if its already present. */
	for (ak = ci->akick, i = 0; i < ci->akickcnt; ak++, i++)
	{
	    if (!ak->level)
		continue;

	    if (!stricmp (ak->mask, akmask))
	    {
		notice (s_ChanServ, source, CS_AKICK_ALREADY, akmask, ci->name);
		return;
	    }
	}

	for (ak = ci->akick, i = 0; i < ci->akickcnt; ak++, i++)
	    if (!ak->level)
		break;

	if (i == ci->akickcnt)
	{
	    ci->akickcnt++;
	    ci->akick = srealloc (ci->akick, sizeof (AKick) * ci->akickcnt);
	    ak = &ci->akick[ci->akickcnt - 1];
	}

	ak->mask = sstrdup (akmask);

	if (isnick)
	    ak->level = 2;
	else
	    ak->level = 1;

	if (reason)
	    ak->reason = sstrdup (reason);
	else
	    ak->reason = NULL;

	ak->time = time (NULL);

	strscpy (ak->setter, source, NICKLEN);

	notice (s_ChanServ, source, CS_AKICK_ADDED, akmask, ci->name);

	if (ci->flags & CF_VERBOSE)
	    opnotice (s_ChanServ, ci->name, CS_VERBOSE_AKICKED, source, akmask);

	c = findchan (ci->name);

	/* Kick anyone in the channel who matches the AKick */
	if (c)
	{
	    NickInfo *ni;
	    struct c_userlist *cu, *next;
	    char *av[3], tmp[BUFSIZE];
	    int found = 0;

	    av[0] = c->name;

	    snprintf (tmp, sizeof (tmp), "%s@%s", u->user, u->host);

	    if (reason)
		av[2] = (char *)reason;
	    else
		av[2] = (char *)def_akick_reason;

	    for (cu = c->users; cu; cu = next)
	    {
		int akicked = 0;
		ni = get_nick (cu->user);
		next = cu->next;
		av[1] = cu->user->nick;

		if (ak->level == 2)
		{
		    if (ni && !stricmp (ni->nick, ak->mask))
			akicked = 1;

		    if (!akicked && !stricmp (u->nick, ak->mask))
			akicked = 1;
		}
    
		if (ak->level == 1)
		    if (match_wild_nocase (tmp, ak->mask))
			akicked = 1;

		if (akicked)
		{
		    if (!found)
		    {
			send_cmd (s_ChanServ, "MODE %s +b %s", c->name, ak->mask);
			found = 1;
		    }

		    send_cmd (s_ChanServ, "KICK %s %s :%s", av[0], av[1], av[2]);
		    do_kick (s_ChanServ, 3, av);
		}
	    }
	}

	return;
    }

    /* DEL */
    if (!stricmp (option, "DEL"))
    {
	int success = 0;

	if (!mask)
	{
	    notice (s_ChanServ, source, RPL_SYNTAX, "AKICK Channel DEL Nick|Mask|Num");
	    errmoreinfo (s_ChanServ, source, "AKICK");
	    return;
	}

	if (ulev < 4)
	{
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "SOP");
	    return;
	}

	for (ak = ci->akick, i = 0; i < ci->akickcnt; ak++, i++)
	{
	    if (!ak->level)
		continue;

	    j++;

	    if (!stricmp (mask, ak->mask))
	    {
		success = 1;
		break;
	    }

	    if (atoi (mask) == j)
	    {
		success = 1;
		break;
	    }
	}

	if (!success)
	{
	    notice (s_ChanServ, source, CS_AKICK_NOT_FOUND, mask, ci->name);
	    return;
	}
	else
	{
	    notice (s_ChanServ, source, CS_AKICK_REMOVED, ak->mask, ci->name);

	    if (ci->flags & CF_VERBOSE)
		opnotice (s_ChanServ, ci->name, CS_VERBOSE_UNAKICK, source, ak->mask);

	    if (ak->mask)
		free (ak->mask);

	    if (ak->reason)
		free (ak->reason);

	    ak->level = 0;
	}

	return;
    }

    /* LIST */
    if (!stricmp (option, "LIST"))
    {
	if (ulev < 1)
	{
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "VOP");
	    return;
	}

	notice (s_ChanServ, source, CS_AKICK_LIST_START, ci->name);
	notice (s_ChanServ, source, CS_AKICK_LIST_HEADER);

	for (ak = ci->akick, i = 0; i < ci->akickcnt; ak++, i++)
	{
	    if (!ak->level)
		continue;

	    j++;

	    notice (s_ChanServ, source, CS_AKICK_LIST, j, ak->mask,
		ak->setter, time_ago (ak->time));

	    if (ak->reason)
		notice (s_ChanServ, source, "     Reason: %s", ak->reason);
	}

	notice (s_ChanServ, source, CS_AKICK_LIST_END, j);
    }
}

/* Freeze a channel, making it unusable. */
static void do_freeze (const char *source)
{
    ChanInfo *ci;
    Channel *c;
    User *u = finduser (source);
    char *chan = strtok (NULL, " ");
    char *reason = strtok (NULL, "");

    if (!u)
	return;

    if (!chan || !reason)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "FREEZE Channel Reason");
	errmoreinfo (s_ChanServ, source, "FREEZE");
	return;
    }

    ci = cs_findchan (chan);
    c = findchan (chan);

    if (!ci)
    {	
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (ci->flags & CF_FROZEN)
    {
	notice (s_ChanServ, source, CS_ALREADY_FROZEN, ci->name);
	return;
    }

    /* Freeze! */
    ci->flags |= CF_FROZEN;

    /* We'll abuse the topic fields for frozen channels. */
    strscpy (ci->topic, reason, TOPICLEN);
    strscpy (ci->topicsetter, source, NICKLEN);
    ci->topictime = time (NULL);

    snoop (s_ChanServ, "CS:FREEZE: %s %s!%s@%s", ci->name, source, u->user,
	u->host);
    log ("CS:FREEZE: %s %s!%s@%s", ci->name, source, u->user, u->host);

    if (c)
    {
	struct c_userlist *cu, *next;
	char *av[3];

	send_cmd (s_ChanServ, "JOIN %s", ci->name);
	add_cs_timeout (ci, TO_UNINHABIT, 60);

	/* Ban everyone. */
	send_cmode (s_ChanServ, ci->name, "+b", "*!*@*");
	send_cmode (NULL);

	/* Now empty the channel. */
	av[0] = c->name;
	av[2] = (char *)reason;

	for (cu = c->users; cu; cu = next)
	{
	    next = cu->next;
	    av[1] = cu->user->nick;

	    send_cmd (s_ChanServ, "KICK %s %s :This channel is frozen: %s", av[0], av[1], av[2]);
	    do_kick (s_ChanServ, 3, av);
	}
    }

    notice (s_ChanServ, source, CS_FROZEN, ci->name);
}

/* UnFreeze a channel */
static void do_unfreeze (const char *source)
{
    ChanInfo *ci;
    User *u = finduser (source);
    char *chan = strtok (NULL, " ");

    if (!u)
	return;

    ci = cs_findchan (chan);

    if (ci)
    {
	if (!(ci->flags & CF_FROZEN))
	{
	    notice (s_ChanServ, source, CS_NOT_FROZEN, ci->name);
	    return;
	}

	ci->flags &= ~CF_FROZEN;

	/* Set the topic info to nothing, to get rid of our freeze info */
	strscpy (ci->topic, "", TOPICLEN);
	strscpy (ci->topicsetter, s_ChanServ, NICKLEN);
	ci->topictime = time (NULL);

	snoop (s_ChanServ, "CS:FREEZE: %s %s!%s@%s", ci->name, source, u->user,
 	    u->host);
	log ("CS:FREEZE: %s %s!%s@%s", ci->name, source, u->user, u->host);

	/* Send a part just in case ChanServ is inhabiting the channel */
	send_cmd (s_ChanServ, "PART %s", ci->name);

	notice (s_ChanServ, source, CS_UNFROZEN, ci->name);
    }
    else
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
}

/* Delete a registered channel. Note: We ignore the channel successor here.
   Successors are only used when the founders nick expires.
 */
int delchan (ChanInfo *ci)
{
#ifndef IRCUP9
    Channel *c = findchan (ci->name);
#endif
    Verify *vf;
    ChanAccess *ca;
    AKick *ak;
    int i, j = 0;

    /* If this channel was dropped while awaiting verification, delete it from the
       Verify list.
     */
    vf = findverify (ci->name);

    if (vf)
    {
	if (vf->prev)
	    vf->prev->next = vf->next;
	else
	    verifylist[VHASH(vf->name)] = vf->next;

	if (vf->next)
	    vf->next->prev = vf->prev;

	free (vf);
    }

    if (ci->successor)
    {
	NickInfo *sni = findnick (ci->successor);
	char **successor;
	int i;

	/* First, we have to take this channel out of the successors
	   successor list.
	 */
	if (sni)
	{
	    for (successor = sni->successor, i = 0; i < sni->successorcnt; successor++, i++)
		if (!stricmp (sni->successor[i], ci->name))
		    break;

	    if (*successor)
		free (*successor);

	    --sni->successorcnt;

	    if (i < sni->successorcnt)
		bcopy (successor + 1, successor, (sni->successorcnt - i)
		    * sizeof (char *));

	    if (sni->successorcnt)
		sni->successor = srealloc (sni->successor, sni->successorcnt
		    * sizeof (char *));
	    else
	    {
		if (sni->successor)
		    free (sni->successor);
		sni->successor = NULL;
	    }
	}

	free (ci->successor);
    }

    if (ci->greet)
	free (ci->greet);

    if (ci->mlock_key)
	free (ci->mlock_key);

#ifdef UNREAL3
    if (ci->mlock_flood)
	free (ci->mlock_flood);

    if (ci->mlock_link)
	free (ci->mlock_link);
#endif

    for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
    {
	if (!ca->level)
	    continue;

	if (ca->mask)
	    free (ca->mask);

	ca->level = 0;

	j++;
    }

    if (j != ci->accesscnt)
	log ("delchan(): Expected %d access list entries, got %d. Memory leak?",
	    ci->accesscnt, j);

    j = 0;

    for (ak = ci->akick, i = 0; i < ci->akickcnt; ak++, i++)
    {
	if (ak->mask)
	    free (ak->mask);

	if (ak->reason)
	    free (ak->reason);

	j++;
    }

    if (j != ci->akickcnt)
	log ("delchan(): Expected %d AKick entries, got %d. Memory leak?",
	    ci->akickcnt, j);

    if (ci->prev)
	ci->prev->next = ci->next;
    else
	cs_chanlist[CSHASH(ci->name)] = ci->next;

    if (ci->next)
	ci->next->prev = ci->prev;

    free (ci);

#ifndef IRCUP9
    /* No longer registered. */
    if (c && (c->mode & CMODE_r))
    {
	c->mode &= ~CMODE_r;
	send_cmode (s_ChanServ, c->name, "-r");
    }
#endif

    return 1;
}

/* Check all channels and expire any nessecary */
void expire_chans ()
{
    ChanInfo *ci, *next;
    int j = 0;

    if (logupdates_on == TRUE)
	log_sameline (1, "CS:SYNC: Checking expires... ");

    for (ci = firstci (); ci; ci = next)
    {
	next = nextci ();

	if ((time (NULL) >= ci->lastused + chan_expire) ||
	    (ci->flags & CF_WAITAUTH && time (NULL) >=
	    ci->registered + tempexpire))
	{
	    log ("CS:SYNC:EXPIRE: %s", ci->name);

	    j++;

	    delchan (ci);
	}
    }

    /* Only log this if we have something to report, or if we should
       log it regardless.
     */
    if (logupdates_on == TRUE)
	log_sameline (0, "%d channels expired.", j);
    else if (logupdates_on == FALSE && j > 0)
	log ("CS:SYNC:EXPIRE: %d channels expired.", j);
}

/* Retrieve the password for a channel */
static void do_getpass (const char *source)
{
    ChanInfo *ci;
    char *chan = strtok (NULL, " ");

    /* Do we have all needed info? */
    if (!chan)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "GETPASS Channel");
	errmoreinfo (s_ChanServ, source, "GETPASS");
	return;
    }

    /* Is the channel registered? */
    if (!(ci = cs_findchan (chan)))
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (!check_cs_auth (s_ChanServ, source, ci))
	return;

    notice (s_ChanServ, source, RPL_GETPASS, ci->name, " (", ci->founder, ")",
	ci->pass);

    globops (s_ChanServ, RPL_PASS_GLOBAL, source, "GETPASS", ci->name,
	" (", ci->founder, ")");

    snoop (s_ChanServ, "CS:GETPASS: %s %s (%s)", source, ci->name, ci->founder);

    log ("CS:GETPASS: %s %s (%s)", source, ci->name, ci->founder);
}

/* Set the password for a channel */
static void do_setpass (const char *source)
{
    ChanInfo *ci;
    char *chan = strtok (NULL, " ");
    char *pass1 = strtok (NULL, " ");
    char *pass2 = strtok (NULL, " ");

    /* Do we have all needed info? */
    if (!chan || !pass1 || !pass2)
    {
	notice (s_ChanServ, source, RPL_SYNTAX,
	    "SETPASS Channel NewPass NewPass");
	errmoreinfo (s_ChanServ, source, "SETPASS");
	return;
    }

    /* Is the channel registered? */
    if (!(ci = cs_findchan (chan)))
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (!check_cs_auth (s_ChanServ, source, ci))
	return;

    if (strcmp (pass1, pass2))
    {
	notice (s_ChanServ, source, ERR_PASS_DONT_MATCH);
	return;
    }

    strscpy (ci->pass, pass1, sizeof (ci->pass));

    notice (s_ChanServ, source, RPL_SETPASS, ci->name, " (", ci->founder, ")",
	ci->pass);

    globops (s_ChanServ, RPL_PASS_GLOBAL, source, "SETPASS", ci->name, " (",
	ci->founder, ")");

    snoop (s_ChanServ, "CS:SETPASS: %s %s (%s)", source, ci->name, ci->founder);

    log ("CS:SETPASS: %s %s (%s)", source, ci->name, ci->founder);
}

/* Email the password for a channel */
static void do_sendpass (const char *source)
{
    ChanInfo *ci;
    NickInfo *ni;
    char *chan = strtok (NULL, " ");

    /* Do we have all needed info? */
    if (!chan)
    {
	notice (s_ChanServ, source, RPL_SYNTAX, "SENDPASS Channel");
	errmoreinfo (s_ChanServ, source, "SENDPASS");
	return;
    }

    /* Is the channel registered? */
    if (!(ci = cs_findchan (chan)))
    {
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
	return;
    }

    if (!check_cs_auth (s_ChanServ, source, ci))
	return;

    ni = findnick (ci->founder);

    if (!ni->email)
    {
	notice (s_ChanServ, source, CS_CANT_SENDPASS, ci->name);
	return;
    }

    notice (s_ChanServ, source, RPL_SENDPASS, ci->name, " (", ci->founder,
	")", ni->email);

    globops (s_ChanServ, RPL_PASS_GLOBAL, source, "SENDPASS", ci->name,
	" (", ci->founder, ")");

    snoop (s_ChanServ, "CS:SENDPASS: %s %s (%s)", source, ci->name, ci->founder);

    log ("CS:SENDPASS: %s %s (%s)", source, ci->name, ci->founder);

    sendemail (ci->name, ni->email, 4);
}

/* Load the ChanServ DB from disk */
void load_cs_dbase (void)
{
    FILE *f = fopen (CHANSERV_DB, "r");
    ChanInfo *ci = NULL;
    ChanAccess *ca;
    AKick *ak;
    char *item, dBuf[2048], backup [2048];
    int log = 0, cin = 0;
#ifdef HAVE_GETTIMEOFDAY
    struct timeval start, now, tmp;
#endif

#ifdef HAVE_GETTIMEOFDAY
    log_sameline (1, "CS:DB: Loading... ");
#else
    log ("CS:DB: Loading... ");
#endif

    if (!f)
    {
#ifdef HAVE_GETTIMEOFDAY
	log_sameline (0, "No %s found! ", CHANSERV_DB);
#else
	log ("CS:DB: No %s found!", CHANSERV_DB);
#endif
	log = 1;

	/* Check for a db.save */
	strcpy (backup, CHANSERV_DB);
	strcat (backup, ".save");

	if (rename (backup, CHANSERV_DB) > 0)
	{
#ifdef HAVE_GETTIMEOFDAY
	    log_sameline (0, "Recovering %s.save", CHANSERV_DB);
#else
	    log ("CS:DB: Recovering %s.save", CHANSERV_DB);
#endif
	    log = 1;

	    f = fopen (CHANSERV_DB, "r");
	}
	else
	    return;
    }

#ifdef HAVE_GETTIMEOFDAY
    gettimeofday (&(start), NULL);
#endif

    while (fgets (dBuf, 2047, f))
    {
	item = strtok (dBuf, " ");

	/* Ignore certain things.. */
	if (*item == '#' || *item == '\n' || *item == '\t' || *item == ' ' ||
	    *item == '\0' || *item == '\r')
	    continue;

	/* ChanInfo */
	if (!strcmp (item, "CI"))
	{
	    ChanInfo **list;

	    ci = scalloc (sizeof (ChanInfo), 1);

	    strscpy (ci->name, strtok (NULL, " "), CHANNELLEN);
	    strscpy (ci->founder, strtok (NULL, " "), NICKLEN);
	    strscpy (ci->pass, strtok (NULL, " "), PASSWDLEN);
	    ci->registered = atol (strtok (NULL, " "));
	    ci->lastused = atol (strtok (NULL, " "));
	    ci->flags = atol (strtok (NULL, " "));
	    ci->mlock_on = atol (strtok (NULL, " "));
	    ci->mlock_off = atol (strtok (NULL, " "));
	    ci->topiclock = atol (strtok (NULL, " "));
	    ci->key = atol (strtok (NULL, "\n"));

	    list = &cs_chanlist[CSHASH(ci->name)];
	    ci->next = *list;
	    if (*list)
		(*list)->prev = ci;
	    *list = ci;

	    cin++;

	    continue;
	}

	/* Access list */
	if (!strcmp (item, "CA"))
	{
	    ci->accesscnt++;
	    ci->access = srealloc (ci->access, sizeof (ChanAccess) * ci->accesscnt);
	    ca = &ci->access[ci->accesscnt - 1];

	    ca->mask = sstrdup (strtok (NULL, " "));
	    ca->level = atoi (strtok (NULL, "\n"));

	    continue;
	}

	/* AKick list */
	if (!strcmp (item, "AK"))
	{
	    char *tmp;

	    ci->akickcnt++;
	    ci->akick = srealloc (ci->akick, sizeof (AKick) * ci->akickcnt);
	    ak = &ci->akick[ci->akickcnt - 1];

	    ak->mask = sstrdup (strtok (NULL, " "));
	    strscpy (ak->setter, strtok (NULL, " "), NICKLEN);
	    ak->level = atoi (strtok (NULL, " "));
	    ak->time = atol (strtok (NULL, " "));
	    tmp = strtok (NULL, "\n");

	    if (tmp)
		ak->reason = sstrdup (tmp);
	    else
		ak->reason = NULL;

	    continue;
	}

	/* Successor */
	if (!strcmp (item, "SU"))
	{
	    ci->successor = sstrdup (strtok (NULL, "\n"));

	    continue;
	}

	/* Greet */
	if (!strcmp (item, "GR"))
	{
	    ci->greet = sstrdup (strtok (NULL, "\n"));

	    continue;
	}

	/* URL */
	if (!strcmp (item, "UR"))
	{
	    ci->url = sstrdup (strtok (NULL, "\n"));

	    continue;
	}

	/* Channel topic */
	if (!strcmp (item, "CT"))
	{
	    strscpy (ci->topicsetter, strtok (NULL, " "), NICKLEN);
	    ci->topictime = atol (strtok (NULL, " "));
	    strscpy (ci->topic, strtok (NULL, "\n"), TOPICLEN);

	    continue;
	}

	/* MLock Key */
	if (!strcmp (item, "KY"))
	{
	    ci->mlock_key = sstrdup (strtok (NULL, "\n"));
	    continue;
	}

	/* MLock Limit */
	if (!strcmp (item, "LM"))
	{
	    ci->mlock_limit = atol (strtok (NULL, "\n"));
	    continue;
	}

#ifdef UNREAL3
	/* MLock Flood */
	if (!strcmp (item, "FL"))
	{
	    ci->mlock_flood = sstrdup (strtok (NULL, "\n"));
	    continue;
	}

	/* MLock Link */
	if (!strcmp (item, "LK"))
	{
	    ci->mlock_link = sstrdup (strtok (NULL, "\n"));
	    continue;
	}
#endif

	/* Channel Verify */
	if (!strcmp (item, "VF"))
	{
	    Verify *vf, **vlist;

	    vf = scalloc (sizeof (Verify), 1);
	    strscpy (vf->name, strtok (NULL, " "), 64);
	    vf->since = atoi (strtok (NULL, "\n"));

	    vlist = &verifylist[VHASH(vf->name)];
	    vf->next = *vlist;
	    if (*vlist)
		(*vlist)->prev = vf;
	    *vlist = vf;
	}
    }

    fclose (f);

#ifdef HAVE_GETTIMEOFDAY
    if (!log)
    {
	gettimeofday (&now, NULL);
	timersub (&now, &start, &tmp);
	log_sameline (0, "finished in %d ms. (%d channels)", tv2ms (&tmp), cin);
    }
#endif
}

/* Write the ChanServ db to disk. */
void save_cs_dbase ()
{
    FILE *f;
    ChanInfo *ci;
    ChanAccess *ca;
    AKick *ak;
    Verify *vf;
    int i, j, log = 0, cout = 0;
    char backup[2048];
#ifdef HAVE_GETTIMEOFDAY
    struct timeval start, now, tmp;
#endif

    if (logupdates_on == TRUE)
#ifdef HAVE_GETTIMEOFDAY
	log_sameline (1, "CS:DB: Saving... ");
#else
	log ("CS:DB: Saving... ");
#endif

    strcpy (backup, CHANSERV_DB);
    strcat (backup, ".save");
    remove (backup);

    if (rename (CHANSERV_DB, backup) < 0)
    {
#ifdef HAVE_GETTIMEOFDAY
	log_sameline (0, "Can't back up %s!", CHANSERV_DB);
#else
	log ("CS:DB: Can't back up %s!", CHANSERV_DB);
#endif
	log = 1;
    }

    f = fopen (CHANSERV_DB, "w");

    if (!f)
    {
#ifdef HAVE_GETTIMEOFDAY
	log_sameline (0, "Can't write to %s! ", CHANSERV_DB);
#else
	log ("CS:DB: Can't write to %s!", CHANSERV_DB);
#endif
	log = 1;

	if (rename (backup, CHANSERV_DB) < 0)
	    fatal ("CS:DB: Can't restore backup of %s!", CHANSERV_DB);
	return;
    }

#ifdef HAVE_GETTIMEOFDAY
    gettimeofday (&(start), NULL);
#endif

    /* Write each channel's info */
    for (ci = firstci (); ci; ci = nextci ())
    {
	/* ChanInfo */
	fprintf (f, "CI %s %s %s %lu %lu %d %d %d %d %lu\n", ci->name,
	    ci->founder, ci->pass, ci->registered, ci->lastused,
	    ci->flags, ci->mlock_on, ci->mlock_off, ci->topiclock,
	    ci->key);

	cout++;

	/* Access list */
	for (ca = ci->access, j = 0; j < ci->accesscnt; ca++, j++)
	{
	    if (!ca->level)
		continue;

	    fprintf (f, "CA %s %d\n", ca->mask, ca->level);
	}

	/* AKick list */
	for (ak = ci->akick, j = 0; j < ci->akickcnt; ak++, j++)
	{
	    if (!ak->level)
		continue;

	    fprintf (f, "AK %s %s %d %lu %s\n", ak->mask, ak->setter,
		ak->level, ak->time, ak->reason ? ak->reason : "");
	}

	/* Successor */
	if (ci->successor)
	    fprintf (f, "SU %s\n", ci->successor);

	/* Greet */
	if (ci->greet)
	    fprintf (f, "GR %s\n", ci->greet);

	/* URL */
	if (ci->url)
	    fprintf (f, "UR %s\n", ci->url);

	/* Last topic info */
	if (strlen (ci->topic))
	    fprintf (f, "CT %s %lu %s\n", ci->topicsetter ? ci->topicsetter : s_ChanServ,
		ci->topictime, ci->topic);

	/* MLock Key */
	if (ci->mlock_key)
	    fprintf (f, "KY %s\n", ci->mlock_key);

	/* MLock Limit */
	if (ci->mlock_limit)
	    fprintf (f, "LM %d\n", ci->mlock_limit);

#ifdef UNREAL3
	/* MLock Flood */
	if (ci->mlock_flood)
	    fprintf (f, "FL %s\n", ci->mlock_flood);

	/* MLock Link */
	if (ci->mlock_link)
	    fprintf (f, "LK %s\n", ci->mlock_link);
#endif
    }

    /* Write out channel verifies */
    for (i = 0; i < 1024; i++)
	for (vf = verifylist[i]; vf; vf = vf->next)
	{
	    ci = cs_findchan (vf->name);

	    if (!ci)
	    {
		/* It's not there now .. Delete it. */
		if (vf->prev)
		    vf->prev->next = vf->next;
		else
		    verifylist[VHASH(vf->name)] = vf->next;

		if (vf->next)
		    vf->next->prev = vf->prev;

		free (vf);
	    }
	    else
		fprintf (f, "VF %s %lu\n", vf->name, vf->since);
	}

    fclose (f);
    remove (backup);

#ifdef HAVE_GETTIMEOFDAY
    if (logupdates_on == TRUE && !log)
    {
	gettimeofday (&now, NULL);
	timersub (&now, &start, &tmp);
	log_sameline (0, "finished in %d ms. (%d channels)", tv2ms (&tmp), cout);
    }
#endif
}


/* Check if the given channel is registered. If it is,
   restore the topic, modes, etc.
 */
void restore_channel (Channel *c)
{
    ChanInfo *ci = cs_findchan (c->name);

    if (!ci || (ci && (ci->flags & CF_WAITAUTH)))
	return;

    /* Restore the topic if it's present. */
#if defined (BAHAMUT) || defined (PROMETHEUS) || defined (UNREAL3)
    if (strlen (ci->topic) && !bursting)
#else
    if (strlen (ci->topic))
#endif
    {
	strscpy (c->topic, ci->topic, TOPICLEN);
	strscpy (c->topicsetter, ci->topicsetter, NICKLEN);
	c->topictime = ci->topictime;

	send_cmd (s_ChanServ, "TOPIC %s %s %lu :%s", ci->name,
	    ci->topicsetter, ci->topictime, ci->topic);
    }

    /* Restore the channel modes. */
#if defined (BAHAMUT) || defined (PROMETHEUS) || defined (UNREAL3)
    if (!bursting)
#endif
	check_modes (ci->name);
}

/* Check the current modes on the given channel. Make any mode changes
   nessecary to keep with the modelock.
 */
void check_modes (const char *chan)
{
    Channel *c = findchan (chan);
    ChanInfo *ci = cs_findchan (chan);
    char newmodes[40], *newkey = NULL;
    char *end = newmodes, params[256];
    int32 newlimit = 0;
    int modes, set_limit = 0, set_key = 0;
#ifdef UNREAL3
    char *newlink = NULL, *newflood = NULL;
    int set_link = 0, set_flood = 0;
#endif

    if (!c)
	return;

    if (!ci)
    {
#ifndef IRCUP9
	if (c->mode & CMODE_r)
	{
	    c->mode &= ~CMODE_r;
	    send_cmode (s_ChanServ, c->name, "-r");
	}
#endif

	return;
    }

    if (debuglevel > 1)
	debug ("check_modes(): %s", ci->name);

#ifndef IRCUP9
    modes = ~c->mode & (ci->mlock_on | CMODE_r);
#else
    modes = ~c->mode & (ci->mlock_on);
#endif

    end += snprintf (end, sizeof (newmodes) - (end - newmodes) - 2,
	"+%s", flags_to_string (modes));
    c->mode |= modes;

    if (ci->mlock_limit && ci->mlock_limit != c->limit)
    {
	*end++ = 'l';
	newlimit = ci->mlock_limit;
	c->limit = newlimit;
	set_limit = 1;
    }

    if (ci->mlock_key)
    {
	if (c->key && strcmp (c->key, ci->mlock_key))
	{
	    send_cmode (s_ChanServ, c->name, "-k", c->key);
	    free (c->key);
	    c->key = NULL;
	}

	if (!c->key)
	{
	    *end++ = 'k';
	    newkey = ci->mlock_key;
	    c->key = sstrdup (newkey);
	    set_key = 1;
	}
    }

#ifdef UNREAL3
    if (ci->mlock_flood)
    {
	if (c->flood && strcmp (c->flood, ci->mlock_flood))
	{
	    send_cmode (s_ChanServ, c->name, "-f", c->flood);
	    free (c->flood);
	    c->flood = NULL;
	}

	if (!c->flood)
	{
	    *end++ = 'f';
	    newflood = ci->mlock_flood;
	    c->flood = sstrdup (newflood);
	    set_flood = 1;
	}
    }

    if (ci->mlock_link)
    {
	if (c->link && strcmp (c->link, ci->mlock_link))
	{
	    send_cmode (s_ChanServ, c->name, "-L", c->link);
	    free (c->link);
	    c->link = NULL;
	}

	if (!c->link)
	{
	    *end++ = 'L';
	    newlink = ci->mlock_link;
	    c->link = sstrdup (newlink);
	    set_link = 1;
	}
    }
#endif

    if (end[-1] == '+')
	end--;

    modes = c->mode & ci->mlock_off;
    modes &= ~(CMODE_k | CMODE_l);
#ifdef UNREAL3
    modes &= ~(CMODE_L | CMODE_f);
#endif
    end += snprintf (end, sizeof (newmodes) - (end - newmodes) - 1,
	"-%s", flags_to_string (modes));
    c->mode &= ~modes;

    if (c->limit && (ci->mlock_off & CMODE_l))
    {
	*end++ = 'l';
	c->limit = 0;
    }

    if (c->key && (ci->mlock_off & CMODE_k))
    {
	*end++ = 'k';
	newkey = sstrdup (c->key);
	free (c->key);
	c->key = NULL;
	set_key = 1;
    }

#ifdef UNREAL3
    if (c->flood && (ci->mlock_off & CMODE_f))
    {
	*end++ = 'f';
	newflood = sstrdup (c->flood);
	free (c->flood);
	c->flood = NULL;
	set_flood = 1;
    }

    if (c->link && (ci->mlock_off & CMODE_L))
    {
	*end++ = 'L';
	newlink = sstrdup (c->link);
	free (c->link);
	c->link = NULL;
	set_link = 1;
    }
#endif

    if (end[-1] == '-')
	end--;

    if (end == newmodes)
	return;

    *end = 0;
    *params = 0;

    if (set_limit)
	strcat (params, itoa (newlimit));

    if (set_key)
    {
	if (*params)
	    strcat (params, " ");

	strcat (params, newkey);
    }

#ifdef UNREAL3
    if (set_flood)
    {
	if (*params)
	    strcat (params, " ");

	strcat (params, newflood);
    }

    if (set_link)
    {
	if (*params)
	    strcat (params, " ");

	strcat (params, newlink);
    }
#endif

    send_cmode (s_ChanServ, c->name, newmodes, params ? params : "");

    if (newkey && !c->key)
	free (newkey);

#ifdef UNREAL3
    if (newflood && !c->flood)
	free (newflood);

    if (newlink && !c->link)
	free (newlink);
#endif
}

/* Change the topic of a registered channel. Store the new one in ChanInfo. */
void set_topic (ChanInfo *ci, const char *topic, const char *setter)
{
    Channel *c = findchan (ci->name);
    time_t now = time (NULL);

    if (!c)
	return;

    /* Update the ChanInfo */
    strscpy (ci->topic, topic, TOPICLEN);
    strscpy (ci->topicsetter, setter, NICKLEN);
    ci->topictime = now;

    /* Update the channel's topic info */
    strscpy (c->topic, topic, TOPICLEN);
    strscpy (c->topicsetter, setter, NICKLEN);
    c->topictime = now;

    /* Now change the topic. */
    send_cmd (s_ChanServ, "TOPIC %s %s %lu :%s", ci->name, setter, now, topic);
}

/* Return the user's access on the given channel. */
int get_access (User *u, ChanInfo *ci)
{
    NickInfo *ni;
    ChanAccess *ca;
    int i;

    if (ci->flags & CF_FROZEN)
	return 0;

    ni = get_nick (u);

    if (!ni)
	return 0;

    for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
    {
	if (!ca->level)
	    continue;

	/* We may have the linked nick on the access list, in which
	   case it won't match the host nick. Check for that.
	 */
	if (islinked (ca->mask, ni))
	    return ca->level;

	/* It doesn't appear to be linked, does it match? */
	if (!stricmp (ca->mask, ni->nick))
	    return ca->level;
    }

    return 0;
}

/* Do various ChanServ specific on join things. */
#if defined (BAHAMUT) || defined (PROMETHEUS) || defined (UNREAL3)
int cs_join (User *u, const char *c, int fromburst)
#else
int cs_join (User *u, const char *c)
#endif
{
    ChanInfo *ci = cs_findchan (c);
    NickInfo *ni = get_nick (u);
    Channel *ch;
    AKick *ak;
#if defined (BAHAMUT) || defined (PROMETHEUS) || defined (UNREAL3)
    char *av[3];
#endif
    char mask[128];
    int i, akicked = 0;
#ifdef UNREAL3
    char **chans;
#endif

    if (!ci || (ci && (ci->flags & CF_WAITAUTH)))
	return 0;

    /* Get their entire address. We'll match this against any AKicks. */
    snprintf (mask, sizeof (mask), "%s@%s", u->user, u->host);

    /* Check if the channel is froze. Kick and ban if true. */
    if (ci->flags & CF_FROZEN && !is_oper (u->nick))
    {
	/* Join it to keep the ban from being removed */
	if (!(ch = findchan (c)))
	{
	    send_cmd (s_ChanServ, "JOIN %s", ci->name);
	    add_cs_timeout (ci, TO_UNINHABIT, 60);
	}

#if defined (BAHAMUT) || defined (PROMETHEUS) || defined (UNREAL3)
	/* If we were bursting, they've already been added to memory. We need to
	   remove them. c should be the channel, from the above if..
	 */
	if (fromburst && ch)
	{
	    av[0] = ci->name;
	    av[1] = u->nick;
	    av[2] = "";

	    do_kick (s_ChanServ, 3, av);

	    /* They're out of the channel. NOW check if the channel is empty, and
	       inhabit if needed.
	     */
	    if (!findchan (c))
	    {
		send_cmd (s_ChanServ, "JOIN %s", ci->name);
		add_cs_timeout (ci, TO_UNINHABIT, 60);
	    }
	}
#endif

	/* The user isn't yet in memory for this channel, so we can just ban and kick.
	   Since it's a frozen channel, we'll ban everyone.
	 */
	send_cmd (s_ChanServ, "MODE %s +b *!*@*.*", ci->name);
	send_cmd (s_ChanServ, "KICK %s %s :This channel is frozen: %s", ci->name,
	    u->nick, ci->topic);

	return 1;
    }

    /* Check for existing AKicks against this user. */
    for (ak = ci->akick, i = 0; i < ci->akickcnt; ak++, i++)
    {
	if (!ak->level)
	    continue;

	/* Nick AKick */
	if (ak->level == 2)
	{
	    /* Check if they've identified to the AKicked nick */
	    if (ni && !stricmp (ni->nick, ak->mask))
		akicked = 1;

	    /* If not, check if they're just using it. */
	    if (!akicked && !stricmp (u->nick, ak->mask))
		akicked = 1;
	}

	/* Mask AKick */
	if (ak->level == 1)
	    if (match_wild_nocase (mask, ak->mask))
		akicked = 1;

	if (akicked)
	{
	    /* Join it to keep the ban from being removed */
	    if (!findchan (c))
	    {
		send_cmd (s_ChanServ, "JOIN %s", ci->name);
		add_cs_timeout (ci, TO_UNINHABIT, 60);
	    }

#if defined (BAHAMUT) || defined (PROMETHEUS) || defined (UNREAL3)
	    /* If we were bursting, they've already been added to memory. We need to
	       remove them. c should be the channel, from the above if..
	     */
	    if (fromburst && c)
	    {
		av[0] = ci->name;
		av[1] = u->nick;
		av[2] = "";

		do_kick (s_ChanServ, 3, av);

		/* They're out of the channel. NOW check if the channel is empty, and
		   inhabit if needed.
		 */
		if (!findchan (c))
		{
		    send_cmd (s_ChanServ, "JOIN %s", ci->name);
		    add_cs_timeout (ci, TO_UNINHABIT, 60);
		}
	    }
#endif

	    /* The user isn't yet in memory for this channel, so we can just
	       ban and kick.
	     */
	    send_cmd (s_ChanServ, "MODE %s +b %s", ci->name, ak->mask);
	    send_cmd (s_ChanServ, "KICK %s %s :%s", ci->name, u->nick,
		ak->reason ? ak->reason : def_akick_reason);

	    return 1;
	}
    }

    /* Check if the channel is restricted */
    if ((ci->flags & CF_RESTRICTED) && !is_oper (u->nick))
    {
	if (!get_access (u, ci))
	{
	    /* Is anyone else in it? We need to inhabit it otherwise. */
	    if (!findchan (c))
	    {
		send_cmd (s_ChanServ, "JOIN %s", ci->name);
		add_cs_timeout (ci, TO_UNINHABIT, 60);
	    }
    
	    send_cmd (s_ChanServ, "MODE %s +b %s", ci->name, create_mask (u));
	    send_cmd (s_ChanServ, "KICK %s %s :%s", ci->name, u->nick,
		CS_RESTRICTED_CHAN);

	    return 1;
	}
    }

#if defined (BAHAMUT) || defined (PROMETHEUS) || defined (UNREAL3)
    if (!fromburst)
    {
#endif
	/* Check for a channel greet message. */
	if (ci->greet)
	    notice (s_ChanServ, u->nick, CS_GREET, ci->name, ci->greet);

	/* Check for a channel URL. */
	if (ci->url)
	    send_cmd (me.name, "328 %s %s :%s", u->nick, ci->name, ci->url);
#if defined (BAHAMUT) || defined (PROMETHEUS) || defined (UNREAL3)
    }
#endif

#ifdef UNREAL3
    /* Check if they're identified for the channel. Set +q if they are. */
    if (u->idchancnt)
	for (chans = u->idchans, i = 0; i < u->idchancnt; chans++, i++)
	    if (!stricmp (*chans, ci->name))
		send_cmd (s_ChanServ, "MODE %s +q %s", ci->name, u->nick);
#endif

    return 0;
}
