/* Config file routines

   Cygnus IRC Services - Copyright (c) 2001-2002 Darcy Grexton
   Contact: skold@habber.net, skold @ HabberNet, ICQ #6285178

   This program is free but copyrighted software; see the file COPYING
   for details.
 */

#include "../inc/services.h"

SRA *sralist = NULL;

/* Some vars */
char s_RootServ[NICKLEN], s_NickServ[NICKLEN], s_MemoServ[NICKLEN];
char s_ChanServ[NICKLEN], s_GlobalNoticer[NICKLEN];
char tlds[64], snoopchan[32], def_akill_reason[256], joinchan[32];
char sendmail_path[512], network_name[32], nsregister_url[256];
char memomail_url[256], csregister_url[256], collision_url[256];
char def_topic[350], def_modes[32], def_akick_reason[256], globalmsg[64];
Boolean_T rootserv_on = FALSE, nickserv_on = FALSE, memoserv_on = FALSE;
Boolean_T chanserv_on = FALSE, globalnoticer_on = FALSE, fakehost_on = FALSE;
Boolean_T globalonakill_on = FALSE, logupdates_on = FALSE, age_on = FALSE;
Boolean_T learnakills_on = FALSE, forgetakills_on = FALSE, sex_on = FALSE;
Boolean_T name_on = FALSE, location_on = FALSE, privmsg_on = FALSE;
Boolean_T showmail_on = FALSE, sendmail_on = FALSE, email_on = FALSE;
Boolean_T url_on = FALSE, uin_on = FALSE, memomail_on = FALSE;
Boolean_T securesra_on = FALSE, globalonauth_on = FALSE, haveserv_on = FALSE;
Boolean_T statglobal_on = FALSE, noisyflood_on = FALSE;
int update_timeout, nonwildreq, def_akill_time, unreal_numeric, passfails, chanlimit;
int wait_restart, maxclones, maxclonekills, nick_expire, maxlinks, memo_delay;
int nick_delay, chan_delay, memobox_size, memomailmax, memosfromuser, csregistertype;
int securitysetting, emaillimit, nsregistertype, tempexpire, chan_expire, collidetype;
int floodmsgs, akickmax, helpindex;
time_t floodtime;
long defnickflags, defchanflags, defmlock_on, defmlock_off;

/* Add a new SRA into the sra list. */ 
static SRA *new_sra (const char *nick, const char *pass)
{
    SRA *sra;

    sra = scalloc (sizeof (SRA), 1);

    sra->nick = sstrdup (nick);
    sra->pass = sstrdup (pass);

    sra->next = sralist;

    if (sra->next)
        sra->next->prev = sra;

    sralist = sra;

    return sra;
}

/* Load and parse the conf file */
void load_conf (void)
{
    FILE *f = fopen (CONFIGFILE, "r");
    char *item, *s, dBuf[BUFSIZE];
#ifdef HAVE_GETTIMEOFDAY
    struct timeval start, now, tmp;
#endif

#ifdef HAVE_GETTIMEOFDAY
    log_sameline (1, "CONF: Loading.... ");
#else
    log ("CONF: Loading....");
#endif

    if (!f)
    {
	log (ERR_NO_CONFIG);
	exit (1);
    }

#ifdef HAVE_GETTIMEOFDAY
    gettimeofday (&(start), NULL);
#endif

    while (fgets (dBuf, 2047, f))
    {
	item = strtok (dBuf, " ");
	strip (item);

	if (*item == '#' || *item == '\t' || *item == ' ')
	    continue;

	if (!stricmp (item, "Info"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s && strchr (s, '.'))
		memcpy (me.name, s, sizeof (me.name));
	    else
		fatal ("Config: Info is not set properly: No server name.");

	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		memcpy (me.desc, s, sizeof (me.desc));
	    else
	    {
		char tmp[25];

		sprintf (tmp, "Cygnus IRC Services %s", version);
		memcpy (me.desc, tmp, sizeof (me.desc));
	    }

	    continue;
	}

	if (!stricmp (item, "BindTo"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (me.bindhost, s, sizeof (me.bindhost));
	    else
		fatal ("Config: BindTo is not set properly: No host.");

	    s = strtok (NULL, " ");   
	    strip (s);

	    if (s)
		me.bindport = atoi (s);
	    else  
		fatal ("Config: BindTo is not set properly: No port.");

	    continue;
	}

	if (!stricmp (item, "Connect"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (me.uplink, s, sizeof (me.uplink));
	    else
		fatal ("Config: Connect is not set properly: No uplink.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (me.pass, s, sizeof (me.pass));
	    else
		fatal ("Config: Connect is not set properly: No password.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (me.ruplink, s, sizeof (me.ruplink));
	    else
		fatal ("Config: Connect is not set properly: No uplink name.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		me.port = atoi (s);
	    else
		fatal ("Config: Connect is not set properly: No port.");

	    continue;
	}

	if (!stricmp (item, "SRAService"))
	{
	    /* In case the client info changes */
	    char nick[NICKLEN], user[USERLEN], host[HOSTLEN];
	    char real[REALLEN], mode[12];

	    rootserv_on = TRUE;

	    strscpy (nick, s_RootServ, sizeof (nick));
	    strscpy (user, rs.user, sizeof (user));
	    strscpy (host, rs.host, sizeof (host));
	    strscpy (real, rs.real, sizeof (real));
	    strscpy (mode, rs.mode, sizeof (mode));

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (s_RootServ, s, sizeof (s_RootServ));
	    else
		fatal ("Config: SRAService is not set properly: No nickname.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (rs.user, s, sizeof (rs.user));
	    else
		fatal ("Config: SRAService is not set properly: No username.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (rs.host, s, sizeof (rs.host));
	    else
		fatal ("Config: SRAService is not set properly: No hostname.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
	    {
		if (!strchr (s, '+'))
		    fatal ("Config: SRAService: You need a + for modes!");

		memcpy (rs.mode, s, sizeof (rs.mode));
	    }
	    else
		fatal ("Config: SRAService is not set properly: No modes.");

	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		memcpy (rs.real, s, sizeof (rs.real));
	    else
		fatal ("Config: SRAService is not set properly: No realname.");

	    if (!(runflags & RUN_STARTING))
	    {
		/* Check if the client info changed. If so, reintroduce the
		   client with the new info.
		 */
		if (rs.isonline && (strcmp (s_RootServ, nick) ||
		    strcmp (rs.user, user) || strcmp (rs.host, host) ||
		    strcmp (rs.mode, mode) || strcmp (rs.real, real)))
		{
		    send_cmd (nick, "%s :Client info changed",
			me.token ? "," : "QUIT");
		    intro_user (s_RootServ, rs.user, rs.host, rs.real, rs.mode);
		}

		/* If RootServ was previously off, and is now on, introduce
		   the pseudo client if needed.
		 */
		if (!rs.isonline)
		{
		    intro_user (s_RootServ, rs.user, rs.host, rs.real, rs.mode);
		    userstat++;
		    rs.isonline = 1;
		}
	    }

	    continue;
	}

	if (!stricmp (item, "NickNameService"))
	{
	    /* In case the client info changes */
	    char nick[NICKLEN], user[USERLEN], host[HOSTLEN];
	    char real[REALLEN], mode[12];

	    nickserv_on = TRUE;

	    strscpy (nick, s_NickServ, sizeof (nick));
	    strscpy (user, ns.user, sizeof (user));
	    strscpy (host, ns.host, sizeof (host));
	    strscpy (real, ns.real, sizeof (real));
	    strscpy (mode, ns.mode, sizeof (mode));

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (s_NickServ, s, sizeof (s_NickServ));
	    else
		fatal ("Config: NickNameService is not set properly: No nickname.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (ns.user, s, sizeof (ns.user));
	    else
		fatal ("Config: NickNameService is not set properly: No username.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (ns.host, s, sizeof (ns.host));
	    else
		fatal ("Config: NickNameService is not set properly: No hostname.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
	    {
		if (!strchr (s, '+'))
		    fatal ("Config: NickNameService: You need a + for modes!");

		memcpy (ns.mode, s, sizeof (ns.mode));
	    }
	    else
		fatal ("Config: NickNameService is not set properly: No modes.");

	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		memcpy (ns.real, s, sizeof (ns.real));
	    else
		fatal ("Config: NickNameService is not set properly: No realname.");

	    if (!(runflags & RUN_STARTING))
	    {
		/* Check if the client info changed. If so, reintroduce the  
		   client with the new info.
		 */
		if (ns.isonline && (strcmp (s_NickServ, nick) ||
		    strcmp (ns.user, user) || strcmp (ns.host, host) ||
		    strcmp (ns.mode, mode) || strcmp (ns.real, real)))
		{
		    send_cmd (nick, "%s :Client info changed",
			me.token ? "," : "QUIT");
		    intro_user (s_NickServ, ns.user, ns.host, ns.real, ns.mode);
		}

		/* If NickServ was previously off, and is now on, introduce
		   the pseudo client if needed.
		 */
		if (!ns.isonline)
		{
		    intro_user (s_NickServ, ns.user, ns.host, ns.real, ns.mode);
		    userstat++;
		    ns.isonline = 1;
		}
	    }

	    continue;
	}

	if (!stricmp (item, "ChannelService"))
	{
	    /* In case the client info changes */
	    char nick[NICKLEN], user[USERLEN], host[HOSTLEN];
	    char real[REALLEN], mode[12];

	    chanserv_on = TRUE;

	    strscpy (nick, s_ChanServ, sizeof (nick));
	    strscpy (user, cs.user, sizeof (user));
	    strscpy (host, cs.host, sizeof (host));
	    strscpy (real, cs.real, sizeof (real));
	    strscpy (mode, cs.mode, sizeof (mode));

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (s_ChanServ, s, sizeof (s_ChanServ));
	    else
		fatal ("Config: ChannelService is not set properly: No nickname.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (cs.user, s, sizeof (cs.user));
	    else
		fatal ("Config: ChannelService is not set properly: No username.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (cs.host, s, sizeof (cs.host));
	    else
		fatal ("Config: ChannelService is not set properly: No hostname.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
	    {
		if (!strchr (s, '+'))
		    fatal ("Config: ChannelService: You need a + for modes!");

		memcpy (cs.mode, s, sizeof (cs.mode));
	    }
	    else
		fatal ("Config: ChannelService is not set properly: No modes.");

	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		memcpy (cs.real, s, sizeof (cs.real));
	    else
		fatal ("Config: ChannelService is not set properly: No realname.");

	    if (!(runflags & RUN_STARTING))
	    {
		/* Check if the client info changed. If so, reintroduce the  
		   client with the new info.
		 */
		if (cs.isonline && (strcmp (s_ChanServ, nick) ||
		    strcmp (cs.user, user) || strcmp (cs.host, host) ||
		    strcmp (cs.mode, mode) || strcmp (cs.real, real)))
		{
		    send_cmd (nick, "%s :Client info changed",
			me.token ? "," : "QUIT");
		    intro_user (s_ChanServ, cs.user, cs.host, cs.real, cs.mode);
		}

		/* If ChanServ was previously off, and is now on, introduce
		   the pseudo client if needed.
		 */
		if (!cs.isonline)
		{
		    intro_user (s_ChanServ, cs.user, cs.host, cs.real, cs.mode);
		    userstat++;
		    cs.isonline = 1;
		}
	    }

	    continue;
	}

	if (!stricmp (item, "MemoService"))
	{
	    /* In case the client info changes */
	    char nick[NICKLEN], user[USERLEN], host[HOSTLEN];
	    char real[REALLEN], mode[12];

	    memoserv_on = TRUE;

	    strscpy (nick, s_MemoServ, sizeof (nick));
	    strscpy (user, ms.user, sizeof (user));
	    strscpy (host, ms.host, sizeof (host));
	    strscpy (real, ms.real, sizeof (real));
	    strscpy (mode, ms.mode, sizeof (mode));

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (s_MemoServ, s, sizeof (s_MemoServ));
	    else
		fatal ("Config: MemoService is not set properly: No nickname.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (ms.user, s, sizeof (ms.user));
	    else
		fatal ("Config: MemoService is not set properly: No username.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (ms.host, s, sizeof (ms.host));
	    else
		fatal ("Config: MemoService is not set properly: No hostname.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
	    {
		if (!strchr (s, '+'))
		    fatal ("Config: MemoService: You need a + for modes!");

		memcpy (ms.mode, s, sizeof (ms.mode));
	    }
	    else
		fatal ("Config: MemoService is not set properly: No modes.");

	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		memcpy (ms.real, s, sizeof (ms.real));
	    else
		fatal ("Config: MemoService is not set properly: No realname.");

	    if (!(runflags & RUN_STARTING))
	    {
		/* Check if the client info changed. If so, reintroduce the  
		   client with the new info.
		 */
		if (ms.isonline && (strcmp (s_MemoServ, nick) ||
		    strcmp (ms.user, user) || strcmp (ms.host, host) ||
		    strcmp (ms.mode, mode) || strcmp (ms.real, real)))
		{
		    send_cmd (nick, "%s :Client info changed",
			me.token ? "," : "QUIT");
		    intro_user (s_MemoServ, ms.user, ms.host, ms.real, ms.mode);
		}

		/* If MemoServ was previously off, and is now on, introduce
		   the pseudo client if needed.
		 */
		if (!ms.isonline)
		{
		    intro_user (s_MemoServ, ms.user, ms.host, ms.real, ms.mode);
		    userstat++;
		    ms.isonline = 1;
		}
	    }

	    continue;
	}

	if (!stricmp (item, "GlobalNoticer"))
	{
	    /* In case the client info changes */
	    char nick[NICKLEN], user[USERLEN], host[HOSTLEN];
	    char real[REALLEN], mode[12];

	    globalnoticer_on = TRUE;

	    strscpy (nick, s_GlobalNoticer, sizeof (nick));
	    strscpy (user, gn.user, sizeof (user));
	    strscpy (host, gn.host, sizeof (host));
	    strscpy (real, gn.real, sizeof (real));
	    strscpy (mode, gn.mode, sizeof (mode));

	    globalnoticer_on = TRUE;

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (s_GlobalNoticer, s, sizeof (s_GlobalNoticer));
	    else
		fatal ("Config: GlobalNoticer is not set properly: No nickname.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (gn.user, s, sizeof (gn.user));
	    else
		fatal ("Config: GlobalNoticer is not set properly: No username.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (gn.host, s, sizeof (gn.host));
	    else
		fatal ("Config: GlobalNoticer is not set properly: No hostname.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
	    {
		if (!strchr (s, '+'))
		    fatal ("Config: GlobalNoticer: You need a + for modes!"); 

		memcpy (gn.mode, s, sizeof (gn.mode));
	    }
	    else
		fatal ("Config: GlobalNoticer is not set properly: No modes.");

	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		memcpy (gn.real, s, sizeof (gn.real));
	    else
		fatal ("Config: GlobalNoticer is not set properly: No realname.");

	    if (!(runflags & RUN_STARTING))
	    {
		/* Check if the client info changed. If so, reintroduce the
		   client with the new info.
		 */
		if (gn.isonline && (strcmp (s_GlobalNoticer, nick) ||
		    strcmp (gn.user, user) || strcmp (gn.host, host) ||
		    strcmp (gn.mode, mode) || strcmp (gn.real, real)))
		{
		    send_cmd (nick, "%s :Client info changed",
			me.token ? "," : "QUIT");
		    intro_user (s_GlobalNoticer, gn.user, gn.host, gn.real,
			gn.mode);
		}

		/* If the GlobalNoticer was previously off, and is now on,
		   introduce the pseudo client if needed.
		 */
		if (!gn.isonline)
		{
		    intro_user (s_GlobalNoticer, gn.user, gn.host, gn.real,
			gn.mode);
		    userstat++;
		    gn.isonline = 1;
		}
	    }

	    continue;
	}

	if (!stricmp (item, "SRA"))
	{
	    char *nick, *pass;

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		nick = s;
	    else
		nick = "";

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		pass = s;
	    else
		pass = "";

	    if (!pass)
		fatal ("Config: Missing password for SRA %s", nick);
	    else
		new_sra (nick, pass);

	    continue;
	}

	if (!stricmp (item, "Trigger"))
	{
	    char *host;
	    int limit;

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		host = s;
	    else
		host = "";

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		limit = atoi (s);
	    else
		limit = 0;

	    if (!limit)
		fatal ("Config: Missing limit for Trigger %s", host);
	    else
		new_trigger (host, limit);

	    continue;
	}

	if (!stricmp (item, "Exception"))
	{
	    char *host;

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		host = s;
	    else
		host = "";

	    new_exception (host);

	    continue;
	}

	if (!stricmp (item, "SyncTime"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		update_timeout = dotime (s);  
	    else
		fatal ("Config: SyncTime is not set properly: No time value.");

	    continue;
	}

	if (!stricmp (item, "NickExpire"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		nick_expire = dotime (s);
	    else
		fatal ("Config: NickExpire is not set properly: No time value.");

	    continue;
	}

	if (!stricmp (item, "ChanExpire"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		chan_expire = dotime (s);
	    else
		fatal ("Config: ChanExpire is not set properly: No time value.");

	    continue;
	}

	if (!stricmp (item, "MaxConnections"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		maxclones = atoi (s);
	    else
		fatal ("Config: MaxConnections is not set properly: No value.");

	    continue;
	}

        if (!stricmp (item, "MaxCloneKills"))
        {
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		maxclonekills = atoi (s);
	    else
		fatal ("Config: MaxCloneKills is not set properly: No value.");

	    continue;
	}

	if (!stricmp (item, "SnoopChan"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (snoopchan, s, sizeof (snoopchan));
	    else
		fatal ("Config: SnoopChan is not set properly: No channel.");

	    continue;
	}

	if (!stricmp (item, "NSRegisterURL"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (nsregister_url, s, sizeof (nsregister_url));
	    else
		fatal ("Config: NSRegisterURL is not set properly: No URL.");

	    continue;
	}

	if (!stricmp (item, "CSRegisterURL"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (csregister_url, s, sizeof (csregister_url));
	    else
		fatal ("Config: CSRegisterURL is not set properly: No URL.");

	    continue;
	}

	if (!stricmp (item, "MemoMailURL"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (memomail_url, s, sizeof (memomail_url));
	    else
		fatal ("Config: MemoMailURL is not set properly: No URL.");

	    continue;
	}

	if (!stricmp (item, "DefaultTopic"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		memcpy (def_topic, s, sizeof (def_topic));
	    else
		fatal ("Config: DefaultTopic is not set properly: No topic.");

	    continue;
	}

	if (!stricmp (item, "CollisionURL"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (collision_url, s, sizeof (collision_url));
	    else
		fatal ("Config: CollisionURL is not set properly: No URL.");

	    continue;
	}

	if (!stricmp (item, "NetworkName"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (network_name, s, sizeof (network_name));
	    else
		fatal ("Config: NetworkName is not set properly: No value.");

	    continue;
	}

	if (!stricmp (item, "SendMailPath"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
	    {
		sendmail_on = TRUE;
		memcpy (sendmail_path, s, sizeof (sendmail_path));
	    }
	    else
		fatal ("Config: SendMailPath is not set properly: No value.");

	    continue;
	}

	if (!stricmp (item, "JoinChan"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memcpy (joinchan, s, sizeof (joinchan));
	    else
		fatal ("Config: JoinChan is not set properly: No channel.");

	    continue;
	}

	if (!stricmp (item, "TLD"))
	{  
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		memcpy (tlds, s, sizeof (tlds));
	    else
		fatal ("Config: TLD is not set properly: No TLDs");

	    continue;
        }

	if (!stricmp (item, "AKillExpire"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		def_akill_time = dotime (s);
	    else
		fatal ("Config: AKillExpire is not set properly: No time value.");

	    continue;
	}

	if (!stricmp (item, "TempExpire"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		tempexpire = dotime (s);
	    else
		fatal ("Config: TempExpire is not set properly: No time value.");

	    continue;
	}

	if (!stricmp (item, "NonWildCards"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		nonwildreq = atol (s);
	    else
		fatal ("Config: NonWildCards is not set properly: No value.");

	    continue;
	}

	if (!stricmp (item, "NSRegisterType"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
	    {
		if (!stricmp (s, "Normal"))
		    nsregistertype = 0;
		if (!stricmp (s, "EMailAuth"))
		    nsregistertype = 1;
		if (!stricmp (s, "EMailNoAuth"))
		    nsregistertype = 2;
		if (!stricmp (s, "PassAuth"))
		    nsregistertype = 3;
		if (!stricmp (s, "Verify"))
		    nsregistertype = 4;
		if (!stricmp (s, "EMailVerify"))
		    nsregistertype = 5;
		if (!stricmp (s, "AuthVerify"))
		    nsregistertype = 6;
	    }
	    else
		fatal ("Config: NSRegisterType is not set properly: No value.");

	    continue;
	}

	if (!stricmp (item, "CSRegisterType"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
	    {
		if (!stricmp (s, "Normal"))
		    csregistertype = 0;
		if (!stricmp (s, "EMailAuth"))
		    csregistertype = 1;
		if (!stricmp (s, "Verify"))
		    csregistertype = 2;
		if (!stricmp (s, "AuthVerify"))
		    csregistertype = 3;
		if (!stricmp (s, "CSOpVerify"))
		    csregistertype = 4;
	    }
	    else
		fatal ("Config: CSRegisterType is not set properly: No value.");

	    continue;
	}

	if (!stricmp (item, "AKickMax"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		akickmax = atoi (s);
	    else
		fatal ("Config: AKickMax is not set properly: No value.");

	    continue;
	}

	if (!stricmp (item, "EMailLimit"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		emaillimit = atoi (s);
	    else
		fatal ("Config: EMailLimit is not set properly: No value.");

	    continue;
	}

	if (!stricmp (item, "ChannelLimit"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		chanlimit = atoi (s);
	    else
		fatal ("Config: ChannelLimit is not set properly: No value.");

	    continue;
	}

	if (!stricmp (item, "SecuritySetting"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
	    {
		if (!stricmp (s, "All"))
		    securitysetting = 1;
		if (!stricmp (s, "Passwords"))
		    securitysetting = 2;
		if (!stricmp (s, "None"))
		    securitysetting = 0;
	    }
	    else
		fatal ("Config: SecuritySetting is not set properly: No value.");

	    continue;
	}

	if (!stricmp (item, "GlobalMSG"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		memcpy (globalmsg, s, sizeof (globalmsg));
	    else
		fatal ("Config: GlobalMSG is not set properly: No message.");

	    continue;
	}

#ifdef UNREAL3
	if (!stricmp (item, "UnrealNumeric"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		unreal_numeric = atol (s);
	    else
		fatal ("Config: UnrealNumeric is not set properly: No value.");

	    continue;
	}
#endif

	if (!stricmp (item, "DefaultAKillReason"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		memcpy (def_akill_reason, s, sizeof (def_akill_reason));
	    else
		fatal ("Config: DefaultAKillReason is not set properly: No reason.");

	    continue;
	}

	if (!stricmp (item, "DefaultAKickReason"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		memcpy (def_akick_reason, s, sizeof (def_akick_reason));
	    else
		fatal ("Config: DefaultAKickReason is not set properly: No reason.");

	    continue;
	}

	if (!stricmp (item, "WaitRestart"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		wait_restart = dotime (s);
	    else
		fatal ("Config: WaitRestart is not set properly: No time value.");

	    continue;
	}

	if (!stricmp (item, "PassFails"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		passfails = atol (s);
	    else
		fatal ("Config: PassFails is not set properly.");

	    continue;
	}

	if (!stricmp (item, "MaxLinks"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		maxlinks = atol (s);
	    else
		fatal ("Config: MaxLinks is not set properly.");

	    continue;
	}

	if (!stricmp (item, "MemoMailMax"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		memomailmax = atol (s);
	    else
		fatal ("Config: MemoMailMax is not set properly.");

	    continue;
	}

	if (!stricmp (item, "MemosFromUser"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		memosfromuser = atol (s);
	    else
		fatal ("Config: MemosFromUser is not set properly.");

	    continue;
	}

	if (!stricmp (item, "MemoBoxSize"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (s)
		memobox_size = atol (s);
	    else
		fatal ("Config: MemoBoxSize is not set properly.");

	    continue;
	}

	if (!stricmp (item, "Delays"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		nick_delay = dotime (s);
	    else
		fatal ("Config: Delays is not set properly: No Nick Reg Delay.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		chan_delay = dotime (s);
	    else
		fatal ("Config: Delays is not set properly: No Chan Reg Delay.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		memo_delay = dotime (s);
	    else
		fatal ("Config: Delays is not set properly: No Memo Send Delay.");

	    continue;
	}

	if (!stricmp (item, "NickFlags"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (stristr (s, "Enforce"))
		defnickflags |= NF_ENFORCE;

	    if (stristr (s, "Secure"))
		defnickflags |= NF_SECURE;

	    if (stristr (s, "MailVisibility"))
	    {
		if (showmail_on == TRUE)
		    defnickflags |= NF_SHOWMAIL;
		else
		    defnickflags |= NF_HIDEMAIL;
	    }

	    if (stristr (s, "NeverOp"))
		defnickflags |= NF_NEVEROP;

	    if (stristr (s, "NoOp"))
		defnickflags |= NF_NOOP;

	    if (stristr (s, "NoSuccessor"))
		defnickflags |= NF_NOSUCCESSOR;

	    if (stristr (s, "Private"))
		defnickflags |= NF_PRIVATE;

	    if (stristr (s, "NoMemo"))
		defnickflags |= NF_NOMEMO;

	    if (stristr (s, "Receipts"))
		defnickflags |= NF_RECEIPTS;

	    /* We'll turn this off when we run checkconf if we need to */
	    if (stristr (s, "MemoMail"))
		defnickflags |= NF_MEMOMAIL;

	    continue;
	}

	if (!stricmp (item, "ChanFlags"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (stristr (s, "Verbose"))
		defchanflags |= CF_VERBOSE;

	    if (stristr (s, "VOPAll"))
		defchanflags |= CF_VOPALL;

	    if (stristr (s, "Secure"))
		defchanflags |= CF_SECURE;

	    if (stristr (s, "Restricted"))
		defchanflags |= CF_RESTRICTED;

	    continue;
	}

	if (!stricmp (item, "DefaultModes"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (strstr (s, "+") || strstr (s, "-"))
		fatal ("Config: DefMLock is not set properly. Do not use + or -.");

	    defmlock_on = string_to_flags (s);

	    s = strtok (NULL, " ");
	    strip (s);

	    if (strstr (s, "+") || strstr (s, "-"))
		fatal ("Config: DefMLock is not set properly. Do not use + or -.");

	    defmlock_off = string_to_flags (s);

	    continue;
	}

	if (!stricmp (item, "CollideType"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (stristr (s, "Guest"))
		collidetype = 1;

	    if (stristr (s, "User"))
		collidetype = 2;

	    if (stristr (s, "Nick"))
		collidetype = 3;

	    if (stristr (s, "Underscore"))
		collidetype = 4;

	    if (stristr (s, "Hyphen"))
		collidetype = 5;

	    if (stristr (s, "Kill"))
		collidetype = 6;

	    continue;
	}

	if (!stricmp (item, "HelpIndex"))
	{
	    s = strtok (NULL, "");
	    strip (s);

	    if (stristr (s, "Short"))
		helpindex = 1;

	    if (stristr (s, "Medium"))
		helpindex = 2;

	    if (stristr (s, "Long"))
		helpindex = 3;

	    continue;
	}

	if (!stricmp (item, "FloodTrigger"))
	{
	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		floodmsgs = atoi (s);
	    else
		fatal ("Config: FloodTrigger is not set properly: No message amount.");

	    s = strtok (NULL, " ");
	    strip (s);

	    if (s)
		floodtime = dotime (s);
	    else
		fatal ("Config: FloodTrigger is not set properly: No message time.");

	    continue;
	}

	if (!stricmp (item, "GlobalOnAKill"))
	{
	    globalonakill_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "GlobalOnAuth"))
	{
	    globalonauth_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "LearnAKills"))
	{
	    learnakills_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "ForgetAKills"))
	{
	    forgetakills_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "LogUpdates"))
	{
	    logupdates_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "SecureSRA"))
	{
	    securesra_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "HaveServ"))
	{
	    haveserv_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "UseEMail"))
	{
	    email_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "UseURL"))
	{
	    url_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "UseUIN"))
	{
	    uin_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "UseMemoMail"))
	{
	    memomail_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "UsePrivMSG"))
	{
	    privmsg_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "UseShowMail"))
	{
	    showmail_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "UseName"))
	{
	    name_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "UseAge"))
	{
	    age_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "UseSex"))
	{
	    sex_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "UseLocation"))
	{
	    location_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "StatusGlobals"))
	{
	    statglobal_on = TRUE;
	    continue;
	}

	if (!stricmp (item, "NoisyFlood"))
	{
	    noisyflood_on = TRUE;
	    continue;
	}

#ifdef UNREAL3
	if (!stricmp (item, "UseFakeHost"))
	{
	    fakehost_on = TRUE;
	    continue;
	}
#endif

    }

    fclose (f);

#ifdef HAVE_GETTIMEOFDAY
    gettimeofday (&now, NULL);
    timersub (&now, &start, &tmp);
    log_sameline (0, "finished in %d ms", tv2ms (&tmp));
#endif
}

/* Sanity checks on the config options */
int check_conf (void)
{
    SRA *sra;
    int i = 0;

    for (sra = sralist; sra; sra = sra->next)
	i++;

    if (!i)
	fatal ("Config: You must have at least one SRA.");

    if (rootserv_on == FALSE)
	fatal ("Config: SRAService cannot be disabled.");

    if (memoserv_on == TRUE && nickserv_on == FALSE)
    {
	printf ("Warning: MemoServ requires NickServ. MemoServ disabled.\n");
	memoserv_on = FALSE;
    }

    if (memomail_on == TRUE && sendmail_on == FALSE)
    {
	printf ("Warning: MemoMail requires SendMailPath. MemoMail disabled.\n");

	/* Turn memomail off */
	memomail_on = FALSE;

	/* Take it out of defnickflags if we need to */
	defnickflags &= ~NF_MEMOMAIL;
    }

    if (memomail_on == TRUE && email_on == FALSE)
    {
	printf ("Warning: MemoMail requires UseEMail. MemoMail disabled.\n");

	/* Turn memomail off */
	memomail_on = FALSE;

	/* Take it out of defnickflags if we need to */
	defnickflags &= ~NF_MEMOMAIL;
    }

    if (sendmail_on == TRUE && email_on == FALSE)
    {
	printf ("Warning: SendMailPath can't do anything without UseEMail. SendMailPath disabled.\n");
	sendmail_on = FALSE;
	*sendmail_path = 0;
    }

    if (memomail_on == TRUE && !memomailmax)
    {
	printf ("Warning: MemoMailMax is required for MemoMail. MemoMail disabled.\n");

	/* Turn memomail off */
	memomail_on = FALSE;

	/* Take it out of defnickflags if we need to */
	defnickflags &= ~NF_MEMOMAIL;
    }

    if (sendmail_on == FALSE && (nsregistertype == 1 || nsregistertype == 3 ||
				 nsregistertype == 6 || csregistertype == 1 ||
				 csregistertype == 3))
	fatal ("Config: E-Mail Auth Registration requires SendMail.");

    if (!collidetype)
	collidetype = 1;

    if (!helpindex)
	helpindex = 2;

    if (!memobox_size)
	memobox_size = 25;

    if (!akickmax)
	akickmax = 50;

    if (!update_timeout)
	update_timeout = 1200;

    if (!def_akill_time)
	def_akill_time = 604800; /* One week */

    if (!nonwildreq)
	nonwildreq = 5;

    if (update_timeout < 60)
	printf ("Warning: Setting SyncTime less than 1 minute is unwise!\n");

#ifndef IRCUP9
    defmlock_on |= CMODE_r;
#endif

    if (wait_restart < 5)
	fatal ("Config: WaitRestart cannot be less than 5s.");

    if (!strlen (tlds) && (globalnoticer_on == TRUE))
    {
        printf ("Config: TLD is not set properly. GlobalNoticer disabled.\n");
	globalnoticer_on = FALSE;
    }

    if (globalnoticer_on == FALSE)
	statglobal_on = FALSE;

    if (!strlen (network_name))
	fatal ("Config: NetworkName must be set.");

    if (!strlen (def_akill_reason))
	memcpy (def_akill_reason, "You are banned.", sizeof (def_akill_reason));

    if (!strlen (def_akick_reason))
	memcpy (def_akick_reason, "User has been banned from channel", sizeof (def_akick_reason));

#ifdef UNREAL3
    if (!unreal_numeric)
	fatal ("Config: UnrealNumeric is required for UnrealIRCd.");

    if (unreal_numeric < 1 || unreal_numeric > 255)
	fatal ("Config: UnrealNumeric is not set properly: Must be between 1 and 255.");
#endif

    if (!nick_expire)
	fatal ("Config: NickExpire is not set properly.");

    if (!chan_expire)
	fatal ("Config: ChanExpire is not set properly.");

    return 1;
}
