/* Main source file.

   Cygnus IRC Services - Copyright (c) 2001-2002 Darcy Grexton
   Contact: skold@habber.net, skold @ HabberNet, ICQ #6285178

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program (see the file COPYING); if not, write to the
   Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "../inc/services.h"
#include "../inc/version.h"

Verify *verifylist[1024];
EMail *emaillist[1024];

int servsock, servcnt = 0, usercnt = 0, chancnt = 0, debuglevel = 0, userstat = 0;
#if defined (BAHAMUT) || defined (PROMETHEUS) || defined (UNREAL3)
time_t burststart;
int bursting;
#endif
#ifdef BAHAMUT
int bursted = 1;
#endif
volatile time_t last_update, news_time;
volatile uint32 last_check;
char inbuf[BUFSIZE * 8], quitmsg[BUFSIZE];
long runflags = 0;
static int rehashing = 0;
static jmp_buf panic_jmp;

/* If we get a weird signal, come here. */
static void sighandler (int signum)
{
    /* Rehash */
    if (signum == SIGHUP)
    {
	rehashing = 1;

	log ("Got SIGHUP, reloading %s", CONFIGFILE);

	globops (s_RootServ, "Got SIGHUP, reloading %s", CONFIGFILE);
	do_rehash ("system console");
    }
    else if (signum == SIGTERM)
    {
	signal (SIGTERM, SIG_IGN);
	return;
    }

    if (signum == SIGUSR1)
    {
	snprintf (quitmsg, BUFSIZE, "Out of memory!");
	runflags |= RUN_RESTART;
    }
    else
    {
	if (rehashing == 1)
	{
	    rehashing = 0;
	    return;
	}

	snoop (s_RootServ,
	    "Fatal Error: Services terminating on signal %d", signum);
	snprintf (quitmsg, BUFSIZE, "Services terminating on signal %d.",
	    signum);
	runflags |= RUN_RESTART;
    }

    log (ERR_QUIT, quitmsg);
    exit (1);
}

/* Command-line help message */
void do_help (char *prog)
{
printf("
Cygnus IRC Services - Copyright (c) 2001-2002 Darcy Grexton

Syntax: %s [parms] [options]

-help                This message
-debug [0|1|2|3]     Enable debug mode - log to debug.log
                         Level 0: No debug logging (Off)
                         Level 1: Server<->server traffic logged
                         Level 2: (Some) Function calls logged
                         Level 3: Server<->server traffic and (some) function
				  calls logged
-live                Don't fork (log screen + log file).

", prog);
}

/* Count our servers and clients so the counts we show are accurate. */
int servcount (void)
{
    int tmp = 0;

    /* We have seperate counts for statistics purposes since we need
       the counts we use internally to be accurate, not including us.
     */
    if (rootserv_on == TRUE)
	++tmp;
    if (nickserv_on == TRUE)
	++tmp;
    if (chanserv_on == TRUE)
	++tmp;
    if (memoserv_on == TRUE)
	++tmp;
    if (globalnoticer_on == TRUE)
	++tmp;

    return tmp;
}

/* Main routine .. :P */
int main (int ac, char **av, char **envp)
{
    FILE *pid_file;
    char *s;
    int i;
    struct rlimit rlim;
#ifdef HAVE_GETTIMEOFDAY
    struct timeval start, now, tmp;
#endif

    if (chdir (SERVICES_DIR) < 0)
    {
	perror (SERVICES_DIR);
	return 20;
    }

#if HAVE_UMASK
    umask (077);
#endif

    /* Core file stuff. */
    if (!getrlimit (RLIMIT_CORE, &rlim))
    {
	rlim.rlim_cur = rlim.rlim_max;
	setrlimit (RLIMIT_CORE, &rlim);
    }

#ifdef HAVE_GETTIMEOFDAY
    gettimeofday (&(start), NULL);
#endif

    /* Parse our command line options */
    for (i = 1; i < ac; ++i)
    {
	s = av[i];
	if (*s == '-' || *s == '/')
	{
	    ++s;
	    if (!stricmp (s, "-help") || !stricmp (s, "?") ||
		  !stricmp (s, "help"))
	    {
		do_help (av[0]);
		return 0;
	    }
	    else if (!stricmp (s, "debug"))
	    {
		if (++i >= ac)
		{
		    fprintf (stderr, "Error: -debug requires a parameter.\n");
		    return -1;
		}
		s = av[i];
		debuglevel = atoi (s);
		open_debug ();
		debug ("-----------------------------------------------------");
	    }
	    else if (!stricmp (s, "live") || !stricmp (s, "nofork"))
		runflags |= RUN_LIVE;
	}
    }

    me.since = time (NULL);
    me.reset = time (NULL);
    last_check = time (NULL);

    runflags |= RUN_STARTING;

    snprintf (quitmsg, BUFSIZE, "Reason unknown");
    servsock = -1;

    /* Open the logfile */
    open_log ();

    /* Fork into the background */
    if (!(runflags & RUN_LIVE))
    {
	if ((i = fork ()) < 0)
	{
	    log ("Cannot fork!");
	    return -1;
	}
	else if (i != 0)
	    exit (0);

	if (setpgid (0, 0) < 0)
	    return -1;
    }

    pid_file = fopen (PIDFILE, "w");

    if (pid_file)
    {
	fprintf (pid_file, "%d\n", getpid ());
	fclose (pid_file);
    }
    else
	log (ERR_CANT_WRITE_PID);

    /* Write to the log so we know we're on a new session */
    log ("-----------------------------------------------------");
    log ("Cygnus IRC Services v%s: Compiled for %s", version,
#ifdef UNREAL3_2
	    "UnrealIRCd 3.2");
#elif defined (UNREAL3)
	    "UnrealIRCd 3.1");
#elif defined (PROMETHEUS)
	    "Prometheus");
#elif defined (BAHAMUT)
	    "Bahamut");
#elif defined (IRCUP9)
	    "IRCu (P9)");
#else                    
	    "DreamForge");
#endif

    /* Set signal handlers */
    signal (SIGINT, sighandler);
    signal (SIGTERM, sighandler);
    signal (SIGPIPE, SIG_IGN);
    signal (SIGQUIT, sighandler);
#ifndef DUMP_CORE
    signal (SIGSEGV, sighandler);
#endif
    signal (SIGBUS, sighandler);
    signal (SIGHUP, sighandler);
    signal (SIGILL, sighandler);
    signal (SIGTRAP, sighandler);
    signal (SIGIOT, sighandler);
    signal (SIGFPE, sighandler);
    signal (SIGALRM, SIG_IGN);
    signal (SIGUSR2, SIG_IGN);
    signal (SIGCHLD, SIG_IGN);
    signal (SIGWINCH, SIG_IGN);
    signal (SIGTTIN, SIG_IGN);
    signal (SIGTTOU, SIG_IGN);
    signal (SIGTSTP, SIG_IGN);
    signal (SIGUSR1, sighandler);

    /* Load our conf */
    load_conf ();

    /* Check the conf */
    check_conf ();

    /* Set the news time in case its not in the DB */
    news_time = time (NULL);

    /* Set up RootServ stuff */
    if (rootserv_on == TRUE)
    {
	rs.version = 1;
	load_rs_dbase ();
    }

    /* Set up NickServ stuff */
    if (nickserv_on == TRUE)
    {
	ns.version = 1;
	load_ns_dbase ();
    }

    /* Set up ChanServ stuff */
    if (chanserv_on == TRUE)
    {
	cs.version = 1;
	load_cs_dbase ();
    }

    /* Count our stuff */
    userstat += servcount ();

    log_sameline (1, "Connecting to %s:%d... ", me.uplink, me.port);
    servsock = conn (me.uplink, me.port, me.bindhost, me.bindport);

    if (servsock < 0)
    {
	log_sameline (0, ERR_CANT_CONNECT);
	disconn (servsock);
	close_log();

	if (debuglevel)
	    close_debug();

	remove (PIDFILE);
	return 0;
    }

    log_sameline (0, "connected.");
    log_sameline (1, "Logging in to server... ");

    /* Send our initial login info to the server. We don't tokenize
       at this point since we wouldn't have gotten much if anything
       from the server yet anyhow ..
     */
#ifdef UNREAL3
    send_cmd (NULL, "PROTOCTL TOKEN NICKv2 UMODE2 VL SJOIN NOQUIT");
#elif !defined (BAHAMUT) && !defined (IRCUP9) && !defined (UNREAL3) && !defined (PROMETHEUS)
    send_cmd (NULL, "PROTOCTL TOKEN NOQUIT");
#endif

#if defined (BAHAMUT) || defined (PROMETHEUS)
    send_cmd (NULL, "PASS %s :TS", me.pass);
#else
    send_cmd (NULL, "PASS :%s", me.pass);
#endif

#ifdef UNREAL3
    send_cmd (NULL, "SERVER %s 1 :U0-*-%d %s", me.name, unreal_numeric, me.desc);
#elif defined (IRCUP9)
    send_cmd (NULL, "SERVER %s 1 %lu %lu P09 :%s", me.name, me.since, me.since, me.desc);
#else
    send_cmd (NULL, "SERVER %s 1 :%s", me.name, me.desc);
#endif

#if defined (BAHAMUT) || defined (PROMETHEUS)
    send_cmd (NULL, "SVINFO 3 3 0 :%lu", time (NULL));
    send_cmd (NULL, "CAPAB TS3 NOQUIT");
#endif

    log_sameline (0, "success.");

    /* Now bring on our clients.. */
    if (rootserv_on == TRUE)
    {
	intro_user (s_RootServ, rs.user, rs.host, rs.real, rs.mode);
	rs.isonline = 1;
    }

    if (nickserv_on == TRUE)
    {
	intro_user (s_NickServ, ns.user, ns.host, ns.real, ns.mode);
	ns.isonline = 1;
    }

    if (chanserv_on == TRUE)
    {
	intro_user (s_ChanServ, cs.user, cs.host, cs.real, cs.mode);
	cs.isonline = 1;
    }

    if (memoserv_on == TRUE)
    {
	intro_user (s_MemoServ, ms.user, ms.host, ms.real, ms.mode);
	ms.isonline = 1;
    }

    if (globalnoticer_on == TRUE)
    {
	intro_user (s_GlobalNoticer, gn.user, gn.host, gn.real, gn.mode);
	gn.isonline = 1;
    }

    /* Put all clients in the JoinChan */
    joinchan_join ();

    last_update = time (NULL);
    last_check = time (NULL);

    setjmp (panic_jmp);

    /* Bahamut sends two pings to indicate the end of the burst */
#ifdef BAHAMUT
    bursting = 2;
    burststart = time (NULL);
#elif defined (PROMETHEUS) || defined (UNREAL3)
    bursting = 1;
    burststart = time (NULL);
#endif

    /* Send status global if nessecary. */
    if (statglobal_on == TRUE)
	noticeall (RPL_SERVICESBACK);

    /* How long did that take? */
#ifdef HAVE_GETTIMEOFDAY
    gettimeofday (&now, NULL);
    timersub (&now, &start, &tmp);
    log ("Cygnus startup completed in %d ms", tv2ms (&tmp));
#endif

    runflags &= ~RUN_STARTING;

    /* Main loop */
    if (servsock <= 0)
	log (ERR_CANT_CONNECT);
    else
	read_loop ();

    /* Let the users know we're shutting down */
    if (statglobal_on == TRUE)
	noticeall (RPL_SERVICESLEAVING);

    if (!quitmsg)
	snprintf (quitmsg, BUFSIZE, "Terminating, reason unknown");

    log (ERR_QUIT, quitmsg);

#ifdef IRCUP9
    send_cmd (me.name, "SQUIT %s 0 :%s", me.name, quitmsg);
#else
    send_cmd (me.name, "%s %s :%s", me.token ? "-" : "SQUIT",
	me.name, quitmsg);
#endif

    disconn (servsock);

    /* Should we reboot? Sleep, then restart. */
    if (wait_restart > 0 && runflags & RUN_RESTART)
    {
	log (RPL_RESTARTING, duration (wait_restart, 1));
	close_log ();

	if (debuglevel)
	    close_debug();

	sleep (wait_restart);
	execve (SERVICES_BINARY, av, envp);
	return 1;
    }
    else
    {
	close_log ();

	if (debuglevel)
	    close_debug();

	remove (PIDFILE);
	return 0;
    }

    return 0; 
}

/* Bring a new client onto the network. */
void intro_user (const char *nick, const char *user, const char *host,
		 const char *real, const char *modes)
{
#ifdef UNREAL3
    send_cmd (NULL, "%s %s 1 %lu %s %s %s 1 %s * :%s", me.token ? "&" : "NICK",
	nick, time (NULL), user, host, me.name, modes, real);
#elif defined (BAHAMUT) || defined (PROMETHEUS)
    send_cmd (NULL, "NICK %s 1 %lu %s %s %s %s 0 :%s", nick, time (NULL), modes, user,
	host, me.name, real);
#elif defined (IRCUP9)
    send_cmd (NULL, "NICK %s 1 %lu %s %s %s :%s", nick, time (NULL), user, host,
	me.name, real);
    send_cmd (nick, "MODE %s %s", nick, modes);
#else
    send_cmd (NULL, "%s %s 1 %lu %s %s %s 1 :%s", me.token ? "&" : "NICK",
	nick, time (NULL), user, host, me.name, real);
    send_cmd (nick, "%s %s %s", me.token ? "G" : "MODE", nick, modes);
#endif
}

/* Check for various timeouts */
void chk (void)
{
    time_t curtime = time (NULL);

    check_timeouts ();

    if ((runflags & RUN_SAVEDBS) ||
	   curtime - last_update >= update_timeout)
    {
	expire_emails ();
	expire_akills ();
	save_rs_dbase ();

	if (nickserv_on == TRUE)
	{
	    expire_nicks ();
	    save_ns_dbase ();
	}

	if (chanserv_on == TRUE)
	{
	    expire_chans ();
	    save_cs_dbase ();
	}

	last_update = curtime;
	runflags &= ~RUN_SAVEDBS;
	snoop (s_RootServ, RS_DB_SYNC_COMPLETE);
    }
}
