/* Processing code.

   Cygnus IRC Services - Copyright (c) 2001-2002 Darcy Grexton
   Contact: skold@habber.net, skold @ HabberNet, ICQ #6285178

   This program is free but copyrighted software; see the file COPYING
   for details.
 */

#include "../inc/services.h"

Servers *servlist = NULL;
int serv_size = 0;

/* split_buf:  Split a buffer into arguments and store the arguments in an
               argument vector pointed to by argv (which will be malloc'd
               as necessary); return the argument count.  If colon_special
               is non-zero, then treat a parameter with a leading ':' as
               the last parameter of the line, per the IRC RFC. Destroys
               the buffer by side effect.
 */
int split_buf (char *buf, char ***argv, int colon_special)
{
    int argvsize = 8;
    int argc;
    char *s;

    *argv = smalloc (sizeof(char *) * argvsize);
    argc = 0;
    while (*buf)
    {
	if (argc == argvsize)
	{
	    argvsize += 8;
	    *argv = srealloc (*argv, sizeof(char *) * argvsize);
	}
	if (*buf == ':')
	{
	    (*argv)[argc++] = buf+1;
	    buf = "";
	}
	else
	{
	    s = strpbrk(buf, " ");
	    if (s)
	    {
		*s++ = 0;
		while (isspace(*s))
		    s++;
	    }
	    else
		s = buf + strlen(buf);
	    (*argv)[argc++] = buf;
	    buf = s;
	}
    }
    return argc;
}

/* process:  Main processing routine. */
void process ()
{
    char source[64];
    char cmd[64];
    char buf[512];
    char *s;
    int ac;
    char **av;
    Message *m;

    /* If debugging, log the buffer. */
    if (debuglevel == 1 || debuglevel == 3)
	debug ("--> %s", inbuf);

    /* First make a copy of the buffer so we have the original in case we
       crash - in that case, we want to know what we crashed on.
     */
    strscpy (buf, inbuf, sizeof (buf));

    /* Split the buffer into pieces. */
    if (*buf == ':')
    {
	s = strpbrk (buf, " ");

	if (!s)
	    return;
	*s = 0;

	while (isspace(*++s))
	    ;

	strscpy (source, buf+1, sizeof(source));
	memmove (buf, s, strlen (s)+1);
    }
    else
	*source = 0;

    if (!*buf)
	return;

    s = strpbrk (buf, " ");

    if (s)
    {
	*s = 0;
	while (isspace (*++s))
	    ;
    }
    else
	s = buf + strlen(buf);

    strscpy (cmd, buf, sizeof(cmd));
    ac = split_buf (s, &av, 1);

    /* Do something with the message. */
    m = find_message (cmd);

    if (m)
	if (m->func)
	    m->func (source, ac, av);

    /* Free argument list we created */
    free (av);
}

/* Send our MOTD */
static void m_motd (char *source, int ac, char **av)
{
    NickInfo *ni;
    FILE *f;
    char buf[BUFSIZE];
    int i;

    send_cmd (me.name, "375 %s :- %s Message of the Day -", source, me.name);
    send_cmd (me.name, "372 %s :- ", source);
    send_cmd (me.name,
	"372 %s :- Cygnus IRC Services v%s Copyright (c) 2001-2002 Darcy Grexton",
	source, version);
    send_cmd (me.name, "372 %s :- ", source);

    f = fopen (MOTDFILE, "r");

    if (f)
    {
	while (fgets (buf, sizeof (buf), f))
	{
	    buf[strlen (buf) - 1] = 0;
	    send_cmd (me.name, "372 %s :- %s", source, buf);
	}
	fclose (f);
    }

    /* List the csops. Ones who are online should be in bold. */
    send_cmd (me.name, "372 %s :- ", source);
    send_cmd (me.name, "372 %s :- Channels expire after \2%s\2 of non use.", source,
	duration (chan_expire, 2));
    send_cmd (me.name, "372 %s :- Nicknames expire after \2%s\2 of non use.", source,
	duration (nick_expire, 2));
    send_cmd (me.name, "372 %s :- ", source);
    send_cmd (me.name, "372 %s :- The following people can help you recover lost or", source);
    send_cmd (me.name, "372 %s :- forgotten passwords. \2Bold\2 nicknames are online.", source);
    send_cmd (me.name, "372 %s :- ", source);

    *buf = 0;

    for (i = 0; i < 1024; i++)
    {
	if (strlen (buf) > 70)
	{
	    send_cmd (me.name, "372 %s :- %s", source, buf);
	    *buf = 0;
	}

	for (ni = nicklist[i]; ni; ni = ni->next)
	{   
	    if (can_csop (ni->nick))
	    {
		if (finduser (ni->nick) && is_oper (ni->nick))
		{
		    strcat (buf, "\2");
		    strcat (buf, ni->nick);
		    strcat (buf, "\2");
		}
		else
		    strcat (buf, ni->nick);

		strcat (buf, " ");
	    }
	}
    }

    if (*buf)
	send_cmd (me.name, "372 %s :- %s", source, buf);

    send_cmd (me.name, "372 %s :- ", source);
    send_cmd (me.name, "376 %s :End of /MOTD command.", source);
}

/* Reply to a TIME query */
static void m_time (char *source, int ac, char **av)
{
    time_t t;
    struct tm *tm;
    char buf[80];

    time (&t);
    tm = localtime (&t);
    strftime (buf, sizeof (buf), "%A %B %d %Y -- %H:%M %z", tm);

    send_cmd (me.name, "391 %s %s :%s", source, me.name, buf);
}

/* Send an ADMIN reply */
static void m_admin (char *source, int ac, char **av)
{
    send_cmd (me.name, "256 %s :Administrative info about %s", source,
	me.name);
    send_cmd (me.name, "257 %s :Cygnus IRC Services v%s", source, version);
    send_cmd (me.name, "258 %s :Copyright (c) 2001-2002 Darcy Grexton",
	source);
    send_cmd (me.name,
	"259 %s :For more information E-Mail skold@habber.net.", source);
}

/* Handle an ERROR command */
static void m_error (char *source, int ac, char **av)
{
    snprintf (quitmsg, BUFSIZE, inbuf);
    log (ERR_ERROR, inbuf);
    runflags |= RUN_SHUTDOWN;
}

/* Reply to a ping */
static void m_ping (char *source, int ac, char **av)
{
    if (ac < 1)
	return;

    /* Reply to the ping with the servername the ping came for. This
       stops bad ping replies when we get PINGs for things like jupes.
       Bahamut seems to fuck up pings when bursting though, so we have
       to account for that too..
     */
    if (ac > 1)
	send_cmd (av[1], "%s %s %s", me.token ? "9" : "PONG",
	    ac > 1 ? av[1] : me.name, av[0]);
    else
	send_cmd (me.name, "%s %s %s", me.token ? "9" : "PONG",
	    ac > 1 ? av[1] : me.name, av[0]);

#ifdef BAHAMUT
    /* Bahamut sends two PINGs at the end of the burst. We'll use that to figure
       out when it's over.
     */
    if (bursting > 0)
	bursting--;

    /* Now that the burst is done, run through every channel, and do the modechanges
       and topic changes as needed.
     */
    if (bursted && !bursting)
    {
	globops (s_RootServ, "Sync'd to network in %s: %d users, %d channels, %d servers.",
	    duration (time (NULL) - burststart, 1), userstat, chancnt, servcnt + 1);

	check_eb ();
    }

    /* Since Bahamut uses PINGs to show the end of the burst, we need somethign different
       to track the end of burst. Otherwise we'll keep sending sync'd notices with each
       ping.
     */  
    if (bursting == 0)
        bursted = 0;
#endif
}

/* Handle a KICK command */
static void m_kick (char *source, int ac, char **av)
{
    if (ac != 3)
	return; 

    do_kick (source, ac, av);
}

/* Handle a TOPIC command */
static void m_topic (char *source, int ac, char **av)
{
    if (ac != 4)
	return;

    do_topic (source, ac, av);
}

/* Handle a JOIN command */
static void m_join (char *source, int ac, char **av)
{
    if (ac != 1)
	return;

    do_join (source, ac, av);
}

#if defined (BAHAMUT) || defined (PROMETHEUS) || defined (UNREAL3)
/* Handle an SJOIN command. */
static void m_sjoin (char *source, int ac, char **av)
{
    do_sjoin (source, ac, av);
}
#endif

/* Handle a PART command */
static void m_part (char *source, int ac, char **av)
{
    if (ac < 1 || ac > 2)   
	return;

    do_part (source, ac, av);
}

/* An AKILL was placed elsewhere, learn it. */
static void m_akill (char *source, int ac, char **av)
{
    AutoKill *akill, **list;
    char *mask;

    /* Should we learn new AKills? */
    if (learnakills_on == FALSE)
	return;

    mask = smalloc (strlen (av[1]) + strlen (av[0]) + 2);
    snprintf (mask, strlen (av[1]) + strlen (av[0]) + 2, "%s@%s", av[1],
	av[0]);

    /* Ignore it if we already have it. */
    if (is_akilled (mask))
	return;

    akill = scalloc (sizeof (AutoKill), 1);
    akill->mask = sstrdup (mask);
    akill->reason = sstrdup (av[2]);
    akill->setter = sstrdup (source);
    akill->realname = 0;
    akill->set = time (NULL);
    akill->expires = def_akill_time;
    list = &akilllist[AHASH(akill->mask)];
    akill->next = *list;
    if (*list)
	(*list)->prev = akill;
    *list = akill;

    if (globalonakill_on == TRUE)
	globops (s_RootServ, RS_AKILL_LEARNED, source, mask, av[2]);

    log (RS_AKILL_LEARNED, source, mask, av[2]);

    free (mask);
}

/* An AKill was removed elsewhere, remove it if we have it. */
static void m_rakill (char *source, int ac, char **av)
{
    AutoKill *akill;
    int i, j;
    char *mask;

    /* Should we forget AKills? */
    if (forgetakills_on == FALSE)
	return;

    mask = smalloc (strlen (av[1]) + strlen (av[0]) + 2);
    snprintf (mask, strlen (av[1]) + strlen (av[0]) + 2, "%s@%s", av[1],
	av[0]);

    /* Make sure we have it */
    if (!is_akilled (mask))
	return;

    for (i = 0; i < 1024; i++)
    {
	for (akill = akilllist[i]; akill; akill = akill->next)
	{
	    j++;

	    if (!stricmp (mask, akill->mask))
	    {
		free (akill->setter);
		free (akill->reason);
		if (akill->prev)
		    akill->prev->next = akill->next;
		else
		    akilllist[AHASH(akill->mask)] = akill->next;
		if (akill->next)
		    akill->next->prev = akill->prev;

		free (akill);

		if (globalonakill_on == TRUE)
		    globops (s_RootServ, RS_AKILL_FORGOTTEN, mask, source);
		log (RS_AKILL_FORGOTTEN, mask, source);

		return;
	    }
	}
    }
}

/* Handle a NICK command */
static void m_nick (char *source, int ac, char **av)
{
#ifdef IRCUP9
    if (strchr (source, '.'))
	*source = 0;
#endif

    do_nick (source, ac, av);
}

/* Handle a QUIT command */
static void m_quit (char *source, int ac, char **av)
{
    if (ac != 1)
	return;

    if (!finduser (source))
    {
	if (debuglevel > 1)
	    debug ("do_quit(): Got QUIT for nonexistant user %s", source);

	return;
    }

    do_quit (source, ac, av);
}

/* Send our version reply */
void m_version (char *source, int ac, char **av)
{
    if (source)
    {
	/* Send version replies with the server name that was queried. If it's
	   not the same as our name, theyre likely versioning a juped server,
	   so we'll reply as such.
	 */
	if (!stricmp (me.name, av[0]))
	    send_cmd (me.name, "351 %s cygnus-%s. %s %s%s%s%s%s%s",
		source, version, me.name, (rootserv_on == TRUE) ? "R" : "",
		(nickserv_on == TRUE) ? "N" : "", (chanserv_on == TRUE) ? "C" : "",
		(memoserv_on == TRUE) ? "M" : "", (globalnoticer_on == TRUE) ? "g" : "",
		debuglevel ? "d" : "");
	else
	    send_cmd (av[0], "351 %s juped-server. %s :XX", source, av[0]);
    }
}

/* Handle an AWAY command */
void m_away (char *source, int ac, char **av)
{
    User *u = finduser (source);

    if (u && (ac == 0 || *av[0] == 0))
    {
	check_memos (u);

	if (is_csop (u->nick))
	    check_verifies (u);
    }
}

/* Reply to an INFO command */
void m_info (char *source, int ac, char **av)
{
    int i;

    for (i = 0; infotext[i]; i++)
	send_cmd (me.name, "371 %s :%s", source, infotext[i]);
    send_cmd (me.name, "371 %s :cygnus-%s", source, version);
    send_cmd (me.name,
	      "371 %s :Birth Date: %s, compile # %s", source, compile_time,
	      build);
    send_cmd (me.name, "371 %s :On-line since %s", source, ctime (&me.since));
    send_cmd (me.name, "374 %s :End of /INFO list.", source);
}

#ifdef UNREAL3
/* Reply to a NETINFO command. */
void m_netinfo (char *source, int ac, char **av)
{
    send_cmd (NULL, "%s %d %lu 0 %s * * * :%s", me.token ? "AO" : "NETINFO",
	usercnt, time (NULL), av[3], av[7]);

    globops (s_RootServ, "Sync'd to network in %s: %d users, %d channels, %d servers.",
	duration (time (NULL) - burststart, 1), userstat, chancnt, servcnt + 1);

    bursting = 0;

    /* Now that the burst is done, run through every channel, and do the modechanges
       and topic changes as needed.
     */
    check_eb ();
}
#endif /* UNREAL3 */

/* Reply to a STATS query */
void m_stats (char *source, int ac, char **av)
{
    if (ac < 1)
	return;

    switch (*av[0])
    {
	case 'u':
	{
	    int uptime = time (NULL) - me.since;
	    int tmp = servcount ();

	    send_cmd (me.name, "242 %s :Server Up %d days, %d:%02d:%02d",
		source, uptime/86400, (uptime/3600) % 24, (uptime/60) % 60,
		uptime % 60);
	    send_cmd (me.name,
		"250 %s :Highest connection count: %d (%d clients)",
		source, tmp + 1, tmp);
	    send_cmd (me.name, "219 %s u :End of /STATS report.", source);
	    break;
	}
	default:
	    send_cmd (me.name, "219 %s * :End of /STATS report.", source);
    }
}

/* Reply to a WHOIS query */
void m_whois (char *source, int ac, char **av)
{   
    if (ac < 1)
	return;   

    if ((!stricmp (av[0], s_RootServ)) && rootserv_on == TRUE)
    {
	send_cmd (me.name, "311 %s %s %s %s * :%s", source, s_RootServ,
	    rs.user, rs.host, rs.real);
	send_cmd (me.name, "312 %s %s %s :%s", source, s_RootServ,
	    me.name, me.desc);
	if (strchr (rs.mode, 'o'))
	    send_cmd (me.name, "313 %s %s :is an IRC Operator", source,
		s_RootServ);
	send_cmd (me.name, "317 %s %s 0 %d :seconds idle, signon time",
	    source, s_RootServ, me.since);
    }
    else if ((!stricmp (av[0], s_NickServ)) && nickserv_on == TRUE)
    {
	send_cmd (me.name, "311 %s %s %s %s * :%s", source, s_NickServ,
	    ns.user, ns.host, ns.real);
	send_cmd (me.name, "312 %s %s %s :%s", source, s_NickServ,
	    me.name, me.desc);
	if (strchr (ns.mode, 'o'))
	    send_cmd (me.name, "313 %s %s :is an IRC Operator", source,
		s_NickServ);
	send_cmd (me.name, "317 %s %s 0 %d :seconds idle, signon time",
	    source, s_NickServ, me.since);
    }
    else if ((!stricmp (av[0], s_ChanServ)) && chanserv_on == TRUE)
    {
	send_cmd (me.name, "311 %s %s %s %s * :%s", source, s_ChanServ,
	    cs.user, cs.host, cs.real);
	send_cmd (me.name, "312 %s %s %s :%s", source, s_ChanServ,
	    me.name, me.desc);
	if (strchr (ms.mode, 'o'))
	    send_cmd (me.name, "313 %s %s :is an IRC Operator", source,
		s_ChanServ);
	send_cmd (me.name, "317 %s %s 0 %d :seconds idle, signon time",
	    source, s_ChanServ, me.since);
    }
    else if ((!stricmp (av[0], s_MemoServ)) && memoserv_on == TRUE)
    {
	send_cmd (me.name, "311 %s %s %s %s * :%s", source, s_MemoServ,
	    ms.user, ms.host, ms.real);
	send_cmd (me.name, "312 %s %s %s :%s", source, s_MemoServ,
	    me.name, me.desc);
	if (strchr (ms.mode, 'o'))
	    send_cmd (me.name, "313 %s %s :is an IRC Operator", source,
		s_MemoServ);
	send_cmd (me.name, "317 %s %s 0 %d :seconds idle, signon time",
	    source, s_MemoServ, me.since);
    }
    else if ((!stricmp (av[0], s_GlobalNoticer)) && globalnoticer_on == TRUE)
    {
	send_cmd (me.name, "311 %s %s %s %s * :%s", source, s_GlobalNoticer,
	    gn.user, gn.host, gn.real);
	send_cmd (me.name, "312 %s %s %s :%s", source, s_GlobalNoticer,
	    me.name, me.desc);
	if (strchr (gn.mode, 'o'))
	    send_cmd (me.name, "313 %s %s :is an IRC Operator", source,
		s_GlobalNoticer);
	send_cmd (me.name, "317 %s %s 0 %d :seconds idle, signon time",
	    source, s_GlobalNoticer, me.since);
    }
    else
	send_cmd (me.name, "401 %s %s :No such nick/channel", source, av[0]);
    send_cmd (me.name, "318 %s %s :End of /WHOIS list", source, av[0]);
}

/* Handle a SERVER command */
static void m_server (char *source, int ac, char **av)
{
    if (servcnt >= serv_size)
    {
	if (serv_size < 8)
	    serv_size = 8;
	else
	    serv_size *= 2;
	servlist = srealloc (servlist, sizeof (*servlist) * serv_size);
    }

    servlist[servcnt].server = sstrdup (av[0]);
    servlist[servcnt].uplink = sstrdup (source);
    servcnt++;
}

/* Handle an SQUIT command */
static void m_squit (char *source, int ac, char **av)
{
    User *u;
    int i, delcnt = 0;
    char **delnicks = NULL;

    for (i = 0; i < servcnt && stricmp (servlist[i].server, av[0]); i++);

    if (i < servcnt)
    {
	/* We give servers the NOQUIT token on link .. If they don't
	   understand it, they'll send QUITs first, so we won't find
	   any users for that server.
	 */
	for (u = firstuser (); u; u = nextuser ())
	{
	    if (u && !stricmp (servlist[i].server, u->server))
	    {
		/* If this is a nickname we were waiting for, reclaim it. */
		if (rs.wantnick)
		    if (!stricmp (u->nick, rs.nick))
		    {
			send_cmd (s_RootServ, "%s %s", me.token ? "&" : "NICK",
			    rs.nick);
			strscpy (s_RootServ, rs.nick, sizeof (s_RootServ));
			rs.wantnick = 0;
		    }

		if (ns.wantnick)
		    if (!stricmp (u->nick, ns.nick))
		    {
			send_cmd (s_NickServ, "%s %s", me.token ? "&" : "NICK",
			    ns.nick);
			strscpy (s_NickServ, ns.nick, sizeof (s_NickServ));
			ns.wantnick = 0;
		    }

		if (cs.wantnick)
		    if (!stricmp (u->nick, cs.nick))
		    {
			send_cmd (s_ChanServ, "%s %s", me.token ? "&" :
			    "NICK", cs.nick);
			strscpy (s_ChanServ, cs.nick, sizeof (s_ChanServ));
			cs.wantnick = 0;
		    }

		if (ms.wantnick)
		    if (!stricmp (u->nick, ms.nick))
		    {
			send_cmd (s_MemoServ, "%s %s", me.token ? "&" :
			    "NICK", ms.nick);
			strscpy (s_MemoServ, ms.nick, sizeof (s_MemoServ));
			ms.wantnick = 0;
		    }

		if (gn.wantnick)
		    if (!stricmp (u->nick, gn.nick))
		    {
			send_cmd (s_GlobalNoticer, "%s %s", me.token ? "&" :
			    "NICK", gn.nick);
			strscpy (s_GlobalNoticer, gn.nick, sizeof
			    (s_GlobalNoticer));
			gn.wantnick = 0;
		    }

		++delcnt;
		delnicks = srealloc (delnicks, sizeof (char *) *delcnt);
		delnicks[delcnt - 1] = sstrdup (u->nick);
	    }
	}
    }

#if defined (BAHAMUT) || defined (PROMETHEUS)

    /* Bahamut only sends SQUITs for the splitting server, not servers linked
       to it. So we have to free those servers and their users here.
     */
    for (i = 0; i < servcnt; i++)
    {
	if (!stricmp (servlist[i].uplink, av[0]))
	{
	    for (u = firstuser (); u; u = nextuser ())
	    {
		if (u && !stricmp (servlist[i].server, u->server))
		{
		    if (rs.wantnick)
			if (!stricmp (u->nick, rs.nick))
			{
			    send_cmd (s_RootServ, "%s %s", me.token ? "&" : "NICK",
				rs.nick);
			    strscpy (s_RootServ, rs.nick, sizeof (s_RootServ));
			    rs.wantnick = 0;
			}

		    if (ns.wantnick)
			if (!stricmp (u->nick, ns.nick))
			{
			    send_cmd (s_NickServ, "%s %s", me.token ? "&" : "NICK",
				ns.nick);
			    strscpy (s_NickServ, ns.nick, sizeof (s_NickServ));
			    ns.wantnick = 0;
			}

		    if (cs.wantnick)
			if (!stricmp (u->nick, cs.nick))
			{
			    send_cmd (s_ChanServ, "%s %s", me.token ? "&" :
				"NICK", cs.nick);
			    strscpy (s_ChanServ, cs.nick, sizeof (s_ChanServ));
			    cs.wantnick = 0;
			}

		    if (ms.wantnick)
			if (!stricmp (u->nick, ms.nick))
			{
			    send_cmd (s_MemoServ, "%s %s", me.token ? "&" :
				"NICK", ms.nick);
			    strscpy (s_MemoServ, ms.nick, sizeof (s_MemoServ));
			    ms.wantnick = 0;
			}

		    if (gn.wantnick)
			if (!stricmp (u->nick, gn.nick))
			{
			    send_cmd (s_GlobalNoticer, "%s %s", me.token ? "&" :
				"NICK", gn.nick);
			    strscpy (s_GlobalNoticer, gn.nick, sizeof
				(s_GlobalNoticer));
			    gn.wantnick = 0;
			}

		    ++delcnt;
		    delnicks = srealloc (delnicks, sizeof (char *) *delcnt);
		    delnicks[delcnt - 1] = sstrdup (u->nick);
		}
	    }

	    if (debuglevel > 1)
		debug ("m_squit(): Deleting server %s", servlist[i].server);

	    if (servlist[i].server)
		free (servlist[i].server);

	    if (servlist[i].uplink)
		free (servlist[i].uplink);

	    servcnt--;

	    if (i < servcnt)
		bcopy (servlist + i + 1, servlist + i, sizeof
		    (*servlist) * (servcnt - i));

	    i--;
	}
    }

    /* Free the splitting server */
    for (i = 0; i < servcnt && stricmp (servlist[i].server, av[0]); i++);

    if (debuglevel > 1)
	debug ("m_squit(): Deleting server %s", servlist[i].server);

    if (servlist[i].server)
	free (servlist[i].server);

    if (servlist[i].uplink)
	free (servlist[i].uplink);

    servcnt--;

    if (i < servcnt)
	bcopy (servlist + i + 1, servlist + i, sizeof
	    (*servlist) * (servcnt - i));

    /* Now free all the affected users. */
    for (i = 0; i < delcnt; delnicks++, i++)
    {
	u = finduser (*delnicks);

	if (u)
	    delete_user (u);

	free (*delnicks);
    }

    delnicks = NULL;

    /* We're done, return here */
    return;

#endif /* BAHAMUT */

    for (i = 0; i < servcnt && stricmp (servlist[i].server, av[0]); i++);

    if (i < servcnt)
    {
	if (debuglevel > 1)
	    debug ("m_squit(): Deleting server %s", servlist[i].server);

	if (servlist[i].server)
	    free (servlist[i].server);

	if (servlist[i].uplink)
	    free (servlist[i].uplink);

	servcnt--;

	if (i < servcnt)
	    bcopy (servlist + i + 1, servlist + i, sizeof
		(*servlist) * (servcnt - i));
    }
}

/* Handle a KILL command. */
static void m_kill (char *source, int ac, char **av)
{
    if (ac != 2)
	return;

    /* Before we reintroduce the user, we'll check to see if they've
       been collided more than 5 times in the last 5 seconds. This is
       a common problem with people who forget to rename/disable services
       with the same nicks as some of ours. If they HAVE been collided
       that many times, we'll give them a nick like RootServ986662577 and
       send a globops about it. We'll then watch for a nickchange/quit from
       the client with our nick, and retake it if the opportunity presents
       itself.
     */
    if ((rootserv_on == TRUE) && (!stricmp (av[0], s_RootServ)))
    {
	/* How many times we've been collided */
	++rs.collided;

	/* If this is true, we give up trying to retake our nick.. */
	if (((rs.lastcoll >= (time (NULL) - 5)) && (rs.collided >= 5)))
	{
	    /* Check if it's a server killing us. If it's not, then it's
	       likely a regular kill, and we don't want to take a waiting
	       nick.
	     */
	    if (!strchr (source, '.'))
	    {
		/* It's not a server. We'll reintro with the regular
		   nick and leave it at that.
		 */
		intro_user (s_RootServ, rs.user, rs.host, rs.real, rs.mode);
		return;
	    }

	    /* Are we already waiting for our nick? Some jackass might be
	       killing us off. We don't want to end up with
	       RootServ903218973782743871412849 or something..
	       We'll only copy our current nick over if this isn't true.
	       That way we'll always end up with a new number.
	     */
	    if (!rs.wantnick)
	    {
		/* Make note of the nick we want */
		strscpy (rs.nick, s_RootServ, sizeof (rs.nick));
	    }

	    /* Make note that we want it */
	    rs.wantnick = 1;

	    /* Give ourselves a new nick */
	    snprintf (s_RootServ, sizeof (s_RootServ), "%s%d", rs.nick,
		getpid());

	    /* Introduce with the new nick */
	    intro_user (s_RootServ, rs.user, rs.host, rs.real, rs.mode);

	    /* Let the staff know about it */
	    globops (s_RootServ, ERR_COLLISIONS, rs.nick);

	    rs.collided = 0;

	    return;
	}
	else
	    rs.lastcoll = time (NULL);

	/* Re-introduce our client */
	intro_user (s_RootServ, rs.user, rs.host, rs.real, rs.mode);
    }
    else if ((nickserv_on == TRUE) && !stricmp (av[0], s_NickServ))
    {
	/* How many times we've been collided */
	++ns.collided;

	/* If this is true, we give up trying to retake our nick.. */
	if (((ns.lastcoll >= (time (NULL) - 5)) && (ns.collided >= 5)))
	{
	    /* Check if it's a server killing us. If it's not, then it's
	       likely a regular kill, and we don't want to take a waiting
	       nick.
	     */
	    if (!strchr (source, '.'))
	    {
		/* It's not a server. We'll reintro with the regular
		   nick and leave it at that.
		 */
		intro_user (s_NickServ, ns.user, ns.host, ns.real, ns.mode);
		return;
	    }

	    /* Are we already waiting for our nick? Some jackass might be
	       killing us off. We don't want to end up with
	       RootServ903218973782743871412849 or something..
	       We'll only copy our current nick over if this isn't true.
	       That way we'll always end up with a new number.
	     */
	    if (!ns.wantnick)
	    {
		/* Make note of the nick we want */
		strscpy (ns.nick, s_NickServ, sizeof (ns.nick));
	    }

	    /* Make note that we want it */
	    ns.wantnick = 1;

	    /* Give ourselves a new nick */
	    snprintf (s_NickServ, sizeof (s_NickServ), "%s%d", ns.nick,
		getpid());

	    /* Introduce with the new nick */
	    intro_user (s_NickServ, ns.user, ns.host, ns.real, ns.mode);

	    /* Let the staff know about it */
	    globops (s_NickServ, ERR_COLLISIONS, ns.nick);

	    ns.collided = 0;

	    return;
	}
	else
	    ns.lastcoll = time (NULL);

	/* Re-introduce our client */
	intro_user (s_NickServ, ns.user, ns.host, ns.real, ns.mode);
    }
    else if ((chanserv_on == TRUE) && !stricmp (av[0], s_ChanServ))
    {
	/* How many times we've been collided */
	++cs.collided;

	/* If this is true, we give up trying to retake our nick.. */
	if (((cs.lastcoll >= (time (NULL) - 5)) && (cs.collided >= 5)))
	{
	    /* Check if it's a server killing us. If it's not, then it's
	       likely a regular kill, and we don't want to take a waiting
	       nick.
	     */
	    if (!strchr (source, '.'))
	    {
		/* It's not a server. We'll reintro with the regular
		   nick and leave it at that.
		 */
		intro_user (s_ChanServ, cs.user, cs.host, cs.real, cs.mode);
		return;
	    }

	    /* Are we already waiting for our nick? Some jackass might be
	       killing us off. We don't want to end up with
	       RootServ903218973782743871412849 or something..
	       We'll only copy our current nick over if this isn't true.
	       That way we'll always end up with a new number.
	     */
	    if (!cs.wantnick)
	    {
		/* Make note of the nick we want */
		strscpy (cs.nick, s_ChanServ, sizeof (cs.nick));
	    }

	    /* Make note that we want it */
	    cs.wantnick = 1;

	    /* Give ourselves a new nick */
	    snprintf (s_ChanServ, sizeof (s_ChanServ), "%s%d", cs.nick,
		getpid());

	    /* Introduce with the new nick */
	    intro_user (s_ChanServ, cs.user, cs.host, cs.real, cs.mode);

	    /* Let the staff know about it */
	    globops (s_ChanServ, ERR_COLLISIONS, cs.nick);

	    cs.collided = 0;

	    return;
	}
	else
	    cs.lastcoll = time (NULL);

	/* Re-introduce our client */
	intro_user (s_ChanServ, cs.user, cs.host, cs.real, cs.mode);
    }
    else if ((memoserv_on == TRUE) && !stricmp (av[0], s_MemoServ))
    {
	/* How many times we've been collided */
	++ms.collided;

	/* If this is true, we give up trying to retake our nick.. */
	if (((ms.lastcoll >= (time (NULL) - 5)) && (ms.collided >= 5)))
	{
	    /* Check if it's a server killing us. If it's not, then it's
	       likely a regular kill, and we don't want to take a waiting
	       nick.
	     */
	    if (!strchr (source, '.'))
	    {
		/* It's not a server. We'll reintro with the regular
		   nick and leave it at that.
		 */
		intro_user (s_MemoServ, ms.user, ms.host, ms.real, ms.mode);
		return;
	    }

	    /* Are we already waiting for our nick? Some jackass might be
	       killing us off. We don't want to end up with
	       RootServ903218973782743871412849 or something..
	       We'll only copy our current nick over if this isn't true.
	       That way we'll always end up with a new number.
	     */
	    if (!ms.wantnick)
	    {
		/* Make note of the nick we want */
		strscpy (ms.nick, s_MemoServ, sizeof (ms.nick));
	    }

	    /* Make note that we want it */
	    ms.wantnick = 1;

	    /* Give ourselves a new nick */
	    snprintf (s_MemoServ, sizeof (s_MemoServ), "%s%d", ms.nick,
		getpid());

	    /* Introduce with the new nick */
	    intro_user (s_MemoServ, ms.user, ms.host, ms.real, ms.mode);

	    /* Let the staff know about it */
	    globops (s_MemoServ, ERR_COLLISIONS, ms.nick);

	    ms.collided = 0;

	    return;
	}
	else
	    ms.lastcoll = time (NULL);

	/* Re-introduce our client */
	intro_user (s_MemoServ, ms.user, ms.host, ms.real, ms.mode);
    }
    else if ((globalnoticer_on == TRUE) && !stricmp (av[0], s_GlobalNoticer))
    {
	/* How many times we've been collided */
	++gn.collided;

	/* If this is true, we give up trying to retake our nick.. */
	if (((gn.lastcoll >= (time (NULL) - 5)) && (gn.collided >= 5)))
	{
	    /* Check if it's a server killing us. If it's not, then it's
	       likely a regular kill, and we don't want to take a waiting
	       nick.
	     */
	    if (!strchr (source, '.'))
	    {
		/* It's not a server. We'll reintro with the regular
		   nick and leave it at that.
		 */
		intro_user (s_GlobalNoticer, gn.user, gn.host, gn.real,
		    gn.mode);
		return;
	    }

	    /* Are we already waiting for our nick? Some jackass might be
	       killing us off. We don't want to end up with
	       RootServ903218973782743871412849 or something..
	       We'll only copy our current nick over if this isn't true.
	       That way we'll always end up with a new number.
	     */
	    if (!gn.wantnick)
	    {
		/* Make note of the nick we want */
		strscpy (gn.nick, s_GlobalNoticer, sizeof (gn.nick));
	    }

	    /* Make note that we want it */
	    gn.wantnick = 1;

	    /* Give ourselves a new nick */
	    snprintf (s_GlobalNoticer, sizeof (s_GlobalNoticer), "%s%d",
		gn.nick, getpid());

	    /* Introduce with the new nick */
	    intro_user (s_GlobalNoticer, gn.user, gn.host, gn.real, gn.mode);

	    /* Let the staff know about it */
	    globops (s_GlobalNoticer, ERR_COLLISIONS, gn.nick);

	    gn.collided = 0;

	    return;
	}
	else
	    gn.lastcoll = time (NULL);

	/* Re-introduce our client */
	intro_user (s_GlobalNoticer, gn.user, gn.host, gn.real, gn.mode);
    }
    else
	do_kill (source, ac, av);
}

/* Handle a nick collision */
static void m_collide (char *source, int ac, char **av)
{
    /* Before we reintroduce the user, we'll check to see if they've
       been collided more than 5 times in the last 5 seconds. This is
       a common problem with people who forget to rename/disable services
       with the same nicks as some of ours. If they HAVE been collided
       that many times, we'll give them a nick like RootServ986662577 and
       send a globops about it. We'll then watch for a nickchange/quit from
       the client with our nick, and retake it if the opportunity presents
       itself.
     */
    if ((rootserv_on == TRUE) && (!stricmp (av[1], s_RootServ)))
    {
	/* How many times we've been collided */
	++rs.collided;

	/* If this is true, we give up trying to retake our nick.. */
	if (((rs.lastcoll >= (time (NULL) - 5)) && (rs.collided >= 5)))
	{
	    /* Are we already waiting for our nick? Some jackass might be
	       killing us off. We don't want to end up with
	       RootServ903218973782743871412849 or something..
	       We'll only copy our current nick over if this isn't true.
	       That way we'll always end up with a new number.
	     */
	    if (!rs.wantnick)
	    {
		/* Make note of the nick we want */
		strscpy (rs.nick, s_RootServ, sizeof (rs.nick));
	    }

	    /* Make note that we want it */
	    rs.wantnick = 1;

	    /* Give ourselves a new nick */
	    snprintf (s_RootServ, sizeof (s_RootServ), "%s%d", rs.nick,
		getpid());

	    /* Introduce with the new nick */
	    intro_user (s_RootServ, rs.user, rs.host, rs.real, rs.mode);

	    /* Let the staff know about it */
	    globops (s_RootServ, ERR_COLLISIONS, rs.nick);

	    rs.collided = 0;

	    return;
	}
	else
	    rs.lastcoll = time (NULL);

	/* Re-introduce our client */
	intro_user (s_RootServ, rs.user, rs.host, rs.real, rs.mode);
    }
    else if ((nickserv_on == TRUE) && !stricmp (av[1], s_NickServ))
    {
	/* How many times we've been collided */
	++ns.collided;

	/* If this is true, we give up trying to retake our nick.. */
	if (((ns.lastcoll >= (time (NULL) - 5)) && (ns.collided >= 5)))
	{
	    /* Are we already waiting for our nick? Some jackass might be
	       killing us off. We don't want to end up with
	       RootServ903218973782743871412849 or something..
	       We'll only copy our current nick over if this isn't true.
	       That way we'll always end up with a new number.
	     */
	    if (!ns.wantnick)
	    {
		/* Make note of the nick we want */
		strscpy (ns.nick, s_NickServ, sizeof (ns.nick));
	    }

	    /* Make note that we want it */
	    ns.wantnick = 1;

	    /* Give ourselves a new nick */
	    snprintf (s_NickServ, sizeof (s_NickServ), "%s%d", ns.nick,
		getpid());

	    /* Introduce with the new nick */
	    intro_user (s_NickServ, ns.user, ns.host, ns.real, ns.mode);

	    /* Let the staff know about it */
	    globops (s_NickServ, ERR_COLLISIONS, ns.nick);

	    ns.collided = 0;

	    return;
	}
	else
	    ns.lastcoll = time (NULL);

	/* Re-introduce our client */
	intro_user (s_NickServ, ns.user, ns.host, ns.real, ns.mode);
    }
    else if ((chanserv_on == TRUE) && !stricmp (av[1], s_ChanServ))
    {
	/* How many times we've been collided */
	++cs.collided;

	/* If this is true, we give up trying to retake our nick.. */
	if (((cs.lastcoll >= (time (NULL) - 5)) && (cs.collided >= 5)))
	{
	    /* Are we already waiting for our nick? Some jackass might be
	       killing us off. We don't want to end up with
	       RootServ903218973782743871412849 or something..
	       We'll only copy our current nick over if this isn't true.
	       That way we'll always end up with a new number.
	     */
	    if (!cs.wantnick)
	    {
		/* Make note of the nick we want */
		strscpy (cs.nick, s_ChanServ, sizeof (cs.nick));
	    }

	    /* Make note that we want it */
	    cs.wantnick = 1;

	    /* Give ourselves a new nick */
	    snprintf (s_ChanServ, sizeof (s_ChanServ), "%s%d", cs.nick,
		getpid());

	    /* Introduce with the new nick */
	    intro_user (s_ChanServ, cs.user, cs.host, cs.real, cs.mode);

	    /* Let the staff know about it */
	    globops (s_ChanServ, ERR_COLLISIONS, cs.nick);

	    cs.collided = 0;

	    return;
	}
	else
	    cs.lastcoll = time (NULL);

	/* Re-introduce our client */
	intro_user (s_ChanServ, cs.user, cs.host, cs.real, cs.mode);
    }
    else if ((memoserv_on == TRUE) && !stricmp (av[1], s_MemoServ))
    {
	/* How many times we've been collided */
	++ms.collided;

	/* If this is true, we give up trying to retake our nick.. */
	if (((ms.lastcoll >= (time (NULL) - 5)) && (ms.collided >= 5)))
	{
	    /* Are we already waiting for our nick? Some jackass might be
	       killing us off. We don't want to end up with
	       RootServ903218973782743871412849 or something..
	       We'll only copy our current nick over if this isn't true.
	       That way we'll always end up with a new number.
	     */
	    if (!ms.wantnick)
	    {
		/* Make note of the nick we want */
		strscpy (ms.nick, s_MemoServ, sizeof (ms.nick));
	    }

	    /* Make note that we want it */
	    ms.wantnick = 1;

	    /* Give ourselves a new nick */
	    snprintf (s_MemoServ, sizeof (s_MemoServ), "%s%d", ms.nick,
		getpid());

	    /* Introduce with the new nick */
	    intro_user (s_MemoServ, ms.user, ms.host, ms.real, ms.mode);

	    /* Let the staff know about it */
	    globops (s_MemoServ, ERR_COLLISIONS, ms.nick);

	    ms.collided = 0;

	    return;
	}
	else
	    ms.lastcoll = time (NULL);

	/* Re-introduce our client */
	intro_user (s_MemoServ, ms.user, ms.host, ms.real, ms.mode);
    }
    else if ((globalnoticer_on == TRUE) && !stricmp (av[1], s_GlobalNoticer))
    {
	/* How many times we've been collided */
	++gn.collided;

	/* If this is true, we give up trying to retake our nick.. */
	if (((gn.lastcoll >= (time (NULL) - 5)) && (gn.collided >= 5)))
	{
	    /* Are we already waiting for our nick? Some jackass might be
	       killing us off. We don't want to end up with
	       RootServ903218973782743871412849 or something..
	       We'll only copy our current nick over if this isn't true.
	       That way we'll always end up with a new number.
	     */
	    if (!gn.wantnick)
	    {
		/* Make note of the nick we want */
		strscpy (gn.nick, s_GlobalNoticer, sizeof (gn.nick));
	    }

	    /* Make note that we want it */
	    gn.wantnick = 1;

	    /* Give ourselves a new nick */
	    snprintf (s_GlobalNoticer, sizeof (s_GlobalNoticer), "%s%d",
		gn.nick, getpid());

	    /* Introduce with the new nick */
	    intro_user (s_GlobalNoticer, gn.user, gn.host, gn.real, gn.mode);

	    /* Let the staff know about it */
	    globops (s_GlobalNoticer, ERR_COLLISIONS, gn.nick);

	    gn.collided = 0;

	    return;
	}
	else
	    gn.lastcoll = time (NULL);

	/* Re-introduce our client */
	intro_user (s_GlobalNoticer, gn.user, gn.host, gn.real, gn.mode);
    }
}

/* Handle a PROTOCTL command */
static void m_protoctl (char *source, int ac, char **av)
{
    int i, token = 0;
#ifdef UNREAL3
    int nickv2 = 0, umode2 = 0;
#endif

    for (i = 0; i < ac; i++)
    {
	if (!stricmp (av[i], "TOKEN"))
	    token = 1;
#ifdef UNREAL3
	if (!stricmp (av[i], "NICKv2"))
	    nickv2 = 1;

	if (!stricmp (av[i], "UMODE2"))
	    umode2 = 1;
#endif
    }

    if (token)
	me.token = 1;

#ifdef UNREAL3
    if (!nickv2 || !umode2)
    {
	send_cmd (NULL,
	    "ERROR :Incompatible IRCd (I'm compiled to support Unreal 3.1+)");
	snprintf (quitmsg, BUFSIZE,
	    "Incompatible IRCd (I'm compiled to support Unreal 3.1+)");
	runflags |= RUN_SHUTDOWN;
    }
#endif
}

/* Handle a MODE command */
static void m_mode (char *source, int ac, char **av)
{
    if (*av[0] == '#')
    {
	if (ac < 2)
	    return;
	do_cmode (source, ac, av);
    }
    else
    {
	if (ac != 2)
	    return;
	do_umode (source, ac, av);
    }
}

#ifdef UNREAL3
/* Handle an UnrealIRCd UMODE2 command */
static void m_umode2 (char *source, int ac, char **av)
{
    av[1] = sstrdup (av[0]);
    strcpy (av[0], source);

    do_umode (source, ac, av);
}

/* Change the users fake hostname */
static void m_sethost (char *source, int ac, char **av)
{
    User *u;

    if (ac != 1)
	return;

    u = finduser (source);

    if (!u)
	return;

    if (fakehost_on == TRUE && u->fakehost);
	free (u->fakehost);

    u->fakehost = sstrdup (av[0]);
}

/* CHGHOST. Similar to SETHOST, use the same function. */
static void m_chghost (char *source, int ac, char **av)
{
    if (ac == 2)
	m_sethost (av[0], ac-1, av+1);
}

/* Change a users ident (aka username) */
static void m_setident (char *source, int ac, char **av)
{
    User *u;

    if (ac != 1)
	return;

    u = finduser (source);

    if (!u)
	return;

    free (u->user);

    u->user = sstrdup (av[0]);
}

/* CHGIDENT. Similar to SETIDENT, use the same function. */
static void m_chgident (char *source, int ac, char **av)
{
    if (ac == 2)
	m_setident (av[0], ac-1, av+1);
}

/* Change a users realname. */
static void m_setname (char *source, int ac, char **av)
{
    User *u;

    if (ac != 1)
	return;

    u = finduser (source);

    if (!u)
        return;

    free (u->real);

    u->real = sstrdup (av[0]);
}

/* CHGNAME. Similar to SETNAME, use the same function. */
static void m_chgname (char *source, int ac, char **av)
{
    if (ac == 2)
	m_setname (av[0], ac-1, av+1);
}
#endif

#ifdef PROMETHEUS
/* Handle the ENDBURST command */
static void m_endburst (char *source, int ac, char **av)
{
    globops (s_RootServ, "Sync'd to network in %s: %d users, %d channels, %d servers.",
	duration (time (NULL) - burststart, 1), userstat, chancnt, servcnt + 1);

    bursting = 0;

    /* Now that the burst is done, run through every channel, and do the modechanges
       and topic changes as needed.
     */
    check_eb ();
}
#endif

/* Handle PRIVMSGs. */
static void m_privmsg (char *source, int ac, char **av)
{
    User *u;
    char buf[BUFSIZE], text[BUFSIZE];

    if (ac != 2)
	return;

    /* Run it through flood checks */
    if ((u = finduser (source)) && floodmsgs)
    {
	time_t now = time (NULL);

	/* If it's a channel PRIVMSG, ignore it. We have nothing to do with
	   channels.
	 */
	if (av[0][0] == '#')
	    return;

	/* Yes, its a user and not a server. Check if they're being
	   ignored.
	 */
	if (u->offences > 10)
	{
	    /* They're being ignored. Check if the ignore expired. */
	    if (now - u->lastmsg > 30)
	    {
		/* They're not being ignored anymore. */
		u->offences = u->offences - 10;
		u->lastmsg = now;
		u->msgs = 0;
	    }
	    else
		return;
	}

	if (now - u->lastmsg > floodtime)
	{
	    u->lastmsg = now;
	    u->msgs = 0;
	}

	u->msgs++;

	if (u->msgs > floodmsgs)
	{
	    /* They set off the trigger. Take action. */
	    if (!u->offences)
	    {
		/* First ignore. */
		u->lastmsg = now;
		u->msgs = 0;

		/* So we know we're ignoring them */
		u->offences = 11;

		/* Tell them. The stern voice of RootServ should be
		   an attention grabber. <g>
		 */
		notice (s_RootServ, source, RS_FLOOD_TRIGGERED);
		notice (s_RootServ, source, RS_FLOOD_TRIGGER, floodmsgs,
		    duration (floodtime, 2));
		notice (s_RootServ, source, RS_FLOOD_FIRST);

		/* Complain in GlobOps */
		if (noisyflood_on == TRUE)
		    globops (s_RootServ, RS_FLOOD_GLOBOPS, source, "");

		return;
	    }

	    if (u->offences == 1)
	    {
		/* Second ignore. */
		u->lastmsg = now;
		u->msgs = 0;

		/* So we know we're ignoring them */
		u->offences = 12;

		notice (s_RootServ, source, RS_FLOOD_TRIGGERED);
		notice (s_RootServ, source, RS_FLOOD_TRIGGER, floodmsgs,
		    duration (floodtime, 2));
		notice (s_RootServ, source, RS_FLOOD_SECOND);

		/* Complain in GlobOps */
		if (noisyflood_on == TRUE)
		    globops (s_RootServ, RS_FLOOD_GLOBOPS, source, "");

		return;
	    }

	    if (u->offences == 2)
	    {
		AutoKill *akill, **list;
		char mask[BUFSIZE], *username, *hostname;

		/* Third offence. AKill. */
		notice (s_RootServ, source, RS_FLOOD_TRIGGERED);
		notice (s_RootServ, source, RS_FLOOD_TRIGGER, floodmsgs,
		    duration (floodtime, 2));

		/* Complain in GlobOps */
		if (noisyflood_on == TRUE)
		    globops (s_RootServ, RS_FLOOD_GLOBOPS, source, " (Killed)");

		u->offences = 13;

		snprintf (mask, sizeof (mask), "*@%s", u->host);

		akill = scalloc (sizeof (AutoKill), 1);
		akill->mask = sstrdup (mask);
		akill->setter = sstrdup (s_RootServ);
		akill->realname = 0;
		akill->set = now;
		akill->expires = 1800;
		akill->reason = sstrdup (RS_AKILL_FLOOD);
		list = &akilllist[AHASH(akill->mask)];
		akill->next = *list;
		if (*list)
		    (*list)->prev = akill;
		*list = akill;

		username = sstrdup (akill->mask);
		hostname = strchr (username, '@');

		if (!hostname)
		    return;

		*hostname++ = 0;

#ifdef IRCUP9
		/* We use 999999999 because we'll expire it ourselves. */
		send_cmd (me.name, "GLINE * +%s@%s 999999999 :%s",
		    username, hostname, akill->reason);
#elif defined (BAHAMUT) || defined (PROMETHEUS)
		send_cmd (me.name, "AKILL %s %s 0 %s %ld :%s", hostname,
		    username, s_RootServ, time (NULL), akill->reason);
#else
		send_cmd (me.name, "%s %s %s :%s", me.token ? "V" : "AKILL",
		    hostname, username, akill->reason);
#endif
		free (username);

		return;
	    }
	}
    }

    /* RootServ? */
    if (stristr (av[0], s_RootServ))
    {
	/* What a nick@server for RootServ looks like */
	snprintf (buf, sizeof (buf), "%s@%s", s_RootServ, me.name);

	/* Do our checks on a copy of the original */
	snprintf (text, sizeof (text), av[1]);

	/* Check if we should deny a message via security settings */
	if (stricmp (av[0], buf) && ((securitysetting == 1 ||
	    (securitysetting == 2 && ispass (text)))
	    && !isctcp (text)))
	{
	    /* Tell them to use nick@server and bail */
	    if (haveserv_on == TRUE)
		notice (s_RootServ, source, ERR_SECURE_MESSAGES, "",
		    s_RootServ, "", "");
	    else
		notice (s_RootServ, source, ERR_SECURE_MESSAGES, "MSG ",
		    s_RootServ, "@", me.name);
	    return;
	}
	else
	{
	    /* Passed security checks, throw it through RootServ */
	    rootserv (source, av[1]);
	    return;
	}
    }

    /* NickServ? */
    if (stristr (av[0], s_NickServ))
    {
	/* What a nick@server for NickServ looks like */
	snprintf (buf, sizeof (buf), "%s@%s", s_NickServ, me.name);

	/* Do our checks on a copy of the original */
	snprintf (text, sizeof (text), av[1]);

	/* Check if we should deny a message via security settings */
	if (stricmp (av[0], buf) && ((securitysetting == 1 ||
	    (securitysetting == 2 && ispass (text)))
	    && !isctcp (text)))
	{
	    /* Tell them to use nick@server and bail */
	    if (haveserv_on == TRUE)
		notice (s_NickServ, source, ERR_SECURE_MESSAGES, "",
		    s_NickServ, "", "");
	    else
		notice (s_NickServ, source, ERR_SECURE_MESSAGES, "MSG ",
		    s_NickServ, "@", me.name);
	    return;
	}
	else
	{
	    /* Passed security checks, throw it through NickServ */
	    nickserv (source, av[1]);
	    return;
	}
    }

    /* ChanServ? */
    if (stristr (av[0], s_ChanServ))
    {
	/* What a nick@server for ChanServ looks like */
	snprintf (buf, sizeof (buf), "%s@%s", s_ChanServ, me.name);

	/* Do our checks on a copy of the original */
	snprintf (text, sizeof (text), av[1]);

	/* Check if we should deny a message via security settings */
	if (stricmp (av[0], buf) && ((securitysetting == 1 ||
	    (securitysetting == 2 && ispass (text)))
	    && !isctcp (text)))
	{
	    /* Tell them to use nick@server and bail */
	    if (haveserv_on == TRUE)
		notice (s_ChanServ, source, ERR_SECURE_MESSAGES, "",
		    s_ChanServ, "", "");
	    else
		notice (s_ChanServ, source, ERR_SECURE_MESSAGES, "MSG ",
		    s_ChanServ, "@", me.name);
	    return;
	}
	else
	{
	    /* Passed security checks, throw it through ChanServ */
	    chanserv (source, av[1]);
	    return;
	}
    }

    /* MemoServ? */
    if (stristr (av[0], s_MemoServ))
    {
	/* What a nick@server for MemoServ looks like */
	snprintf (buf, sizeof (buf), "%s@%s", s_ChanServ, me.name);

	/* Check if we should deny a message via security settings */
	if (stricmp (av[0], buf) && (securitysetting == 1 &&
	    !isctcp (av[1])))
	{
	    /* Tell them to use nick@server and bail */
	    if (haveserv_on == TRUE)
		notice (s_MemoServ, source, ERR_SECURE_MESSAGES, "",
		    s_MemoServ, "", "");
	    else
		notice (s_MemoServ, source, ERR_SECURE_MESSAGES, "MSG ",
		    s_MemoServ, "@", me.name);
	    return;
	}
	else
	{
	    /* Passed security checks, throw it through MemoServ */
	    memoserv (source, av[1]);
	    return;
	}
    }

    snprintf (buf, sizeof (buf), "%s@%s", s_GlobalNoticer, me.name);
    if (!stricmp (av[0], s_GlobalNoticer) || !stricmp (av[0], buf))
    {
	globalnoticer (source, av[1]);
	return;
    }
}

Message messages[] = {
    {"436",	m_collide},
    {"KICK",	m_kick},
    {"KILL",	m_kill},
    {"MODE",	m_mode},
    {"NICK",	m_nick},
    {"PART",	m_part},
    {"PING",	m_ping},
    {"PRIVMSG",	m_privmsg},
    {"JOIN",	m_join},
    {"QUIT",	m_quit},
    {"SERVER",	m_server},
    {"SQUIT",	m_squit},
    {"VERSION",	m_version},
    {"INFO",	m_info},
    {"WHOIS",	m_whois},
    {"STATS",	m_stats},
    {"MOTD",	m_motd},
    {"TIME",	m_time},
    {"ADMIN",	m_admin},
    {"AKILL",	m_akill},
    {"RAKILL",	m_rakill},
    {"TOPIC",	m_topic},
    {"PROTOCTL",m_protoctl},
    {"ERROR",	m_error},
    {"AWAY",	m_away},

    /* IRCd specific commands */
#ifdef UNREAL3
    {"NETINFO",	m_netinfo},
    {"UMODE2",	m_umode2},
    {"SETIDENT",m_setident},
    {"SETHOST",	m_sethost},
    {"SETNAME", m_setname},
    {"CHGHOST",	m_chghost},
    {"CHGIDENT",m_chgident},
    {"CHGNAME",	m_chgname},
#endif

#if defined (BAHAMUT) || defined (PROMETHEUS) || defined (UNREAL3)
    {"SJOIN",	m_sjoin},
#endif

#ifdef PROMETHEUS
    {"EB",	m_endburst},
#endif

    /* Tokens - DreamForge - DreamForge has an array of tokens defined,
       but only ever sends 5 tokenized commands, one of which we don't
       care about.
     */
    {"G",	m_mode},
    {"&",	m_nick},
    {"C",	m_join},
    {")",	m_topic},
    {"6",	m_away},

    /* Tokens - UnrealIRCd. - Unreal sends more than just these, but 
       these are the only ones we care about right now. Others will be
       added later as we add new features. The 4 above are the same on
       Unreal.
     */
#ifdef UNREAL3
    {"H",	m_kick},
    {".",	m_kill},
    {"+",	m_version},
    {"/",	m_info},
    {"-",	m_squit},
    {"#",	m_whois},
    {"2",	m_stats},
    {">",	m_time},
    {"@",	m_admin},
    {"F",	m_motd},
    {"D",	m_part},
    {"8",	m_ping},
    {",",	m_quit},
    {"'",	m_server},
    {"!",	m_privmsg},
    {"|",	m_umode2},
    {"AO",	m_netinfo},
    {"~",	m_sjoin},
#endif /* UNREAL3 */    
    {NULL}
};

Message *find_message (const char *name)
{
    Message *m;

    for (m = messages; m->name; m++)
	if (!stricmp (name, m->name))
	    return m;

    return NULL;
}
